/*
*   connect_gui.c     (part of the eDonkey2000 linux GUI)
*
*   Functions to
*    - show/destroy 'GUI connect' dialogue
*    - cleanup GUI variables after disconnect
*    - init GUI variables after connect
*
*   (c) 2001-2003 Tim-Philipp Mueller <t.i.m@orange.net>
*
*/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/* last code review: 2002-11-01 by Tim (the code in here is still a bit messy) */

#include "global.h"

#include "mainwindow.h"
#include "misc.h"
#include "misc_gtk.h"
#include "notebook.h"
#include "connect_gui.h"
#include "downloads.h"
#include "options.h"
#include "search.h"
#include "statusbar.h"
#include "status_page.h"
#include "shared_dirs.h"
#include "shared_files.h"
#include "shared_page.h"
#include "spawn_core.h"
#include "servers.h"
#include "toolbar.h"
#include "uploads.h"
#include "uploads_page.h"
#include "dns_lookup.h"
#include "gui_flags.h"
#include "misc_strings.h"
#include "guicoreconn.h"
#include "send_command.h"
#include "parse_message.h"
#include "packet.h"

#include "icons.h"

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>


#define QUICK_START_TEXT	\
		_(" * You have the GUI. Good. Now get the eDonkey2000 core from http://ed2k-gtk-gui.sf.net/core.shtml  \n"\
		  " * Download the archive into a subdirectory of your home directory (e.g. /home/joe/ed2k/). \n"\
		  " * Unpack it in there with 'tar xzf donkey_bla.tar.gz'. \n"\
		  " * Put the full path to the binary, e.g. '/home/joe/ed2k/donkey_59-gaps-4' into the 'path' field \n"\
		  " * Choose an arbitrary username and password. Avoid special characters. \n"\
		  " * The hostname should be 'localhost' and the port should be 4663. 'ping localhost' should work. \n"\
		  " * Hit the 'spawn' button. The dialog should display 'running' now. Hit the 'Go!' button. \n"\
		  " * You should now be connected to the core properly. Make sure that on the 'Shared' \n"\
		  "     page there is at least one directory listed. If that is so, you are ready to go. \n"\
		  " * Right-click on the server list and fetch an up-to-date list of servers and connect to one. \n\n"\
			\
		  " If this does not work, please look at the status page and consult the FAQ/documentation.")


enum
{
  DIALOG_RESPONSE_EXIT = 0,
  DIALOG_RESPONSE_HELP,
  DIALOG_RESPONSE_SPAWN,
  DIALOG_RESPONSE_CONNECT,
  DIALOG_RESPONSE_QUICK_HELP,
} ;


/* local variables */

static guint            num_port = 0;
static const gchar     *str_host = NULL;
static const gchar     *str_user = NULL;
static const gchar     *str_pass = NULL;
static const gchar     *str_path = NULL;

static gboolean         host_is_local = TRUE;


// Timer that is running while the 'connect to..'
//   dialog is up to show periodically if donkey
//   is running currently or not (on local machine)

static gint 		 connect_timer_handle = 0;


/* functions */

static void        gui_connect_onConnect (GtkWidget *widget, gpointer data);

static void        gui_connect_do_connect_check_pref_met (void);

static gboolean    gui_connect_do_connect (void);

static gint        gui_connect_find_binary (GtkWidget *entry);

static void        gui_connect_onSpawnCore (GtkWidget *widget, gpointer data);

static void        gui_connect_do_spawn (GtkWidget *dialog);

static void        gui_connect_onMakeDefaultToggled (GtkWidget *widget, gpointer data);

static void        gui_connect_onHostnameChanged (GtkWidget *widget, gpointer data);

static void        gui_connect_update_core_path (GtkWidget *dialog);

static void        gui_connect_onFileSelectorClicked (GtkWidget *widget, gpointer label);

static GtkWidget  *gui_connect_make_dialog_status_spawn_frame (GtkWidget *dialog);

static GtkWidget  *gui_connect_make_dialog (GtkWidget **qhframe);

static GtkWidget  *gui_connect_make_dialog_connect_frame (GtkWidget *dialog);


/******************************************************************************
 *
 *   clean_up_stuff
 *
 ***/

void
clean_up_stuff (void)
{
	show_restart_success_message_on_next_packet = 0;

	search_clean_up_all();
	downloads_cleanup();
	uploads_clear_all();
	servers_cleanup();
	shared_dirs_clear_list();
	shared_files_clear_list();

	status_page_print_clientID_info("");
}


/******************************************************************************
 *
 *   cleanup_after_disconnect
 *
 ***/

void
cleanup_after_disconnect (void)
{
	clean_up_stuff();

	if (about_to_quit)
		return;

	if (guicoreconn_do_reconnect)
	{
		guicoreconn_reconnect();
	}
	else
	{
		gui_connect_dialog();	// connect to dialog if auto-connect failed
	}

	return;
}





/******************************************************************************
 *
 *   gui_connect_find_binary
 *
 ***/

static gint
gui_connect_find_binary (GtkWidget *label)
{
	static gchar *path = NULL;

	g_return_val_if_fail ( label != NULL, 0 );

	G_FREE(path);

	path = g_strdup(find_donkey_binary());

	if (path)
		str_path = path;
	else
		str_path = NULL;

	return (gint)( path != NULL );	
}


/******************************************************************************
 *
 *   gui_connect_onSpawnCore
 *
 *   data: the 'connect to'-dialog widget
 *
 ***/

static void
gui_connect_onSpawnCore (GtkWidget *widget, gpointer data)
{
	gtk_dialog_response (GTK_DIALOG(data), DIALOG_RESPONSE_SPAWN);
}


/******************************************************************************
 *
 *   gui_connect_do_spawn
 *
 ***/

static void
gui_connect_do_spawn (GtkWidget *dialog)
{
	gchar  *dir, *pmfn;
	gchar  *msg_cp = NULL;
	gchar  *msg_ap = NULL;
	gint    ret;
	guint   new_clientport, new_aport, clientport;

	g_return_if_fail ( str_path  != NULL );
	g_return_if_fail ( *str_path != 0x00 );
	g_return_if_fail ( g_file_test(str_path, G_FILE_TEST_EXISTS)        == TRUE );
	g_return_if_fail ( g_file_test(str_path, G_FILE_TEST_IS_EXECUTABLE) == TRUE );
	g_return_if_fail ( g_file_test(str_path, G_FILE_TEST_IS_DIR)  == FALSE );

	if ( !str_user || !str_pass || !(*str_user) || !(*str_pass) )
	{
		misc_gtk_ok_dialog (GTK_MESSAGE_WARNING, _("Please enter a username and password first, so that I can set it up if necessary."));
		return;
	}

//	spawn_core_listening_port_is_in_use(9999);

	/* do entered username/password + port match with what's in the pref.met file */

	dir = g_path_get_dirname(str_path);

	g_return_if_fail ( dir != NULL );

	clientport = 4662;

	ret = spawn_core_pref_met_has_username (dir, str_user, str_pass, num_port, &clientport);

	if ( ret == -1 )
	{
		status_warning (_("GUI: some error occured in spawn_core_pref_met_has_username(), but will try to spawn anyway.\n"));
		(void) spawn_core_and_print_spawn_result_message(str_path);
		g_free(dir);
		return;
	}

	if (  ret == 1 )
	{
		status_message_blue (_("Good news - specified admin username, password and aport match the one in the pref.met file.\n"));
		(void) spawn_core_and_print_spawn_result_message(str_path);
		g_free(dir);
		return;
	}

	/* on or all of username/password/port do not match... */

	pmfn = g_strconcat(dir, G_DIR_SEPARATOR_S, "pref.met", NULL);
	g_return_if_fail ( pmfn != NULL );

	if ( !g_file_test(pmfn, G_FILE_TEST_EXISTS) )
	{
		status_message_blue (_("GUI: '%s' does not exist yet (will create it).\n"),pmfn);
	}
	else
	{
		status_message_blue (_("GUI: '%s' exists but with different name/pass/port (will change it).\n"),pmfn);
	}


	/* is some program already using the client port
	 *  that the donkey is going to use?              */

	new_clientport = clientport;

	if ( spawn_core_listening_port_is_in_use(clientport) )
	{
		new_clientport = clientport + 10;

		while ( spawn_core_listening_port_is_in_use(new_clientport) )
			new_clientport += 10;

		msg_cp = g_strdup_printf(_("\nSome other program is already using the\n"
		                           "port that the donkey wanted to listen on\n"
		                           "for connections from other clients (%u).\n"
		                           "I have changed this port to TCP port %u\n"
		                           "now. If you have port-forwarding configured\n"
		                           "somewhere, you will probably want to update that.\n\n"),
		                           clientport, new_clientport);
	}

	/* is some program already using the GUI admin port
	 *  that the donkey is going to use?                */

	new_aport = num_port;

	if ( spawn_core_listening_port_is_in_use(num_port)  ||  num_port == new_clientport )
	{
		GtkWidget *spin_button;

		new_aport = num_port + 9;

		while (spawn_core_listening_port_is_in_use(new_aport))
		{
			new_aport += 9;
		}

		msg_ap = g_strdup_printf(_("Some other program is already using the\n"
		                           "port that the donkey wanted to listen on\n"
		                           "for GUI connections (aport) (%u).\n"
		                           "I have changed this port to TCP port %u\n"
		                           "now. This is just for your information.\n\n"),
		                           num_port, new_aport);

		spin_button = g_object_get_data (G_OBJECT(dialog), "spin_button");

		if (spin_button)
			gtk_spin_button_set_value (GTK_SPIN_BUTTON(spin_button), new_aport*1.0);

		num_port = new_aport;
	}

	ret = spawn_core_modify_pref_met (dir, str_user, str_pass, new_aport, new_clientport);

	if (!ret)
	{
		/* failure */
		status_system_error_msg (_("Failed to change or create %s/pref.met.\n"));

		misc_gtk_ok_dialog (GTK_MESSAGE_WARNING,
			_("I noticed that the eDonkey2000 core's configuration\n"
			  "file that contains the admin username and password\n"
			  "either does not exist, or contains a different\n"
			  "username/password/port combination than the one you\n"
			  "specified in the 'connect to dialog'.\n\n"
			  "Therefore I tried to change the pref.met file so that\n"
			  "it contains the username/password/port that you have\n"
			  "just entered here, but I couldn't (reason: see statusbar).\n\n"
			  "This means that you will have to set up the admin username\n"
			  "and password for the GUI manually in the core, using the 'pass'\n"
			  "command. Make sure you change into the directory where the core\n"
			  "binary is before you do that (with 'cd /home/joe/ed2k/').\n\n"
			  "Please consult the FAQ/documentation for details.\n\n"
			  "Sorry it didn't work out.\n\n"
			  "Note: I will try to spawn a core anyway now.\n"));
	}
	else
	{
		/* success */
		status_message_blue (_("GUI: %s/pref.met changed or created successfully.\n"),dir);

		if ((msg_ap)||(msg_cp))
		{
			GtkWidget *dialog;

			dialog = gtk_message_dialog_new ( GTK_WINDOW(window),
			                                  GTK_DIALOG_DESTROY_WITH_PARENT,
			                                  GTK_MESSAGE_INFO, GTK_BUTTONS_OK,
			                                  "%s%s",
			                                  (msg_cp) ? msg_cp : "",
			                                  (msg_ap) ? msg_ap : "" );

			(void) gtk_dialog_run (GTK_DIALOG(dialog));

			gtk_widget_destroy(dialog);
		}
	}

	(void) spawn_core_and_print_spawn_result_message(str_path);

	G_FREE(msg_ap);
	G_FREE(msg_cp);
	G_FREE(dir);
	g_free(pmfn);
}


/******************************************************************************
 *
 *   gui_connect_onMakeDefaultToggled
 *
 ***/

static void
gui_connect_onMakeDefaultToggled (GtkWidget *widget, gpointer data)
{
	opt_toggle_bool(OPT_GUI_CONNECT_DIALOG_MAKE_VALUES_DEFAULT);
}


/******************************************************************************
 *
 *  gui_connect_onConnect
 *
 *   data: the 'connect to'-dialog widget
 *
 ***/

static void
gui_connect_onConnect (GtkWidget *widget, gpointer data)
{
	gtk_dialog_response (GTK_DIALOG(data), DIALOG_RESPONSE_CONNECT);
}


/******************************************************************************
 *
 *  gui_connect_do_connect_check_pref_met
 *
 ***/

static void
gui_connect_do_connect_check_pref_met (void)
{
	gchar *dir, *pmfn;

	if ( !str_path )
	{
		status_message_blue (_("GUI: You have not specified the path of the core binary. I can't double-check your settings then.\n"));
		return;
	}

	dir = g_path_get_dirname(str_path);

	g_return_if_fail ( dir != NULL );

	/* do username/password + port not match with what's in the pref.met file? */

	if (spawn_core_pref_met_has_username(dir, str_user, str_pass, num_port,NULL) == 1)
	{
		status_message_blue (_("Good news - specified admin username, password and aport match the one in the pref.met file.\n"));
		g_free(dir);
		return;
	}

	pmfn = g_strconcat(dir, G_DIR_SEPARATOR_S, "pref.met", NULL);
	g_return_if_fail ( pmfn != NULL );

	if ( !g_file_test(pmfn, G_FILE_TEST_EXISTS) )
	{
		status_message_blue (_("GUI: %s/pref.met does not exist yet in the core's directory.\n"),dir);
	}
	else
	{
		status_message_blue (_("GUI: %s/pref.met exists but with different name/pass/port!\n"),dir);
	}

	g_free(pmfn);
	g_free(dir);
}


/******************************************************************************
 *
 *  gui_connect_do_connect
 *
 *  returns TRUE if there are no problems so far, otherwise FALSE
 *
 ***/

static gboolean
gui_connect_do_connect (void)
{
	g_return_val_if_fail ( str_host != NULL && *str_host != 0x00, FALSE);
	g_return_val_if_fail ( str_user != NULL && *str_user != 0x00, FALSE);
	g_return_val_if_fail ( str_pass != NULL && *str_pass != 0x00, FALSE);

	/* only check pref.met if it's a local core */
	if ( host_is_local )
		gui_connect_do_connect_check_pref_met();

	/* make the entered values the default values? */
	if ( opt_get_bool(OPT_GUI_CONNECT_DIALOG_MAKE_VALUES_DEFAULT) )
	{
		opt_set_str(OPT_GUI_DEFAULT_CORE_HOST, str_host);
		opt_set_str(OPT_GUI_DEFAULT_CORE_USER, str_user);
		opt_set_str(OPT_GUI_DEFAULT_CORE_PASS, str_pass);

		if (str_path)
			opt_set_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH, str_path);

		opt_set_int(OPT_GUI_DEFAULT_CORE_PORT, num_port);
	}

	/* remove 'core running' timer for dialog label */
    g_source_remove (connect_timer_handle);

	status_message_blue (_("Preparing to connect to %s:%u\n"), str_host, num_port);

	/* Initiate connection to core */
	if (!guicoreconn_set_connection (str_host, num_port, str_user, str_pass, NULL))
	{
		g_printerr (_("Connect to core failed - couldn't initiate valid connection structure with default values.\n"));
		return FALSE;
	}

	if (!guicoreconn_connect())
	{
		g_printerr (_("Connect to core failed - connection structure hasn't been set up?!\n"));
		return FALSE;
	}

	return TRUE;
}


/******************************************************************************
 *
 *   gui_connect_onEntryChanged
 *
 *   updates variables if value has changed
 *
 *   data: either const gchar ** (for GtkEntry) or guint *  (for GtkSpinButton)
 *
 ***/

static void
gui_connect_onEntryChanged (GtkWidget *widget, gpointer data)
{
	g_return_if_fail ( data != NULL );

	if ( GTK_IS_SPIN_BUTTON(widget) )
	{
		guint *intport = (guint*)data;

		*intport = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON(widget));
	}

	else if ( GTK_IS_ENTRY(widget) )
	{
		*((const gchar **)data) = gtk_entry_get_text(GTK_ENTRY(widget));
	}

	else g_return_if_reached();
}




/******************************************************************************
 *
 *   connect_dialog_timer
 *
 *   'data' should be a pointer to the dialog widget
 *
 ***/

static gint
connect_dialog_timer (gpointer data)
{
	GtkWidget  *ok_button, *status_label;

	if ( CORE_IS_ALIVE )
		return 0; /* don't call us again in this case (dialog is about to be destroyed anyway) */

	g_return_val_if_fail ( data != NULL,       0 );
	g_return_val_if_fail ( GTK_IS_DIALOG(data), 0 );

	ok_button    = g_object_get_data (G_OBJECT(data), "ok_button");
	status_label = g_object_get_data (G_OBJECT(data), "status_label");

	g_return_val_if_fail ( GTK_IS_BUTTON(ok_button)   , 0 );
	g_return_val_if_fail ( GTK_IS_LABEL (status_label), 0 );

	if (ok_button)
	{
		gboolean sensitive = TRUE;

		if ( !str_host || !(*str_host) )
			sensitive = FALSE;

		else if ( !str_user || !(*str_user) )
			sensitive = FALSE;

		else if ( !str_pass || !(*str_pass) )
			sensitive = FALSE;

		gtk_widget_set_sensitive (GTK_WIDGET(ok_button), sensitive );
	}

	if ( !host_is_local )
	{
		gtk_label_set_text ( GTK_LABEL(status_label),
		                     UTF8_SHORT_PRINTF("%s", _(" Core on remote host ")));
		return 1;		/* do call us again (hostname might change) */
	}

	local_donkey_pid = donkey_process_running_check ();

	/* we shouldn't set the spawn button active/inactive due to the detection.
	 * people might want to run two cores on the same box or something
	 */
	if (local_donkey_pid < 0)
	{
		gtk_label_set_text ( GTK_LABEL(status_label),
		                     UTF8_SHORT_PRINTF("%s", _(" Could not check ")));
	}

	if ( local_donkey_pid == 0 )
	{
		gtk_label_set_markup (GTK_LABEL(status_label), misc_string_markup (_(" Not running "), "red", TRUE, FALSE));

		/* now this is a bit tricky... make GO! button insensitive
		 * if it's localhost and if we know that no core is running
		 * (assuming the core to be called 'donkey*'
		 */
		if ( host_is_local && (ok_button) )
			gtk_widget_set_sensitive (GTK_WIDGET(ok_button), FALSE);
	}

	if (local_donkey_pid > 0)
	{
		gchar *bname, *msg;

		bname = misc_get_basename_from_pid (local_donkey_pid,NULL);

		msg = g_strdup_printf ("\n %s (pid %d) \n\n %s \n", misc_string_markup (_("Running"), "darkGreen", TRUE, FALSE), local_donkey_pid, bname);

		if (msg)
			gtk_label_set_markup (GTK_LABEL(status_label), msg);

		G_FREE(msg);
		G_FREE(bname);
	}

	return 1;	/* do call us again */
}




/******************************************************************************
 *
 *   gui_connect_onHostnameChanged
 *
 *   This is not very sophisticated, but then why would anyone type 'berserker.dynisp.com'
 *   instead of 'localhost' if he/she means localhost? (well, if they do it's their fault,
 *   I don't care...)
 *
 *   checks if the hostname/IP in the hostname widget is localhost or not and sets
 *   the path to binary field active/inactive according to that.
 *
 ***/

static void
gui_connect_onHostnameChanged (GtkWidget *widget, gpointer dialog)
{
	GtkWidget *fsbutton;

	fsbutton = g_object_get_data (G_OBJECT(dialog), "fsbutton");

	gui_connect_onEntryChanged (widget,&str_host);

	host_is_local = FALSE;

	if ( str_host && (*str_host) )
	{

		/* in these cases UTF-8 == ASCII */
		if ( strcmp(str_host,"localhost") == 0 )
			host_is_local = TRUE;

		if ( g_ascii_strncasecmp(str_host, "127.0.0.", 8) == 0 )
			host_is_local = TRUE;
	}

	/* will not be set the first time we are called */
	if (fsbutton)
	{
		g_return_if_fail ( GTK_IS_BUTTON(fsbutton)  );
		gtk_widget_set_sensitive(fsbutton, host_is_local );
	}

	gui_connect_update_core_path(dialog);

	return;

/*
 * Let's try not to be too smart and annoy people who want to do nifty stuff
 * like ssh-forwarding GUI connections etc.
 *
	if (!host_is_local)
	{
		// note: regardless of what the documentation says - at least up until gnet 1.1.7,
		// gnet_inetaddr_list_interfaces() does _not_ list loopback interfaces (see source)
		//
		GList *if_list, *node;

		if_list = gnet_inetaddr_list_interfaces();

		for ( node = if_list; node != NULL; node = node->next)
		{
			GInetAddr *item = (GInetAddr*)node->data;

			if (!host_is_local)
			{
				const gchar *cname = gnet_inetaddr_get_canonical_name(item);
				if (cname)
				{
					if (g_strcasecmp(cname,hostname)==0)
						host_is_local = TRUE;
				}
			}
			gnet_inetaddr_delete(item);
		}
		g_list_free(if_list);
	}
*/

}



/******************************************************************************
 *
 *   gui_connect_update_core_path
 *
 ***/

static void
gui_connect_update_core_path (GtkWidget *dialog)
{
	GtkWidget *pathlabel, *spawn_button;

	g_return_if_fail ( dialog != NULL );
	g_return_if_fail ( GTK_IS_DIALOG(dialog) );

	pathlabel  = g_object_get_data (G_OBJECT(dialog), "path_label");

	/* might not be set yet when we're called the first time
	 * (due to hostname entry changed signal when creating the hostname entry field)  */

	if (!pathlabel)
		return;

	g_return_if_fail ( GTK_IS_LABEL (pathlabel) );

	spawn_button = g_object_get_data (G_OBJECT(dialog), "spawn_button");

	if ( !str_path )
	{
		gtk_label_set_markup ( GTK_LABEL(pathlabel),
			UTF8_SHORT_PRINTF("<span size=\"x-small\">%s</span>",
				(host_is_local) ? _("couldn't find binary") : _("none (remote host)")));

		/* might not be set yet when we're called the first time */
		if (spawn_button)
		{
			g_return_if_fail ( GTK_IS_BUTTON(spawn_button) );
			gtk_widget_set_sensitive(spawn_button, FALSE);
		}
	}
	else
	{
		gtk_label_set_markup ( GTK_LABEL(pathlabel),
			UTF8_SHORT_PRINTF("<span size=\"x-small\">%s</span>", str_path));

		/* might not be set yet when we're called the first time */
		if (spawn_button)
		{
			g_return_if_fail ( GTK_IS_BUTTON(spawn_button) );
			gtk_widget_set_sensitive(spawn_button, TRUE);
		}
	}
}


/******************************************************************************
 *
 *   gui_connect_onFileSelectorClicked
 *
 ***/

static void
gui_connect_onFileSelectorClicked (GtkWidget *widget, gpointer dialog)
{
	static gchar   *fname = NULL;
	GtkWidget      *fs;
	gboolean        quit = FALSE;
	gint            ret = 0;

	g_return_if_fail ( dialog != NULL );


	fs = gtk_file_selection_new(UTF8_SHORT_PRINTF("%s",_("select core executable ('donkey_s' or 'overnet_s')")));

	if (opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH))
	{
		gtk_file_selection_set_filename(GTK_FILE_SELECTION(fs), opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH));
	}
	else
	{
		if (fname)
			gtk_file_selection_set_filename(GTK_FILE_SELECTION(fs),fname);
	}

	gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(fs));

	gtk_window_set_transient_for(GTK_WINDOW(fs), GTK_WINDOW(dialog));

	while ( !quit )
	{
		G_FREE(fname);

		ret = gtk_dialog_run(GTK_DIALOG(fs));

		if ( ret == GTK_RESPONSE_CANCEL  ||  ret == GTK_RESPONSE_DELETE_EVENT )
		{
			quit = TRUE;
			continue;
		}

		fname = g_strdup(gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)));

		g_return_if_fail ( fname != NULL );

		if (g_file_test(fname, G_FILE_TEST_IS_DIR))
		{
			misc_gtk_ok_dialog ( GTK_MESSAGE_WARNING, 
			                     _(" You selected a directory.\n\n"
			                       " Please select the core program executable itself.\n"
			                       " (e.g. /home/joe/ed2k/donkey_s)"));

			continue; /* continue running the dialog */
		}

		if (!g_file_test(fname, G_FILE_TEST_IS_EXECUTABLE))
		{
			misc_gtk_ok_dialog ( GTK_MESSAGE_WARNING, _(" The file you selected is not executable. "));

			continue; /* continue running the dialog */
		}

		/* everything's fine. Let's overwrite previous error messages that might be there */

		statusbar_extra_message(" ");

		str_path = fname;

		gui_connect_update_core_path(dialog);

		quit = TRUE;
	}

	gtk_grab_remove(fs);
	gtk_widget_destroy(fs);
}

/******************************************************************************
 *
 *   gui_connect_onHostnameActivated
 *
 *   <Enter> has been pressed in the hostname entry widget
 *
 ***/

static void
gui_connect_onHostnameActivated (GtkWidget *hostentry, gpointer dialog)
{
	if (dialog)
	{
		GtkWidget *okbutton;

		okbutton = g_object_get_data (G_OBJECT(dialog), "ok_button");

		if ((okbutton) && GTK_WIDGET_SENSITIVE(okbutton))
			gtk_button_clicked(GTK_BUTTON(okbutton));
	}
}



/******************************************************************************
 *
 *   gui_connect_make_dialog_connect_frame
 *
 ***/

static GtkWidget *
gui_connect_make_dialog_connect_frame (GtkWidget *dialog)
{
	GtkWidget    *cframe, *cframelabel, *vbox, *hbox, *hbuttonbox, *ok_button;
	GtkWidget    *cb_makedefault, *label, *entry, *fsbutton, *hbox2;
	GtkSizeGroup *sg_labels, *sg_entries;


	/* frame and frame label */
	cframe = gtk_frame_new (NULL);

	cframelabel = gtk_label_new (NULL);
	gtk_label_set_use_markup (GTK_LABEL (cframelabel), TRUE);
	gtk_label_set_markup (GTK_LABEL (cframelabel), misc_string_markup( _("Connect to donkey core"), "blue", TRUE, FALSE));
	gtk_frame_set_label_widget (GTK_FRAME (cframe), cframelabel);
	gtk_container_set_border_width (GTK_CONTAINER(cframe),5);

	vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER (cframe), vbox);
	gtk_container_set_border_width (GTK_CONTAINER(vbox),5);

	sg_labels  = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
	sg_entries = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);

	/* host */

	hbox = gtk_hbox_new (FALSE, 10);

	label = gtk_label_new(UTF8_SHORT_PRINTF("%s", _("host of running core ")));
	gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 0);
	gtk_size_group_add_widget(sg_labels, label);

	entry = gtk_entry_new();
	gtk_size_group_add_widget(sg_entries, entry);

	g_signal_connect(entry, "changed", G_CALLBACK(gui_connect_onHostnameChanged), dialog);
	g_signal_connect(entry, "activate", G_CALLBACK(gui_connect_onHostnameActivated), dialog);

	/* this will set 'host_is_local' from 'changed' callback */
	if (opt_get_str(OPT_GUI_DEFAULT_CORE_HOST))
		gtk_entry_set_text (GTK_ENTRY(entry), opt_get_str(OPT_GUI_DEFAULT_CORE_HOST));
	else
		gtk_entry_set_text (GTK_ENTRY(entry), "localhost");

	gtk_box_pack_start (GTK_BOX(hbox), entry, TRUE, TRUE, 0);

	gtk_box_pack_start (GTK_BOX(vbox), hbox, TRUE, TRUE, 10);

	/* port */

	hbox = gtk_hbox_new (FALSE, 10);

	label = gtk_label_new(UTF8_SHORT_PRINTF("%s", _("core port")));
	gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 0);
	gtk_size_group_add_widget(sg_labels, label);

	entry = gtk_spin_button_new (GTK_ADJUSTMENT(gtk_adjustment_new(opt_get_int(OPT_GUI_DEFAULT_CORE_PORT),10,65536,1.0,1.0,1.0)),1,0);

	gtk_size_group_add_widget(sg_entries, entry);
	g_signal_connect ( G_OBJECT (entry), "value-changed",
	                   G_CALLBACK(gui_connect_onEntryChanged),
	                   &num_port );
	g_object_set_data (G_OBJECT(dialog), "spin_button", entry);

	gtk_spin_button_set_value (GTK_SPIN_BUTTON(entry),
	                           opt_get_int(OPT_GUI_DEFAULT_CORE_PORT)*1.0);

	num_port = opt_get_int(OPT_GUI_DEFAULT_CORE_PORT);

	gtk_box_pack_start (GTK_BOX(hbox), entry, TRUE, TRUE, 0);

	gtk_box_pack_start (GTK_BOX(vbox), hbox, TRUE, TRUE, 10);

/*								_("This is the port the core listens on for GUI connections. "
								  "The standard port for this is 4663. Note: the core will only "
								  "listen for GUIs if it is started with the '-' option (in case "
								  "you start the core yourself and not via the GUI."));
*/


	/* user */

	hbox = gtk_hbox_new (FALSE, 10);
	label = gtk_label_new(UTF8_SHORT_PRINTF("%s", _("core username")));
	gtk_size_group_add_widget(sg_labels, label);
	gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 0);

	entry = gtk_entry_new();
	gtk_size_group_add_widget(sg_entries, entry);

	g_signal_connect ( entry, "changed",
	                   G_CALLBACK(gui_connect_onEntryChanged),
	                   &str_user );

	if (opt_get_str(OPT_GUI_DEFAULT_CORE_USER))
		gtk_entry_set_text (GTK_ENTRY(entry), UTF8_SHORT_PRINTF("%s", opt_get_str(OPT_GUI_DEFAULT_CORE_USER)));

	gtk_box_pack_start (GTK_BOX(hbox), entry, TRUE, TRUE, 0);

	gtk_box_pack_start (GTK_BOX(vbox), hbox, TRUE, TRUE, 10);



	/* password */

	hbox = gtk_hbox_new (FALSE, 10);
	label = gtk_label_new(UTF8_SHORT_PRINTF("%s", _("core password")));
	gtk_size_group_add_widget(sg_labels, label);
	gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 0);

	entry = gtk_entry_new();
	gtk_size_group_add_widget(sg_entries, entry);
	g_signal_connect ( entry, "changed",
	                   G_CALLBACK(gui_connect_onEntryChanged),
	                   &str_pass );

	if (opt_get_str(OPT_GUI_DEFAULT_CORE_PASS))
		gtk_entry_set_text (GTK_ENTRY(entry), UTF8_SHORT_PRINTF("%s",opt_get_str(OPT_GUI_DEFAULT_CORE_PASS)));

	gtk_entry_set_visibility (GTK_ENTRY(entry), FALSE); /* hide password cleartext */
	gtk_box_pack_start (GTK_BOX(hbox), entry, TRUE, TRUE, 0);

	gtk_box_pack_start (GTK_BOX(vbox), hbox, TRUE, TRUE, 10);


	/* core path */

	if (opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH))
	{
		/* Has configured binary vanished/been deleted since last start?
		 * If yes, reset to empty path.                                  */
		if ( access (opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH), F_OK|R_OK|X_OK) )
		{
			opt_set_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH, NULL);
		}
	}

	hbox = gtk_hbox_new (FALSE, 10);
	label = gtk_label_new(UTF8_SHORT_PRINTF("%s", _("path to local core")));
	gtk_size_group_add_widget(sg_labels, label);
	gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 0);

	hbox2 = gtk_hbox_new (FALSE, 5);
	gtk_size_group_add_widget(sg_entries, hbox2);
	gtk_box_pack_start (GTK_BOX(hbox), hbox2, TRUE, TRUE, 0);

	label = gtk_label_new(NULL);
	g_object_set_data (G_OBJECT(dialog), "path_label", label);

	/* if we don't have a path to the donkey binary then try to find it */
	if ( !opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH) || !(*opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH)) )
	{
		gui_connect_find_binary(label);
	}
	else
	{
		str_path = opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH);
	}

	gui_connect_update_core_path(dialog);

	gtk_box_pack_start (GTK_BOX(hbox2), label, TRUE, TRUE, 0);

	fsbutton = gtk_button_new_with_label (" ... ");
	gtk_box_pack_start (GTK_BOX(hbox2), fsbutton, FALSE, FALSE, 0);
	g_signal_connect ( fsbutton, "clicked",
	                   G_CALLBACK(gui_connect_onFileSelectorClicked),
	                   dialog );

	g_object_set_data (G_OBJECT(dialog), "fsbutton", fsbutton);

	gtk_box_pack_start (GTK_BOX(vbox), hbox, TRUE, TRUE, 10);



	/* save as default values */

	hbox = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);

	cb_makedefault = gtk_check_button_new_with_label (UTF8_SHORT_PRINTF("%s",_("save as default values")));

	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(cb_makedefault),
	                              opt_get_bool(OPT_GUI_CONNECT_DIALOG_MAKE_VALUES_DEFAULT));

	g_signal_connect (cb_makedefault, "toggled", G_CALLBACK(gui_connect_onMakeDefaultToggled), NULL);
	gtk_box_pack_start (GTK_BOX (hbox), cb_makedefault, FALSE, FALSE, 5);

	hbuttonbox = gtk_hbutton_box_new ();
	gtk_box_pack_start (GTK_BOX (hbox), hbuttonbox, TRUE, TRUE, 5);
	gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox), GTK_BUTTONBOX_END);

	ok_button = gtk_button_new();
	gtk_container_add( GTK_CONTAINER(ok_button),
                       make_icon_and_label_hbox(misc_string_markup (" GO! ", "blue", TRUE, FALSE), ICON_NUKE));
	gtk_container_add (GTK_CONTAINER (hbuttonbox), ok_button);

	g_signal_connect( ok_button, "clicked", G_CALLBACK (gui_connect_onConnect), dialog);

	g_object_set_data (G_OBJECT(dialog), "ok_button", ok_button);

	gtk_widget_show_all (cframe);

	return cframe;
}


/******************************************************************************
 *
 *   gui_connect_make_dialog_status_spawn_frame
 *
 ***/

static GtkWidget *
gui_connect_make_dialog_status_spawn_frame (GtkWidget *dialog)
{
	GtkWidget *sframe, *sframelabel, *hbox, *hbuttonbox, *slabel, *spawn_button;

	sframe = gtk_frame_new(NULL);
	sframelabel = gtk_label_new(NULL);
	gtk_label_set_use_markup (GTK_LABEL (sframelabel), TRUE);
	gtk_label_set_markup (GTK_LABEL (sframelabel), misc_string_markup( _("core status"), "blue", TRUE, FALSE));
	gtk_frame_set_label_widget (GTK_FRAME (sframe), sframelabel);
	gtk_container_set_border_width (GTK_CONTAINER(sframe),5);

	hbox = gtk_hbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER (sframe), hbox);

	slabel = gtk_label_new (NULL);
	gtk_label_set_use_markup (GTK_LABEL (slabel), TRUE);
	gtk_box_pack_start(GTK_BOX(hbox), slabel, TRUE, FALSE, 0);

	g_object_set_data (G_OBJECT(dialog), "status_label", slabel);

	connect_timer_handle = g_timeout_add (1000, connect_dialog_timer, dialog);

	hbuttonbox = gtk_hbutton_box_new ();
	gtk_box_pack_start (GTK_BOX (hbox), hbuttonbox, TRUE, TRUE, 0);
	gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox), GTK_BUTTONBOX_END);

	spawn_button = gtk_button_new();

	gtk_container_add ( GTK_CONTAINER(spawn_button),
                        make_icon_and_label_hbox(_(" Spawn core "), ICON_CORE));

	g_signal_connect( spawn_button, "clicked", G_CALLBACK(gui_connect_onSpawnCore), dialog );

	g_object_set_data (G_OBJECT(dialog), "spawn_button", spawn_button);

	gtk_container_add (GTK_CONTAINER (hbuttonbox), spawn_button);

	gtk_widget_show_all(sframe);

	return sframe;
}

/******************************************************************************
 *
 *   gui_connect_make_dialog
 *
 ***/

static GtkWidget *
gui_connect_make_dialog (GtkWidget **qhframe)
{
	GtkWidget *help_button, *quickhelp_button, *exit_button;
	GtkWidget *dialog, *cframe, *sframe, *qhlabel;
	gchar     *qh_utf8, *qh_markup_utf8;

	dialog = gtk_dialog_new_with_buttons ( UTF8_SHORT_PRINTF("%s",_("connect to donkey core")),
	                                       GTK_WINDOW(window),
	                                       GTK_DIALOG_DESTROY_WITH_PARENT,
	                                       NULL );

	gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_CENTER);
	gtk_window_set_resizable(GTK_WINDOW (dialog), FALSE);

	/* frames */

	cframe = gui_connect_make_dialog_connect_frame(dialog);
	gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->vbox), cframe, TRUE, TRUE, 0);

	sframe = gui_connect_make_dialog_status_spawn_frame(dialog);
	gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->vbox), sframe, TRUE, TRUE, 0);

	*qhframe = gtk_frame_new (NULL);
	gtk_widget_hide (*qhframe);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG(dialog)->vbox), *qhframe, TRUE, TRUE, 0);
	gtk_container_set_border_width (GTK_CONTAINER(*qhframe),5);

	qhlabel = gtk_label_new (NULL);
	qh_utf8 = TO_UTF8(QUICK_START_TEXT);
	qh_markup_utf8 = g_strconcat ("<span size=\"small\">", qh_utf8, "</span>", NULL);
	gtk_label_set_justify(GTK_LABEL(qhlabel),GTK_JUSTIFY_LEFT);
	gtk_label_set_markup(GTK_LABEL(qhlabel), qh_markup_utf8);
	gtk_container_add (GTK_CONTAINER (*qhframe), qhlabel);
	gtk_widget_show(qhlabel);

	/* set up bottom row of buttons */

	help_button = gtk_button_new();
	gtk_container_add ( GTK_CONTAINER(help_button),
	                    make_icon_and_label_hbox(_(" Help "), ICON_ABOUT));
	gtk_widget_show (help_button);
	gtk_dialog_add_action_widget(GTK_DIALOG(dialog), help_button, DIALOG_RESPONSE_HELP );

	quickhelp_button = gtk_button_new();
	gtk_container_add ( GTK_CONTAINER(quickhelp_button),
	                    make_icon_and_label_hbox(_(" Quick start "), ICON_ABOUT));
	gtk_widget_show (quickhelp_button);
	gtk_dialog_add_action_widget(GTK_DIALOG(dialog), quickhelp_button, DIALOG_RESPONSE_QUICK_HELP );

	exit_button = gtk_button_new();
	gtk_container_add ( GTK_CONTAINER(exit_button),
	                    make_icon_and_label_hbox(_(" Exit "), ICON_MENU_QUIT));
	gtk_widget_show (exit_button);
	gtk_dialog_add_action_widget(GTK_DIALOG(dialog), exit_button, DIALOG_RESPONSE_EXIT );

	G_FREE(qh_utf8);
	G_FREE(qh_markup_utf8);

	return dialog;
}



/******************************************************************************
 *
 *   gui_connect_dialog
 *
 *   Displays the 'connect to...' dialog and makes it transient
 *
 ******************************************************************************/

void
gui_connect_dialog (void)
{
	static gboolean  busy = FALSE; /* make sure there is no recursion */
	GtkWidget       *dialog, *qhframe;
	gboolean         quit_dialog = FALSE;
	gint             ret = 0;

	if ( about_to_quit )
		return;

	if ( CORE_IS_ALIVE )	/* XXX - ??? - was: g_return_if_fail ( !CORE_IS_ALIVE ) */
		return;

	g_return_if_fail ( busy == FALSE  );

	if (!window)
		return;

	busy = TRUE;

	notebook_set_page(NOTEBOOK_PAGE_STATUS);

	dialog = gui_connect_make_dialog(&qhframe);

	gui_connect_update_core_path(dialog);

	/* run dialog until 'exit' or 'connect' response */

	while( !quit_dialog )
	{
		ret = gtk_dialog_run(GTK_DIALOG(dialog));

		switch (ret)
		{
			case GTK_RESPONSE_DELETE_EVENT:

				statusbar_extra_message (_(" GUI: Once you are connected to a core, you'll get rid of this dialog."));

			break;


			case DIALOG_RESPONSE_HELP:

				invoke_browser_with_url (DOCS_URL);

			break;


			case DIALOG_RESPONSE_QUICK_HELP:
			{
				if (GTK_WIDGET_VISIBLE(qhframe))

					gtk_widget_hide (qhframe);

				else

					gtk_widget_show (qhframe);
			}
			break;


			case DIALOG_RESPONSE_EXIT:

				gtk_grab_remove (GTK_WIDGET(dialog));

				g_source_remove (connect_timer_handle);

				quit_dialog = TRUE;

			break;

			case DIALOG_RESPONSE_CONNECT:
			{
				if (gui_connect_do_connect())
					quit_dialog = TRUE;
			}
			break;

			case DIALOG_RESPONSE_SPAWN:
				gui_connect_do_spawn(dialog);
			break;
		}
	}


	gtk_widget_destroy (dialog);
	dialog = NULL;

	/* Exit? => use main window quit function */

	if ( ret == DIALOG_RESPONSE_EXIT )
	{
		about_to_quit = TRUE;

		mainwindow_retrieve_some_options_before_main_window_is_destroyed();

		gtk_widget_destroy(window);
	}

	busy = FALSE;

	return;
}


