/***************************************************************************
                          downloads.c  -  description
                             -------------------
    begin                : Sun Feb 9 2003
    copyright            : (C) 2003 by Tim-Philipp Mller
    email                : t.i.m@orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "blacklist.h"
#include "downloads.h"
#include "downloads_popup.h"
#include "downloads_attrib.h"
#include "icons.h"
#include "global.h"
#include "gui_flags.h"
#include "guicoreconn.h" /* for CORE_IS_LOCAL */
#include "options.h"
#include "misc.h"
#include "misc_gtk.h"
#include "misc_strings.h"
#include "parse_message.h"
#include "send_command.h"
#include "status_page.h"
#include "statusbar.h"

/* for stat(): */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

/* sort IDs for our custom sort function */
enum
{
	DL_SORT_ID_FILENAME = 0,
	DL_SORT_ID_SIZE,
	DL_SORT_ID_STATUS,
	DL_SORT_ID_TRANSFERED,
	DL_SORT_ID_BYTESLEFT,
	DL_SORT_ID_SPEED,
	DL_SORT_ID_TIMELEFT,
	DL_SORT_ID_PRIORITY,
	DL_SORT_ID_AVAILABILITY,
	DL_SORT_ID_SOURCES,
	DL_SORT_ID_LAST_SEEN_DLING, /* this one we abuse for the cell data func as well */
	DL_SORT_ID_LAST_SEEN_CMPLT, /* this one we abuse for the cell data func as well */
	DL_SORT_ID_LAST
};

#define DOWNLOADS_TYPES_PLACEHOLDER \
	G_TYPE_POINTER, /* a pointer to our Download struct */\
	G_TYPE_STRING,  /* DOWNLOADS_FILEEXTENSION_COLUMN   */\
	G_TYPE_STRING,  /* DOWNLOADS_FILETYPE_COLUMN        */\
	G_TYPE_STRING   /* DOWNLOADS_SIZE_STRING_COLUMN     */

enum
{
	DL_RECORD = 0,
	DOWNLOADS_FILEEXTENSION_COLUMN,
	DOWNLOADS_FILETYPE_COLUMN,
	DOWNLOADS_SIZE_STRING_COLUMN,
	DOWNLOADS_N_COLUMNS
};

#define DOWNLOADS_MAX_PROGRESSBAR_WIDTH   2000

GtkListStore             *downloads_store = NULL;       /* exported only to downloads_popup.c */
GtkWidget                *downloads_view = NULL;        /* exported only to downloads_popup.c */
static GtkTreeModel      *downloads_store_sort = NULL;

static GHashTable        *downloads_ht; /* NULL */      /* filehash as keys, Download records as values */

static guint              dlview_font_height; /* 0 */

gfloat                    total_downloaded = 0;
time_t                    seconds_download_was_nonzero = 1; /* so we don't divide by zero */

static time_t             downloads_last_dl_removed; /* 0 */

static GtkWidget         *downloads_page_stats_label; /* NULL */
static GtkWidget         *downloads_page_fd_label;    /* NULL */

static gboolean           downloads_option_cache_hide_junk;        /* FALSE */
static gboolean           downloads_option_cache_trans_in_percent; /* FALSE */

/* we store these, so we can retrieve the current width and
   heigth of the column for drawing the progrss bars */

static GtkTreeViewColumn *downloads_progressbar_column;   /* NULL */

static GtkCellRenderer   *downloads_filename_renderer;    /* NULL */

static GHashTable        *downloads_id_ht; /* NULL */   /* maps index/ID <-> Download record */

static gchar            **status_utf8 = NULL; /* translated status strings in UTF-8 */


/* functions */

static void       download_make_progressbar (GtkTreeViewColumn *column, GtkCellRenderer *renderer, Download *dl);


/******************************************************************************
 *
 *  downloads_cleanup
 *
 *  clears the downloads list completely
 *
 ***/

void
downloads_cleanup(void)
{
	if (downloads_store)
	{
		GSList *all, *node;

		all = downloads_get_all();

		for ( node = all;  node != NULL;  node = node->next )
			download_remove((Download*)node->data);

		g_slist_free(all);
	}

	if (downloads_ht)
		g_hash_table_destroy (downloads_ht);

	if (downloads_id_ht)
		g_hash_table_destroy (downloads_id_ht);

	downloads_ht = g_hash_table_new_full ( misc_hash_hash, misc_hash_equal, g_free, NULL);

	downloads_id_ht = g_hash_table_new(g_direct_hash, g_direct_equal);
}


/******************************************************************************
 *
 *  downloads_view_save_settings
 *
 *  saves downloads view settings: column widths and sort column
 *
 ***/

void
downloads_view_save_settings (void)
{
	misc_gtk_save_tree_view_settings ( downloads_view,
	                                   downloads_store,
	                                   OPT_GUI_COLUMN_WIDTHS_DOWNLOADS,
	                                   OPT_GUI_SORT_TYPE_DOWNLOADS,
	                                   OPT_GUI_SORT_COL_DOWNLOADS);
}


/******************************************************************************
 *
 *  downloads_view_restore_settings
 *
 *  Note: setting column widths only seems to work after the
 *        widget has been realised.
 *
 ***/

static void
downloads_view_restore_settings (void)
{
	misc_gtk_restore_tree_view_settings ( downloads_view,
	                                      downloads_store,
	                                      OPT_GUI_COLUMN_WIDTHS_DOWNLOADS,
	                                      OPT_GUI_SORT_TYPE_DOWNLOADS,
	                                      OPT_GUI_SORT_COL_DOWNLOADS,
	                                      DL_SORT_ID_FILENAME );
}


/******************************************************************************
 *
 *  downloads_get_num
 *
 *  returns number of downloads (excluding 'completed' items)
 *
 ***/

gint
downloads_get_num (void)
{
	g_return_val_if_fail ( downloads_id_ht != NULL, 0 );
	return g_hash_table_size(downloads_id_ht);
}


/******************************************************************************
 *
 *  download_free_download
 *
 *  Frees a download structure and all its content.
 *
 ***/

static void
download_free_download (Download *dl)
{
	g_return_if_fail ( dl != NULL );

	G_FREE(dl->name);
	G_FREE(dl->name_utf8);
	G_FREE(dl->name_collate_key);
	G_FREE(dl->name_nojunk_utf8);
	G_FREE(dl->name_nojunk_collate_key);
	G_FREE(dl->status_completing_string);
	G_FREE(dl->gaps_data);
	G_FREE(dl->ext);
	G_FREE(dl->previewname);

	if (dl->progressbar)
		gdk_pixbuf_unref(dl->progressbar);

	memset(dl, 0x00, sizeof(Download));
	g_free(dl);
}


/******************************************************************************
 *
 *  download_remove
 *
 *  Removes the download from the ListStore and removes all
 *   other references to it in the various structures (hashtables)
 *
 ***/

void
download_remove (Download *dl)
{
	g_return_if_fail ( dl != NULL );

#if ( GTK_MAJOR_VERSION == 2 && GTK_MINOR_VERSION == 2 )
	if (gtk_list_store_iter_is_valid(downloads_store, &dl->iter))
		(void)gtk_list_store_remove(downloads_store, &dl->iter);
#else
	(void)gtk_list_store_remove(downloads_store, &dl->iter);
#endif

	if (downloads_ht)
	{
		if (!g_hash_table_remove (downloads_ht, dl->hash))
			g_warning("%s: g_hash_table_remove() failed.", __FUNCTION__);
	}

	download_free_download(dl);

	downloads_last_dl_removed = time(NULL);
}


/******************************************************************************
 *
 *  downloads_pause_all
 *
 *  pauses all downloads and prints 'only x MB left' message
 *
 ***/

void
downloads_pause_all (gfloat space_left_on_device)
{
	static time_t last_called = 0;

	g_return_if_fail (downloads_store != NULL);

	/* problem: a send_pause() will also request update on the space stats,
	 * which in turn will again trigger an emergency pause all if we don't
	 * break this cycle here */
	if ( last_called != 0 )
	{
		if ( (time(NULL)-last_called) < 120)
			return;
	}

	last_called = time(NULL);

	downloads_onPauseAll (NULL, GUINT_TO_POINTER(PAUSE_MODE_ALL));

	if ( downloads_num_paused > 0 )
		status_message_blue (_("GUI: Only %3.1f MB free in the temp folder! Pausing all downloads.\n"), space_left_on_device);
}



/******************************************************************************
 *
 *   downloads_pause_and_resume_all_active_downloads
 *
 ***/

void
downloads_pause_and_resume_all_active_downloads (void)
{
	GSList            *all, *node;
	static time_t    lastIPchange = 0;

	if (time(NULL)-lastIPchange< 60*60)
	{
		g_printerr ("%s", _(" You can't use |ipchanged| more often than once per hour.\n"));
		return;
	}

	lastIPchange = time(NULL);

	all = downloads_get_all();

	for ( node = all; node != NULL; node = node->next )
	{
		if (node->data)
			send_pause_download (((Download*)node->data)->hash);
	}

	send_disconnect();
	usleep(500);
	send_connect_to_server (0,0);

	for ( node = all; node != NULL; node = node->next )
	{
		if (node->data)
			send_resume_download (((Download*)node->data)->hash);
	}

	statusbar_extra_message ("%s", _(" IP change signalled - paused and resumed all active downloads and changed server."));
	g_slist_free(all);
}


/******************************************************************************
 *
 *   downloads_update_view_appearance
 *
 *   Makes filenames printed in bold or not, according to setting.
 *   Should be called after changing OPT_GUI_DOWNLOADS_FILENAMES_BOLD
 *    or OPT_GUI_DOWNLOADS_HIDE_JUNK or OPT_GUI_DOWNLOADS_SHOW_TRANSFERED_IN_PERCENT
 *
 ***/

void
downloads_update_view_appearance (void)
{
	GSList *all, *node;

	if (downloads_filename_renderer)
	{
		g_object_set (G_OBJECT(downloads_filename_renderer),
		              "weight", PANGO_WEIGHT_BOLD,
		              "weight-set", opt_get_bool(OPT_GUI_DOWNLOADS_FILENAMES_BOLD),
		              NULL );
	}

	/* setting all nojunk names to NULL will regenerate them if needed */
	all = downloads_get_all();
	for ( node = all;  node != NULL;  node = node->next )
	{
		if (node->data)
			G_FREE(((Download*)node->data)->name_nojunk_utf8);
	}
	g_slist_free(all);

	downloads_option_cache_hide_junk = opt_get_bool(OPT_GUI_DOWNLOADS_HIDE_JUNK);

	downloads_option_cache_trans_in_percent = opt_get_bool(OPT_GUI_DOWNLOADS_SHOW_TRANSFERED_IN_PERCENT);
}


/******************************************************************************
 *
 *   downloads_get_download_from_hash
 *
 ***/

Download *
downloads_get_download_from_hash (const guint8 *filehash)
{
	g_return_val_if_fail ( filehash        != NULL, NULL );
	g_return_val_if_fail ( downloads_ht    != NULL, NULL );

	return (Download*) g_hash_table_lookup ( downloads_ht, filehash );
}


/******************************************************************************
 *
 *   downloads_get_completed_items_count
 *
 *   returns the number of completed items in the download list
 *
 ***/

guint
downloads_get_completed_items_count (void)
{
	GSList *all, *node;
	guint   count = 0;

	all = downloads_get_all();

	for ( node = all;  node != NULL;  node = node->next )
	{
		Download *dl = (Download*)node->data;

		if ((dl) && ( dl->status == STATUS_COMPLETE  ||  dl->status == STATUS_CORRUPTED ) )
			count++;
	}

	g_slist_free(all);

	return count;
}



/******************************************************************************
 *
 *   downloads_get_selected
 *
 *   returns a GSList of Download records to all
 *    currently selected items
 *
 ***/

static void
downloads_get_selected_for_each_helper (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data)
{
	Download  *dl = NULL;

	gtk_tree_model_get (model, iter, 0, &dl, -1 );

	*((GSList**)data) = g_slist_prepend ( *((GSList**)data), dl );
}

GSList *
downloads_get_selected (void)
{
	GtkTreeSelection  *selection;
	GSList            *newlist = NULL;

	g_return_val_if_fail ( downloads_view != NULL, NULL );

	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(downloads_view));

	g_return_val_if_fail ( selection != NULL, NULL );

	gtk_tree_selection_selected_foreach ( selection, downloads_get_selected_for_each_helper, &newlist );

	return newlist;
}

/******************************************************************************
 *
 *   downloads_get_all
 *
 *   returns a list of all downloads.
 *    (used by downloadsattrib_list for syncing, and in downloads_popup.c)
 *
 *   (traversing the hash table is A LOT faster than walking the store)
 *
 ***/

static void
downloads_get_all_foreach_helper(gpointer key, gpointer val, GSList **p_list)
{
	*p_list = g_slist_prepend(*p_list, val);
}

GSList *
downloads_get_all (void)
{
	GSList *newlist = NULL;
	if (downloads_ht)
		g_hash_table_foreach(downloads_ht, (GHFunc)downloads_get_all_foreach_helper, &newlist);
	return newlist;
}


/******************************************************************************
 *
 *   downloads_add_download
 *
 *   make a new download record in store + list
 *
 ***/

void
downloads_add_download ( const guint8 *hash,
                         const gchar  *name,
                         guint32       size,
                         const gchar  *previewname,
                         guint8        priority,
                         guint32       id /* used from v49.5 to map status status <->  download */
                       )
{
	const gchar          *ext;
	gchar                *name_utf8, *name_nocase;
	gchar                *ext_utf8, *type_utf8 = NULL;
	Download             *dl;

	g_return_if_fail ( hash            != NULL );
	g_return_if_fail ( name            != NULL );
	g_return_if_fail ( downloads_store != NULL );
	g_return_if_fail ( id              >  0    ); /* as we don't want to store NULL pointers in hash table */

	if(downloads_get_download_from_hash(hash))
	{
		g_warning("%s: Asked to add download that's already in the list!\n", __FUNCTION__);
		return;
	}

	dl = g_new0(Download, 1);
	memcpy(dl->hash, hash, 16);

	dl->name         = g_strdup(name); /* do we need this field? */
	dl->name_utf8    = TO_UTF8(name);

	dl->size         = size;
	dl->priority     = priority;
	dl->id           = id; /* used from v49.5 to map status status <->  download */

	dl->previewname = g_strdup(previewname);

	dl->name_nojunk_utf8 = NULL;  /* filled on demand later */

	name_nocase = g_utf8_casefold(dl->name_utf8,-1);
	dl->name_collate_key = g_utf8_collate_key(name_nocase,-1);
	G_FREE(name_nocase);

	name_utf8 = TO_UTF8(name);

	ext = misc_get_extension_from_filename(name);
	ext_utf8 = TO_UTF8(ext);
	if (ext_utf8)
		type_utf8 = TO_UTF8(misc_get_filetype_from_extension(ext));

	dl->ext = g_strdup(ext); /* TODO: why not ext_utf8? */

	gtk_list_store_append (downloads_store, &dl->iter);


	g_hash_table_insert ( downloads_ht, g_memdup(hash,16), dl );
	g_hash_table_insert ( downloads_id_ht, GUINT_TO_POINTER(id), dl);

	dl->time_added = time(NULL);
	(void) downloadsattriblist_retrieve_attributes ( hash, &dl->last_complete, &dl->last_downloading, &dl->time_added );

	dl->status = STATUS_HASHING;

	gtk_list_store_set ( downloads_store, &dl->iter,
	                     DOWNLOADS_SIZE_STRING_COLUMN, misc_get_human_size_utf8 (NULL,0, size),
	                     DOWNLOADS_FILETYPE_COLUMN, type_utf8,
	                     DOWNLOADS_FILEEXTENSION_COLUMN, ext_utf8,
	                     -1 );

	dl->last_gotten_gaps = (time_t) 0;

	if ( opt_get_bool(OPT_GUI_DOWNLOADS_BLACKLIST_ON_DOWNLOAD) )
		add_to_blacklist (dl->hash, size, name);

	FREE_UTF8(ext_utf8);
	FREE_UTF8(type_utf8);
	FREE_UTF8(name_utf8);

	gtk_list_store_set ( downloads_store, &dl->iter, 0, dl, -1);
}


/******************************************************************************
 *
 *   downloads_remove_download
 *
 *   removes the download with the specified hash from the list.
 *
 *   Returns TRUE if the download could be found and removed, and the
 *    variables 'name', 'corrupted' and 'cancelled' have been set to
 *    the last values of the removed download.
 *
 *    ===> Caller needs to free 'name' string returned <===
 *
 *   Returns FALSE if no download with this hash could be found.
 *
 ***/

gboolean
downloads_remove_download ( const guint8 *hash,
                            gchar       **name,
                            gboolean     *corrupted,
                            gboolean     *cancelled,
                            guint        *size )
{
	Download *dl = NULL;

	g_return_val_if_fail ( hash      != NULL, FALSE );
	g_return_val_if_fail ( name      != NULL, FALSE );
	g_return_val_if_fail ( corrupted != NULL, FALSE );
	g_return_val_if_fail ( cancelled != NULL, FALSE );
	g_return_val_if_fail ( size      != NULL, FALSE );

	dl = downloads_get_download_from_hash(hash);
	g_return_val_if_fail ( dl != NULL, FALSE );

	*size      = dl->size;
	*name      = g_strdup(dl->name);
	*corrupted = dl->corrupted;
	*cancelled = dl->cancelled;


	dl->speed = 0.0;
	dl->done  = TRUE;
	G_FREE(dl->status_completing_string); /* will set it to NULL */

	(void) g_hash_table_remove (downloads_id_ht, GUINT_TO_POINTER(dl->id));

	/* we leave it in downloads_ht though. It will be removed
	 * from there in download_remove() called on 'clear completed downloads',
	 * if that has not already been done above (in case the DL was cancelled  */

	/* keep it in list as 'completed' download until cleared */
	if (dl->cancelled)
	{
		download_remove(dl);
		dl = NULL; /* from here on the dl record has been freed! */
	}
	else
	{
		dl->status = (dl->corrupted) ? STATUS_CORRUPTED : STATUS_COMPLETE;
		gtk_list_store_set(downloads_store, &dl->iter, 0, dl, -1); /* make row in the view update */
	}

	return TRUE;
}




/******************************************************************************
 *
 *   downloads_get_percent_completed
 *
 *   If the file system is local, and we are using the new x.part filesystem,
 *    then we can peek at the size of the x.1.part file and see how close it
 *    is to the filesize-to-be. From that we can calculate how much percent
 *    of the completing... process is already done.
 *
 *    NOTE: Of course this does not include the copying process to another
 *          filesystem if incoming/ and temp/ are on different filesystems :-/
 *          (XXX - this might be confusing for the user)
 *
 *    returns gfloat between 0.0 and 100.0 (0.0 on error or if not detectable)
 *
 ***/

G_INLINE_FUNC gfloat
downloads_get_percent_completed (const Download *dl)
{
	struct stat   logstat;

	if ( (!dl) || (!dl->previewname) || dl->size == 0)
		return 0.0;

	if (!option_temporary)
		return 0.0;

	if ( !g_file_test (dl->previewname, G_FILE_TEST_EXISTS) )
		return 0.0;

	if (stat (dl->previewname, &logstat) != 0)
		return 0.0;

	if ( logstat.st_size > dl->size )
		return 0.0;

	return ( ((gfloat)logstat.st_size / (gfloat)dl->size) * 100.0 );
}


/******************************************************************************
 *
 *   download_set_completing_string
 *
 *   ONLY CALL IF STATUS IS 'completing...' and CORE IS LOCAL!
 *
 ***/

G_INLINE_FUNC void
download_set_completing_string (Download *dl)
{
	gfloat percentcompleted = 0.0;

	g_return_if_fail ( dl != NULL );

	/* Status is 'Completing...' and core is local */

	percentcompleted = downloads_get_percent_completed(dl);

	/*  if we couldn't get a percentage value for some reason ... */
	if ( percentcompleted <= 0.0 )
	{
		G_FREE(dl->status_completing_string);	/* sets it to NULL as well */
		return;
	}

	if ( percentcompleted < 100.0 )
	{
		G_FREE(dl->status_completing_string);
		dl->status_completing_string = g_strconcat( status_utf8[STATUS_COMPLETING],
		                               UTF8_SHORT_PRINTF(_(" (%.0f%% done)"), percentcompleted),
		                               NULL);
		return;
	}

	dl->status_completing_string = g_strconcat( status_utf8[STATUS_COMPLETING],
	                                            UTF8_SHORT_PRINTF("%s", _(" (moving)")),
	                                            NULL);
}


/******************************************************************************
 *
 *   downloads_send_for_gaps_timeout
 *
 ***/

static gboolean
downloads_send_for_gaps_timeout (gpointer hash)
{
	if (hash)
	{
		send_get_download_gaps((const guint8*)hash);
		g_free(hash);
	}
	return FALSE;
}


/******************************************************************************
 *
 *   download_determine_dot
 *
 ***/

G_INLINE_FUNC IconName
download_determine_dot (Download *dl)
{
	gfloat fpercent;
	guint  percent;

	switch (dl->status)
	{
		case STATUS_PAUSED:
		case STATUS_PAUSING:
		case STATUS_RESUMING:
			return ICON_DL_STATUS_PAUSED;

		case STATUS_QUEUED:
			return ICON_DL_STATUS_PAUSED;

		case STATUS_HASHING:
		case STATUS_HASHING2:
			return ICON_NONE;

		case STATUS_COMPLETE:
			return ICON_DL_STATUS_AVAIL_FULL;

		default:
		break;
	}

	if ( dl->avail >= 100 )
		return ICON_DL_STATUS_AVAIL_FULL;

	if ( dl->sources == 0  ||  dl->avail == 0 )
		return ICON_DL_STATUS_ERROR;	/* colour of 'not available' = colour of 'error' */

	fpercent = (dl->transfered * 100.0) / ( dl->size * 1.0 );
	percent  = (gint) fpercent;

	if ( percent == 100  &&  fpercent < 100.0 )
		percent=99;

	if ( dl->avail <= percent )
		return ICON_DL_STATUS_AVAIL_LOW;

	return ICON_DL_STATUS_AVAIL_HIGH;
}


/******************************************************************************
 *
 *   progressbar_set_sources_from_gaplist
 *
 ***
   *
   * MORE THAN 255 SOURCES are a problem, because the GUI-core
   * protocol only provides 8-bit for the number of sources.
   * Currently the core effectively sends us num_sources&0xff, so
   * when there are 257 sources, we'd see a '2'. Which is not really
   * what we want.
   *
   * workaround:
   * if the core is gap-enabled and we show progress bars, then we
   * can scan through the progress bar gaplist and look at the gaps
   * which mark the parts that are available but don't have yet.
   * We will know the number of sources which have this one part
   * (luckily it's a 16-bit value there). Although we cannot deduct
   * from the gaplist how many different sources there are overall,
   * we know can know the minimum number of different sources from
   * the highest number of sources that has one and the same part of
   * the file.
   * Now, if the number of sources as provided by the core is LOWER
   * then the minimum number of sources according to the gaplist,
   * then we can conclude with some probability that we have to add
   * another 256 to get the real number of sources.
   * But of course our gaplist could just be horribly out of date or
   * wrong as well....
   *
   ***/

G_INLINE_FUNC void
download_set_sources_from_gaplist (Download *dl)
{
	/* require difference to be at least 20 to avoid wrong
	 * judgements due to less frequent update of the gaplist */
	while ( dl->min_sources > 0  &&  dl->sources < dl->min_sources  &&  (dl->min_sources - dl->sources) >= 20 )
	{
		/* XXX - do we need to specify a minimum threshold value for the
		 * 'gaplist' value (e.g. 64)? */
		dl->sources += 256;
	}
}

/******************************************************************************
 *
 *   downloads_set_download_status
 *
 *   sets the status values of a download (current speed etc.)
 *
 *   Note:
 *   - core versions up to incl. 48.2 use a 16-bit index number to
 *     map the status info to the DL (e.g. = number in list)
 *   - core versions from 49.5 onwards will send a unique 32-bit ID
 *     (a pointer essentially, so always non-NULL) to map the status
 *     info to a download. That unique ID will also be sent at the end
 *     of the NEW_DOWNLOAD message, so we know which ID goes with which
 *     download.
 *   - for core versions <= 48.2, ID = index number (also unique)
 *
 ***/

void
downloads_set_download_status (guint32 id,
                               guint8 status,
                               gfloat speed,
                               guint32 trans,
                               guint8 avail,
                               guint8 sources_num)
{
	guint8      oldstatus;
	Download   *dl = NULL;

	if (about_to_quit)
		return;

	g_return_if_fail ( downloads_id_ht != NULL );

	/* Get the download from the unique ID */
	dl = (Download*) g_hash_table_lookup(downloads_id_ht, GUINT_TO_POINTER(id)); /* new-style >= v49.5 */

	/* maybe a download has just been removed and we are still getting the status for it */
	if (!dl)
	{
		if ((time(NULL)-downloads_last_dl_removed) <= 10)
			return;
		g_return_if_fail (dl != NULL);
	}

	/* If the download is already finished, we are dealing with old and
	 *  delayed status messages. Ignore them. */
	if (dl->done)
		return;

	dl->sources = sources_num;
	download_set_sources_from_gaplist(dl);

	oldstatus      = dl->status;
	dl->status     = status;
	dl->speed      = speed;
	dl->transfered = trans;
	dl->left       = (dl->transfered > dl->size) ? G_MAXUINT : (dl->size - dl->transfered);
	dl->avail      = avail;

	/* check if it is time to get new gaps again */
	if (CORE_IS_GAP_ENABLED  &&  !dl->done )
	{
		gint   get_gap_interval = (opt_get_bool(OPT_GUI_SLOW_GUI_CORE_CONNECTION)) ? 30: 15;

		if ( (time(NULL) - dl->last_gotten_gaps) > get_gap_interval )
		{
			if ( oldstatus != STATUS_PAUSED  || dl->last_gotten_gaps == 0 )
			{
				/* randomly delay request for new gaps, so the cpu load gets spread better and
				 * doesn't bulk up when all the pixmaps have to be redrawn at the same time.
				 * (and yes, I do know that this method of getting a random number is bad) */
				srand (dl->last_gotten_gaps);
				g_timeout_add ( 500 + (rand()%(get_gap_interval/3)), (GSourceFunc) downloads_send_for_gaps_timeout, g_memdup(dl->hash,16));

				/* set this back in order to avoid asking over and over again
				 * if we don't get a reply from the core to this message      */
				dl->last_gotten_gaps = time(NULL);
			}
		}
	}

	/* to make progress bars grey-ish; we might receive first stats after first gaps */
	if (CORE_IS_GAP_ENABLED  &&  oldstatus == STATUS_HASHING  &&  status == STATUS_PAUSED)
		send_get_download_gaps(dl->hash);

	/* don't change status from 'pausing...' into looking
	 * or 'resuming...' into paused ('cause of delay of core response) */

	if ( oldstatus == STATUS_PAUSING   &&  status == STATUS_LOOKING  )
		dl->status = STATUS_PAUSING;

	if ( oldstatus == STATUS_RESUMING   &&   status == STATUS_PAUSED  )
		dl->status = STATUS_RESUMING;

	if ( status == STATUS_CORRUPTED )
		dl->corrupted = TRUE;


	if ( ( avail == 100 ) && (status == STATUS_LOOKING || status == STATUS_DOWNLOADING || status == STATUS_COMPLETING) )
		dl->last_complete = time(NULL);


	if ( speed > 0.05 )
		dl->last_downloading = time(NULL);


	if ( speed < 0.0   ||  speed > 2048.0 )
		dl->speed = 0.0;

	/* Note: Insuff. Disk Space and dlspeed>0 is possible, but probably
	 * wrong most of the time (core bug) (let's not print an error msg though,
	 * as this is a confirmed core bug) */
	if ( status == STATUS_ERRLOADING || status == STATUS_ERRHASHING || status == STATUS_IDS )
		dl->speed = 0.0;

	/* if it is downloading or was in the last 6 seconds, make it 'downloading...' */
	if ( status == STATUS_LOOKING )
	{
		if (  dl->speed > 0.0   ||   (time(NULL) - dl->last_downloading) <= 6  )
			dl->status = STATUS_DOWNLOADING;
	}

	if ( (dl->gaps_data) &&  trans >= dl->size  )
	{
		progressbar_set_transfered_from_gaplist (dl->gaps_data, &dl->transfered);
		dl->left = (dl->transfered > dl->size) ? G_MAXUINT : (dl->size - dl->transfered);
	}

	if ( dl->status == STATUS_COMPLETING && CORE_IS_LOCAL )
		download_set_completing_string(dl);

	/* we need to do this in order to force the model to re-sort itself */
	gtk_list_store_set (downloads_store, &dl->iter, 0, dl, -1);
}


/******************************************************************************
 *
 *   downloads_get_gaps_again_when_column_realised_timeout
 *
 *   This timeout is called if the width of a column was too wide at the
 *    time we last got the gaps (usually only the first time, when the
 *    column is not realised properly yet or so)
 *
 ***/

static gboolean
downloads_get_gaps_again_when_column_realised_timeout (gpointer hash)
{
	g_return_val_if_fail ( hash != NULL, FALSE );

	send_get_download_gaps((guint8*)hash);

	g_free(hash);

	return FALSE;	/* only call once */
}

/******************************************************************************
 *
 *   download_make_progressbar
 *
 *   Makes a progressbar from the gaplist
 *
 ***/

static void
download_make_progressbar (GtkTreeViewColumn *column, GtkCellRenderer *renderer, Download *dl)
{
	guint width = 0, height = 0, xpad;

	g_return_if_fail ( dl              != NULL );
	g_return_if_fail ( dl->progressbar == NULL );

	progressbar_sort_gaps (dl->gaps_data); /* only sort now that we actually draw the progress bar */

	width = gtk_tree_view_column_get_width (column);

	/* make sure progress bar height varies with font size */
	height = ( dlview_font_height == 0 ) ? 18 : dlview_font_height + 4;

	g_object_get ( G_OBJECT(renderer), "xpad", &xpad, NULL);

	width  = width  - (2*xpad) - 2;

	if ( gtk_tree_view_column_get_visible(column) == FALSE )
		width = 0;

	if ( width < DOWNLOADS_MAX_PROGRESSBAR_WIDTH )
	{
		/* note: might return NULL if width is too small */
		dl->progressbar = progressbar_make_pixbuf_from_gaplist(dl->gaps_data, width, height, dl->size,
		                                   (dl->status == STATUS_PAUSED || dl->status == STATUS_PAUSING));
	}
	else
		g_timeout_add(5000, downloads_get_gaps_again_when_column_realised_timeout, g_memdup(dl->hash,16));
}

/******************************************************************************
 *
 *   downloads_set_download_gaplist
 *
 *   sets the gaplist, taking ownership of the allocated 'pbgaps' array.
 *
 *   returns FALSE if the download record with the given hash was not
 *    found (then pbgaps needs to be freed by caller), otherwise TRUE.
 *
 ***/

gboolean
downloads_set_download_gaplist (const guint8 *hash, progressbarGap *pbgaps)
{
	Download *dl = NULL;

	if (about_to_quit)
		return FALSE;

	dl = downloads_get_download_from_hash(hash);
	g_return_val_if_fail ( dl != NULL, FALSE );

	/* make sure progress bar is re-generated with new gaps data later */
	if (dl->progressbar)
	{
		gdk_pixbuf_unref(dl->progressbar); /* free old progress bar */
		dl->progressbar = NULL;
	}

	if(dl->gaps_data) /* free old gaps data if it exists */
		g_free(dl->gaps_data);

	dl->gaps_data = pbgaps;
	dl->min_sources = progressbar_find_min_num_of_sources_from_gaplist(pbgaps);
	dl->last_gotten_gaps = time(NULL);

	return TRUE;
}


/******************************************************************************
 *
 *   downloads_set_download_gaplist_from_id
 *
 *   Upcoming cores (after v49.12) will send the download ID instead of
 *    the full file hash with the progress bar message
 *
 ***/

gboolean
downloads_set_download_gaplist_from_id ( guint32 id, progressbarGap *pbgaps)
{
	Download *dl = NULL;

	if (about_to_quit)
		return FALSE;

	/* Get the download from the unique ID */
	dl = (Download*) g_hash_table_lookup(downloads_id_ht, GUINT_TO_POINTER(id));
	g_return_val_if_fail ( dl != NULL, FALSE );

	/* make sure progress bar is re-generated with new gaps data later */
	if (dl->progressbar)
	{
		gdk_pixbuf_unref(dl->progressbar); /* free old progress bar */
		dl->progressbar = NULL;
	}

	if(dl->gaps_data) /* free old gaps data if it exists */
		g_free(dl->gaps_data);

	dl->gaps_data = pbgaps;
	dl->min_sources = progressbar_find_min_num_of_sources_from_gaplist(pbgaps);
	dl->last_gotten_gaps = time(NULL);

	return TRUE;
}


/******************************************************************************
 *
 *   downloads_redraw_all_progressbars
 *
 *   makes sure all progress bars are redrawn
 *
 ***/

void
downloads_redraw_all_progressbars (void)
{
	Download    *dl;
	GSList      *all, *node;

	if ( about_to_quit == 1 )
		return;

	all = downloads_get_all();

	for ( node=all;  node != NULL; node = node->next )
	{
		dl = (Download*)node->data;
		if ((dl) && (dl->progressbar))
		{
			gdk_pixbuf_unref(dl->progressbar);
			dl->progressbar = NULL;
			gtk_list_store_set(downloads_store, &dl->iter, 0, dl, -1); /* force row to be re-drawn */
		}
	}

	g_slist_free(all);
}


/******************************************************************************
 *
 *   downloads_onTimeout
 *
 *   Checks regularly whether the progress bar column has been resized,
 *   and if yes, makes sure that the pixmaps are redrawn with the new width
 *
 ***/

static gboolean
downloads_onTimeout (gpointer data)
{
	static guint lastwidth = G_MAXUINT;
	guint  width;

	if (!downloads_progressbar_column)
		return TRUE;

	width = gtk_tree_view_column_get_width (downloads_progressbar_column);

	if ( lastwidth != G_MAXUINT )
	{
		if ( width != lastwidth )
			downloads_redraw_all_progressbars();
	}

	lastwidth = width;

	return TRUE;
}


/******************************************************************************
 *
 *   downloads_onRowActivated
 *
 *   a row has been double-clicked.
 *
 ***/

static gboolean
downloads_onRowActivated (GtkWidget *view, GtkTreePath *path, GtkTreeViewColumn *col, gpointer data)
{
	GtkTreeIter   iter;
	Download     *dl = NULL;

	if (opt_get_bool(OPT_GUI_IGNORE_DOUBLE_CLICKS))
		return TRUE;

	if (!gtk_tree_model_get_iter (GTK_TREE_MODEL(downloads_store), &iter, path))
		return TRUE;

	gtk_tree_model_get ( GTK_TREE_MODEL(downloads_store), &iter, 0, &dl, -1);

	g_return_val_if_fail ( dl != NULL, TRUE );

	if (dl->done)
		return TRUE;

	switch (dl->status)
	{
		case STATUS_DOWNLOADING:
		case STATUS_LOOKING:
		case STATUS_QUEUED:
		case STATUS_IDS:
		case STATUS_PAUSED:
		{
			if (dl->status == STATUS_PAUSED)
			{
				send_resume_download (dl->hash);
				dl->status = STATUS_RESUMING;
			}
			else
			{
				send_pause_download (dl->hash);
				dl->status = STATUS_PAUSING;
 			}

			if (dl->progressbar)  /* redraw pixmap later with new 'active' or 'paused' colors */
			{
				gdk_pixbuf_unref(dl->progressbar);
				dl->progressbar = NULL;
			}

			gtk_list_store_set(downloads_store, &dl->iter, 0, dl, -1); /* force row to be re-drawn */

			send_get_space_stats();

			break;
		}

		default:
			break;
	}

	return TRUE;
}


/******************************************************************************
 *
 *   downloads_onButtonPress
 *
 *   a single click on a row has happened
 *
 ***/

static gboolean
downloads_onButtonPress (GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	gtk_widget_grab_focus(widget);

	/* right-click -> popup the menu */
	if (event->button == 3)
	{
		GtkTreeSelection *selection;

		selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(downloads_view));

		if ( misc_gtk_tree_selection_count_selected_rows (selection) <= 1 )
		{
			GtkTreePath *path = NULL;

			gtk_tree_selection_unselect_all(selection);

			if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW(downloads_view), (gint)event->x, (gint)event->y, &path, NULL, NULL, NULL))
				gtk_tree_selection_select_path (selection, path);

			if (path)
				gtk_tree_path_free(path);
		}

		downloads_popup_menu (widget, event, NULL);

		return TRUE; /* event has been dealt with */
	}

	return FALSE;
}



/******************************************************************************
 *
 *   downloads_onPopUpMenu
 *
 *   this is so that user-defined stuff like Shift-F10 works.
 *
 ***/

static gboolean
downloads_onPopUpMenu (GtkWidget *widget, gpointer data)
{
	downloads_popup_menu (widget, NULL, NULL);
	return TRUE;
}


/******************************************************************************
 *
 *  downloads_view_filename_func
 *
 *  Renders our filename column
 *
 ***/

static void
downloads_view_filename_func (GtkTreeViewColumn *column,
                              GtkCellRenderer   *renderer,
                              GtkTreeModel      *model,
                              GtkTreeIter       *iter,
                              gpointer           data)
{
	Download *dl = NULL;

	gtk_tree_model_get ( GTK_TREE_MODEL(model), iter, 0, &dl, -1);

	if (!downloads_option_cache_hide_junk)
	{
		g_object_set (renderer, "text", dl->name_utf8, NULL);
		return;
	}

	if (!dl->name_nojunk_utf8)
	{
		gchar *name_no_junk;
		gchar *nojunk_nocase;

		name_no_junk = g_strdup(dl->name);
		remove_junk_from_filename (name_no_junk);
		dl->name_nojunk_utf8 = TO_UTF8(name_no_junk);

		nojunk_nocase = g_utf8_casefold(dl->name_nojunk_utf8,-1);
		dl->name_nojunk_collate_key = g_utf8_collate_key(nojunk_nocase,-1);

		G_FREE(name_no_junk);
		G_FREE(nojunk_nocase);
	}

	g_object_set (renderer, "text", dl->name_nojunk_utf8, NULL);
}


/******************************************************************************
 *
 *  downloads_view_priority_func
 *
 *  Renders our priority column
 *
 ***/

static void
downloads_view_priority_func (GtkTreeViewColumn *column,
                              GtkCellRenderer   *renderer,
                              GtkTreeModel      *model,
                              GtkTreeIter       *iter,
                              gpointer           data)
{
	Download *dl = NULL;

	gtk_tree_model_get ( GTK_TREE_MODEL(model), iter, 0, &dl, -1);

	/* display no priority if DL is done */
	if ( (!dl) || (dl->done) )
		g_object_set (renderer, "text", NULL, NULL);
	else
		g_object_set (renderer, "text", misc_get_prio_string(dl->priority), NULL);
}


/******************************************************************************
 *
 *  downloads_view_icon_func
 *
 ***/

static void
downloads_view_icon_func (GtkTreeViewColumn *column,
                          GtkCellRenderer   *renderer,
                          GtkTreeModel      *model,
                          GtkTreeIter       *iter,
                          gpointer           data)
{
	Download *dl = NULL;

	gtk_tree_model_get ( GTK_TREE_MODEL(model), iter, 0, &dl, -1);

	if (dl)
		g_object_set (renderer, "pixbuf", get_icon(download_determine_dot(dl)), NULL);
	else
		g_object_set(renderer, "pixbuf", NULL, NULL);
}


/******************************************************************************
 *
 *  downloads_view_status_func
 *
 *  Renders our status column
 *
 ***/

static void
downloads_view_status_func (GtkTreeViewColumn *column,
                              GtkCellRenderer   *renderer,
                              GtkTreeModel      *model,
                              GtkTreeIter       *iter,
                              gpointer           data)
{
	Download *dl = NULL;

	gtk_tree_model_get ( GTK_TREE_MODEL(model), iter, 0, &dl, -1);

	if (dl)
	{
		if ((status_utf8) && dl->status < STATUS_LAST && (!dl->status_completing_string))
			g_object_set (renderer, "text", status_utf8[dl->status], NULL);
		else if (dl->status_completing_string)
			g_object_set (renderer, "text", dl->status_completing_string, NULL);
		else
			g_object_set (renderer, "text", FIXME_STRING, NULL);
	}
	else
		g_object_set(renderer, "text", FIXME_STRING, NULL);
}


/******************************************************************************
 *
 *  downloads_gobject_set_renderer_text_float
 *
 *  Transform the given value into a float string with one fraction
 *   digit (e.g. 23456 => 2345.6 ) and an optional unit at the end,
 *   and puts that string into the given cell renderer's "text" property.
 *
 *  Note: yes, it's horrible. So what? Shoot me. It's fast.
 *        A LOT faster than snprintf(). Multiply by number of
 *        visible downloads...
 *
 ***/

G_INLINE_FUNC void
downloads_gobject_set_renderer_text_float (GtkCellRenderer *renderer, guint val, gchar unit, gboolean twodigits)
{
	static gchar  *pos, num[7]; /* all 0x00 */

	num[0] = ( (val/10000)       | 0x30);  /* X000.0  or X00.00 */
	num[1] = (((val%10000)/1000) | 0x30);  /* 0X00.0  or 0X0.00 */
	num[2] = (((val%1000)/100)   | 0x30);  /* 00X0.0  or 00X.00 */

	if (!twodigits)
	{
		num[3] = (((val%100)/10)     | 0x30);  /* 000X.0 */
		num[4] = '.';
	}
	else
	{
		num[3] = '.';
		num[4] = (((val%100)/10)     | 0x30);  /* 000.X0 */
	}
	num[5] = ( (val%10)          | 0x30);  /* 0000.X  or 000.0X*/
	num[6] = unit;
	/* num[7] = 0x00;  is 0x00 due to static allocation */
	pos = num;
	if (*pos == '0') /* num[0] */
		++pos;
	if (*pos == '0') /* num[1] */
		++pos;
	if (*pos == '0') /* num[2] */
		++pos;

	g_object_set (renderer, "text", pos, "foreground-set", FALSE, NULL);

}


/******************************************************************************
 *
 *  downloads_gobject_set_renderer_text_uint
 *
 *  Transform the given value into a unsigned int string and
 *   an optional unit at the end, and puts that string into
 *   the given cell renderer's "text" property.
 *
 *  Note: yes, it's horrible. So what? Shoot me. It's fast.
 *        A LOT faster than snprintf(). Multiply by number of
 *        visible downloads...
 *
 ***/

G_INLINE_FUNC void
downloads_gobject_set_renderer_text_uint (GtkCellRenderer *renderer, guint val, gchar unit)
{
	static gchar  *pos, num[6]; /* all 0x00 */

	/* This is ca. 12x faster than printf() Note: '0' = 0x30 */
	num[0] = ( (val/1000)      | 0x30);
	num[1] = (((val%1000)/100) | 0x30);
	num[2] = (((val%100)/10)   | 0x30);
	num[3] = ( (val%10)        | 0x30);
	num[4] = unit;
	/* num[5] = 0x00;  due to static allocation */
	pos = num;
	if (*pos == '0') /* num[0] */
		++pos;
	if (*pos == '0') /* num[1] */
		++pos;
	if (*pos == '0') /* num[2] */
		++pos;

	g_object_set (renderer, "text", pos, "foreground-set", FALSE, NULL);
}


/******************************************************************************
 *
 *  downloads_view_transfered_func
 *
 *  Renders the bytes left column. Ugly but fast.
 *
 ***/

static void
downloads_view_transfered_func (GtkTreeViewColumn *column,
                                GtkCellRenderer   *renderer,
                                GtkTreeModel      *model,
                                GtkTreeIter       *iter,
                                gpointer           data)
{
	Download  *dl = NULL;

	gtk_tree_model_get ( GTK_TREE_MODEL(model), iter, 0, &dl, -1);

	if (!dl)
	{
		g_object_set(renderer, "text", FIXME_STRING, NULL);
		return;
	}

	if (dl->done)
	{
		g_object_set(renderer, "text", NULL, NULL);
		return;
	}

	if (!downloads_option_cache_trans_in_percent)
	{
		if ( dl->transfered < 1024 )
			downloads_gobject_set_renderer_text_uint (renderer, dl->transfered, 'b');
		else if ( dl->transfered < (1024*1024) )
			downloads_gobject_set_renderer_text_uint (renderer, (dl->transfered >> 10) , 'k');
		else if ( dl->transfered <= (1024*1024*1024))
			downloads_gobject_set_renderer_text_float(renderer, (((dl->transfered >> 10)*100) >> 10), 'M', TRUE);
		else
			downloads_gobject_set_renderer_text_float (renderer, (((dl->transfered >> 20)*10) >> 10), 'G', FALSE);
	}
	else
		downloads_gobject_set_renderer_text_uint (renderer, (guint)(((gfloat)dl->transfered*100.0)/(gfloat)dl->size)	, '%');
}

/******************************************************************************
 *
 *  downloads_view_bytes_left_func
 *
 *  Renders the bytes left column. Ugly but fast.
 *
 ***/

static void
downloads_view_bytes_left_func (GtkTreeViewColumn *column,
                               GtkCellRenderer   *renderer,
                               GtkTreeModel      *model,
                               GtkTreeIter       *iter,
                               gpointer           data)
{
	Download  *dl = NULL;

	gtk_tree_model_get ( GTK_TREE_MODEL(model), iter, 0, &dl, -1);

	if (!dl)
	{
		g_object_set(renderer, "text", FIXME_STRING, NULL);
		return;
	}

	if ( dl->done || dl->left == G_MAXUINT )
	{
		g_object_set(renderer, "text", NULL, NULL);
		return;
	}

	if ( dl->left < 1024 )
		downloads_gobject_set_renderer_text_uint (renderer, dl->left, 'b');
	else if ( dl->left < (1024*1024) )
		downloads_gobject_set_renderer_text_uint (renderer, (dl->left >> 10) , 'k');
	else if ( dl->left <= (1024*1024*1024))
		downloads_gobject_set_renderer_text_float(renderer, (((dl->left >> 10)*100) >> 10), 'M', TRUE);
	else
		downloads_gobject_set_renderer_text_float (renderer, (((dl->left >> 20)*10) >> 10), 'G', FALSE);
}


/******************************************************************************
 *
 *  downloads_view_speed_func
 *
 *  Renders the speed column
 *
 ***/

static void
downloads_view_speed_func (GtkTreeViewColumn *column,
                           GtkCellRenderer   *renderer,
                           GtkTreeModel      *model,
                           GtkTreeIter       *iter,
                           gpointer           data)
{
	Download  *dl = NULL;

	gtk_tree_model_get ( GTK_TREE_MODEL(model), iter, 0, &dl, -1);

	if (!dl)
	{
		g_object_set(renderer, "text", FIXME_STRING, NULL);
		return;
	}

	if ( dl->speed > 0.0  ||  dl->status == STATUS_DOWNLOADING )
		downloads_gobject_set_renderer_text_float (renderer, (guint)(dl->speed*10.0), 0x00, FALSE);
	else
		g_object_set(renderer, "text", NULL, NULL);
}


/******************************************************************************
 *
 *  downloads_view_time_left_func
 *
 *  Renders the time left column. Ugly but fast.
 *
 ***/

static void
downloads_view_time_left_func (GtkTreeViewColumn *column,
                               GtkCellRenderer   *renderer,
                               GtkTreeModel      *model,
                               GtkTreeIter       *iter,
                               gpointer           data)
{
	Download  *dl = NULL;

	gtk_tree_model_get ( GTK_TREE_MODEL(model), iter, 0, &dl, -1);

	if (!dl)
	{
		g_object_set(renderer, "text", FIXME_STRING, NULL);
		return;
	}

	if ( dl->speed < 0.05  &&  dl->status != STATUS_DOWNLOADING )
	{
		g_object_set(renderer, "text", NULL, NULL);
		return;
	}

	if ( dl->transfered > dl->size )
	{
		g_object_set(renderer, "text", UTF8_SHORT_PRINTF(_("not long")), NULL);
		return;
	}

	/* dl->secs_left = bytes_left / (speed_in_bytes_per_sec + 0.001) */
	dl->secs_left = (gint) (((dl->size - dl->transfered)*1.0) / ((dl->speed*1024.0) + 0.001));

	if ( dl->secs_left >= 3600 )
		downloads_gobject_set_renderer_text_uint (renderer, dl->secs_left/3600, 'h');
	else if ( dl->secs_left >= 60 )
		downloads_gobject_set_renderer_text_uint (renderer, dl->secs_left/60, 'm');
	else if (dl->secs_left > 0)
		downloads_gobject_set_renderer_text_uint (renderer, dl->secs_left, 's');
	else
		g_object_set(renderer, "text", NULL, NULL); /* usually if dl->size < dl->transfered */
}

/******************************************************************************
 *
 *  downloads_view_progress_bar_func
 *
 ***/

static void
downloads_view_progress_bar_func (GtkTreeViewColumn *column,
                                  GtkCellRenderer   *renderer,
                                  GtkTreeModel      *model,
                                  GtkTreeIter       *iter,
                                  gpointer           data)
{
	Download *dl = NULL;

	gtk_tree_model_get ( GTK_TREE_MODEL(model), iter, 0, &dl, -1);

	if ((!dl) || dl->done)
	{
		g_object_set(renderer, "pixbuf", NULL, NULL);
		return;
	}

	/* got gaps data, but no progress bar? => make progress bar */
	if ((!dl->progressbar) && (dl->gaps_data))
		download_make_progressbar(column, renderer, dl);

	if (dl->progressbar)
		g_object_set (renderer, "pixbuf", dl->progressbar, NULL);
	else
		g_object_set (renderer, "pixbuf", NULL, NULL);

}


/******************************************************************************
 *
 *  downloads_view_sources_func
 *
 *  Renders our sources column
 *
 ***/

static void
downloads_view_sources_func (GtkTreeViewColumn *column,
                             GtkCellRenderer   *renderer,
                             GtkTreeModel      *model,
                             GtkTreeIter       *iter,
                             gpointer           data)
{
	Download *dl = NULL;

	gtk_tree_model_get ( GTK_TREE_MODEL(model), iter, 0, &dl, -1);

	if (dl)
	{
		if (dl->status != STATUS_COMPLETE && dl->status != STATUS_CORRUPTED)
			downloads_gobject_set_renderer_text_uint(renderer, dl->sources, 0x00);
		else
			g_object_set(renderer, "text", NULL, NULL); /* don't show num. sources if completed */
	}
	else
		g_object_set(renderer, "text", FIXME_STRING, NULL);
}


/******************************************************************************
 *
 *  cell_data_function_download_last_seen
 *
 ***/

#define SECS_TWO_YEARS   (60*60*24*365*2)
#define SECS_ONE_YEAR    (60*60*24*365  )
#define SECS_ONE_DAY     (60*60*24      )
#define SECS_ONE_HOUR    (60*60         )

static void
cell_data_function_download_last_seen (GtkTreeViewColumn *column,
                                       GtkCellRenderer   *renderer,
                                       GtkTreeModel      *model,
                                       GtkTreeIter       *iter,
                                       gpointer           which)
{
	static gchar         *now_utf8, *never_utf8; /* all NULL */
	Download             *dl = NULL;
	time_t                lastseen, delta_t;

	gtk_tree_view_column_set_min_width (column, 0); /* TODO: why do we do this here? makes no sense */

	if (!now_utf8)
	{
		now_utf8   = TO_UTF8(_("now"));
		never_utf8 = TO_UTF8(_("never"));
	}

	gtk_tree_model_get ( GTK_TREE_MODEL(model), iter, 0, &dl, -1);

	g_return_if_fail ( dl != NULL );

	/* display nothing if DL is done */
	if (dl->done)
	{
		g_object_set (renderer, "text", NULL, "foreground-set", FALSE, NULL);
		return;
	}

	if ( GPOINTER_TO_UINT(which) == DL_SORT_ID_LAST_SEEN_DLING )
		lastseen = dl->last_downloading;
	else
		lastseen = dl->last_complete; /* DL_SORT_ID_LAST_SEEN_CMPLT */

	delta_t = time(NULL) - lastseen;

	if ( delta_t >= 0  &&  delta_t <= 6 )
	{
		g_object_set (renderer, "text", now_utf8, "foreground-set", FALSE, NULL);
		return;
	}

	/* more than 2 years ago? => never */
	if ( delta_t > SECS_TWO_YEARS )
	{
		if ( dl->status == STATUS_PAUSED || dl->status == STATUS_HASHING ) /* not red if paused or hashing */
			g_object_set (renderer, "text", never_utf8, "foreground-set", FALSE, NULL);
		else
			g_object_set (renderer, "text", never_utf8, "foreground", "red", "foreground-set", TRUE, NULL);

		return;
	}

	/* make this as fast as possible (who cares about correct rounding anyway?) */
	if ( delta_t >= SECS_ONE_YEAR )
		downloads_gobject_set_renderer_text_float (renderer, (delta_t*10)/SECS_ONE_YEAR, 'y', FALSE);
	else if ( delta_t >= SECS_ONE_DAY )
		downloads_gobject_set_renderer_text_float (renderer, (delta_t*10)/SECS_ONE_DAY, 'd', FALSE);
	else if ( delta_t >= SECS_ONE_HOUR )
		downloads_gobject_set_renderer_text_float (renderer, (delta_t*10)/SECS_ONE_HOUR, 'h', FALSE);
	else if ( delta_t >= 60 )
		downloads_gobject_set_renderer_text_float (renderer, (delta_t*10)/60, 'm', FALSE);
	else /* something between 0 and 60 seconds */
		downloads_gobject_set_renderer_text_uint(renderer, delta_t, 's');

#if 0
	/* THIS IS THE SLOW VARIANT : */
	g_object_set (renderer,  /* TODO: does seconds_into_human_time() not already return UTF8? */
	              "text", UTF8_SHORT_PRINTF("%s", seconds_into_human_time(delta_t)),
	              "foreground-set", FALSE,
	              NULL);
#endif
}


/******************************************************************************
 *
 *   downloads_update_rules_hint
 *
 ***/

void
downloads_update_rules_hint (void)
{
	gtk_tree_view_set_rules_hint (GTK_TREE_VIEW(downloads_view), opt_get_bool(OPT_GUI_SET_RULES_HINT_ON_LISTS));
}


/******************************************************************************
 *
 *   downloads_save_attributes_before_destroy
 *
 *   no use calling this after the main loop has been quit and everything
 *   has been destroyed...
 *
 ***/

static void
downloads_save_attributes_before_destroy (gpointer data)
{
	downloadsattriblist_save();
}

/******************************************************************************
 *
 *   downloads_tree_sortable_sort_function
 *
 ***/

G_INLINE_FUNC gint
dl_comp_int(guint a, guint b, const gchar *alta, const gchar *altb)
{
	if ( a == b )
	{
		if ((alta)&&(altb))
			return strcmp(alta,altb);
		return 0;
	}

	return ( a > b ) ? 1 : -1;
}

G_INLINE_FUNC gint
dl_comp_float(gfloat a, gfloat b, const gchar *alta, const gchar *altb)
{
	if ( a == b )
	{
		if ((alta)&&(altb))
			return strcmp(alta,altb);
		return 0;
	}

	return ( a > b ) ? 1 : -1;
}

/* dl_get_status_rating: returns a 'ranking' for a status */
/* the 'better' the status, the higher the returned value   */

G_INLINE_FUNC gint
dl_get_status_rating (guint status)
{
	static gint *status_value_map; /* NULL */
	const  gint  status_ranking[] = { STATUS_HASHING, STATUS_HASHING2, STATUS_ERRLOADING, STATUS_IDS,
	                                  STATUS_PAUSED, STATUS_PAUSING, STATUS_RESUMING, STATUS_QUEUED,
	                                  STATUS_NOSRCS, STATUS_LOOKING, STATUS_TRANSFERRING, STATUS_DOWNLOADING,
	                                  STATUS_ERRHASHING, STATUS_CORRUPTED, STATUS_COMPLETING, STATUS_COMPLETE,
	                                  STATUS_DONE, -1 }; /* the different status from 'worst' to 'best' */

	if ((!status_value_map)&&(downloads_store))
	{
		gint i;
		status_value_map = g_new0(gint,STATUS_LAST+1);

		/* free allocated memory with store */
		g_object_set_data_full (G_OBJECT(downloads_store), "status_value_map",
		                        status_value_map, (GDestroyNotify)g_free);
		g_object_set_data_full (G_OBJECT(downloads_store), "status_value_map_ptr",
		                        &status_value_map, (GDestroyNotify)g_nullify_pointer);

		for (i=0;  status_ranking[i] >= 0; i++)
			status_value_map[status_ranking[i]] = i;
	}

	if (status < STATUS_LAST)
		return status_value_map[status];

	g_return_val_if_reached(0);
}

static gint
downloads_tree_sortable_sort_function (GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gpointer sortid)
{
	const gchar  *namea, *nameb;
	Download     *dla = NULL, *dlb = NULL;

	gtk_tree_model_get (model, a, 0, &dla, -1);
	gtk_tree_model_get (model, b, 0, &dlb, -1);

	if ((!dla)||(!dlb)) /* happens for some reason when filling the model */
		return 0;

	/* use name as secondary sort criterion */
	if (downloads_option_cache_hide_junk)
	{
		namea = dla->name_nojunk_collate_key;
		nameb = dlb->name_nojunk_collate_key;
	}
	else
	{
		namea = dla->name_collate_key;
		nameb = dlb->name_collate_key;
	}

	switch (GPOINTER_TO_UINT(sortid))
	{
		case DL_SORT_ID_FILENAME:
			if ((namea)&&(nameb))
				return strcmp(namea,nameb);
			else
				return 0;
		case DL_SORT_ID_SIZE:
			return dl_comp_int(dla->size,dlb->size, namea, nameb);
		case DL_SORT_ID_STATUS:
			return dl_comp_int(dl_get_status_rating(dla->status),dl_get_status_rating(dlb->status), namea, nameb);

		case DL_SORT_ID_TRANSFERED:
			if (!downloads_option_cache_trans_in_percent)
				return dl_comp_int(dla->transfered,dlb->transfered, namea, nameb);
			else
				return dl_comp_float( ((gfloat)dla->transfered/(gfloat)dla->size), ((gfloat)dlb->transfered/(gfloat)dlb->size), namea, nameb);

		case DL_SORT_ID_BYTESLEFT:
			return dl_comp_int(dla->left,dlb->left, namea, nameb);

		case DL_SORT_ID_SPEED:
		{
			gint ret = dl_comp_float(dla->speed,dlb->speed, NULL, NULL);
			if (ret == 0) /* we want 0.0 Looking... records sorted before 0.0 Paused records */
				ret = downloads_tree_sortable_sort_function (model, a, b, GUINT_TO_POINTER(DL_SORT_ID_STATUS));
			if (ret == 0 && (namea) && (nameb)) /* if even the status is the same, sort alphabetically by name */
				ret = strcmp(namea,nameb);
			return ret;
		}

		case DL_SORT_ID_TIMELEFT:
			return dl_comp_int(dla->secs_left,dlb->secs_left, namea, nameb);

		case DL_SORT_ID_PRIORITY:
			return dl_comp_int(dla->priority,dlb->priority, namea, nameb);

		case DL_SORT_ID_AVAILABILITY:
			return dl_comp_int(dla->avail,dlb->avail, namea, nameb);

		case DL_SORT_ID_SOURCES:
			return dl_comp_int(dla->sources,dlb->sources, namea, nameb);

		case DL_SORT_ID_LAST_SEEN_DLING:
			return dl_comp_int(dla->last_downloading,dlb->last_downloading, namea, nameb);

		case DL_SORT_ID_LAST_SEEN_CMPLT:
			return dl_comp_int(dla->last_complete,dlb->last_complete, namea, nameb);

		default:
			break;
	}

	/* invalid sort column? Set sort column to something sane. */
	gtk_tree_sortable_set_sort_column_id ( GTK_TREE_SORTABLE(model), DL_SORT_ID_FILENAME, GTK_SORT_ASCENDING );
	gtk_tree_sortable_sort_column_changed ( GTK_TREE_SORTABLE(model) );
	g_return_val_if_reached(0);
}


/******************************************************************************
 *
 *   downloads_view_append_column
 *
 *   Appends a column with a cell data function to the downloads view
 *
 ***/

static GtkTreeViewColumn *
downloads_view_append_column (const gchar          *title,
                              GtkTreeCellDataFunc   cell_data_func,
                              gpointer              cell_data_func_data,
                              gint                  sortid,
                              gfloat                align,
                              guint                 cell_xpad,
                              gfloat                cell_xalign,
                              guint                 minwidth)
{
	GtkCellRenderer    *renderer;
	GtkTreeViewColumn  *column;
	gint                colnum;

	renderer = gtk_cell_renderer_text_new();
	colnum = gtk_tree_view_insert_column_with_data_func ( GTK_TREE_VIEW(downloads_view), -1,
	                                                      UTF8_SHORT_PRINTF("%s",title),
	                                                      renderer, cell_data_func,
	                                                      cell_data_func_data, NULL );
	column = gtk_tree_view_get_column(GTK_TREE_VIEW(downloads_view), colnum-1);
	gtk_tree_view_column_set_resizable(column, TRUE);
	gtk_tree_view_column_set_sort_column_id(column, sortid);
	gtk_tree_view_column_set_alignment(column, align);
	g_object_set (G_OBJECT(renderer), "xpad", cell_xpad, "xalign", cell_xalign, NULL);
	gtk_tree_view_column_set_min_width (column, minwidth);

	return column;
}

/******************************************************************************
 *
 *   downloads_create_store_model_and_view
 *
 *   Creates the downloads view, store and model, embedded in a scrollwin.
 *   Returns the scrollwin widget.
 *
 ***/

static GtkWidget *
downloads_create_store_model_and_view (void)
{
	GtkWidget          *scrollwin;
	GtkCellRenderer    *renderer;
	GtkTreeViewColumn  *column;
	GtkTreeSelection   *selection;
	gint                column_timeout_handle, sortid;

	downloads_store = gtk_list_store_new ( DOWNLOADS_N_COLUMNS,
	                                       DOWNLOADS_TYPES_PLACEHOLDER );

	downloads_store_sort = gtk_tree_model_sort_new_with_model (GTK_TREE_MODEL(downloads_store));

	downloads_view = gtk_tree_view_new_with_model ( GTK_TREE_MODEL(downloads_store));

	/* icon and filename */

	column = gtk_tree_view_column_new();
	gtk_tree_view_column_set_resizable(GTK_TREE_VIEW_COLUMN(column), TRUE);

	gtk_tree_view_column_set_title(column, UTF8_SHORT_PRINTF("%s",_("Filename")));
	gtk_tree_view_column_set_alignment(column, 0.02);

	renderer = gtk_cell_renderer_pixbuf_new();
	gtk_tree_view_column_pack_start(GTK_TREE_VIEW_COLUMN(column), renderer, FALSE);
	gtk_tree_view_column_set_cell_data_func(GTK_TREE_VIEW_COLUMN(column), renderer,
	                                        downloads_view_icon_func, NULL, NULL);

	renderer = gtk_cell_renderer_text_new();
	downloads_filename_renderer = renderer;
	gtk_tree_view_column_pack_start (GTK_TREE_VIEW_COLUMN(column), renderer, FALSE);
	gtk_tree_view_column_set_cell_data_func(GTK_TREE_VIEW_COLUMN(column), renderer,
	                                        downloads_view_filename_func, NULL, NULL);
	gtk_tree_view_column_set_sort_column_id(column, DL_SORT_ID_FILENAME);
	gtk_tree_view_append_column(GTK_TREE_VIEW(downloads_view), column);
	g_object_set (G_OBJECT(renderer), "xpad", 5, NULL);

	downloads_update_view_appearance();

	/* status */

	downloads_view_append_column (_("Status"), downloads_view_status_func, NULL,
	                              DL_SORT_ID_STATUS, 0.1, 5, 0.0, 5); /* align, cell_xpad, cell_xalign, minwidth */

	/* size */

	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(UTF8_SHORT_PRINTF("%s",_("Size")), renderer,
	                                                  "text", DOWNLOADS_SIZE_STRING_COLUMN, NULL);
	gtk_tree_view_column_set_resizable(GTK_TREE_VIEW_COLUMN(column), TRUE);
	gtk_tree_view_column_set_sort_column_id(column, DL_SORT_ID_SIZE);
	gtk_tree_view_append_column(GTK_TREE_VIEW(downloads_view), column);
	gtk_tree_view_column_set_alignment(column, 0.9);
	g_object_set (G_OBJECT(renderer), "xpad", 5, NULL);
	g_object_set (G_OBJECT(renderer), "xalign", 1.0, NULL);
	gtk_tree_view_column_set_min_width (column, 5);


	downloads_view_append_column (_("Transfered"), downloads_view_transfered_func, NULL,
	        DL_SORT_ID_TRANSFERED, 0.9, 5, 1.0, 5); /* align, cell_xpad, cell_xalign, minwidth */

	downloads_view_append_column (_("Left"), downloads_view_bytes_left_func, NULL,
	        DL_SORT_ID_BYTESLEFT, 0.9, 5, 1.0, 5); /* align, cell_xpad, cell_xalign, minwidth */

	downloads_view_append_column (_("Speed"), downloads_view_speed_func, NULL,
	        DL_SORT_ID_SPEED, 0.9, 5, 1.0, 5); /* align, cell_xpad, cell_xalign, minwidth */

	downloads_view_append_column (_("Time"), downloads_view_time_left_func, NULL,
	        DL_SORT_ID_TIMELEFT, 0.9, 5, 1.0, 5); /* align, cell_xpad, cell_xalign, minwidth */

	downloads_view_append_column (_("Priority"), downloads_view_priority_func, NULL,
	        DL_SORT_ID_PRIORITY, 0.2, 5, 0.0, 5); /* align, cell_xpad, cell_xalign, minwidth */

	column = gtk_tree_view_column_new();
	downloads_progressbar_column = column; /* save it so we can check whether it's been resized */
	gtk_tree_view_column_set_resizable(GTK_TREE_VIEW_COLUMN(column), TRUE);
	renderer = gtk_cell_renderer_pixbuf_new();
	gtk_tree_view_column_pack_start(GTK_TREE_VIEW_COLUMN(column), renderer, FALSE);
	gtk_tree_view_column_set_title(column, UTF8_SHORT_PRINTF("%s",_("Availability")));
	gtk_tree_view_column_set_sort_column_id(column, DL_SORT_ID_AVAILABILITY);
	gtk_tree_view_column_set_cell_data_func(column, renderer, downloads_view_progress_bar_func, NULL, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(downloads_view), column);
	gtk_tree_view_column_set_alignment(column, 0.02);
	g_object_set (G_OBJECT(renderer), "xalign", 0.0, "xpad", 0, NULL);
	gtk_tree_view_column_set_min_width (column, 5);

	/* foo ... , align, cell_xpad, cell_xalign, minwidth */
	downloads_view_append_column (_("Sources"), downloads_view_sources_func, NULL,
	                              DL_SORT_ID_SOURCES, 0.9, 5, 1.0, 5);

	downloads_view_append_column (_("Last Downloading"), cell_data_function_download_last_seen,
	                              GUINT_TO_POINTER(DL_SORT_ID_LAST_SEEN_DLING),
	                              DL_SORT_ID_LAST_SEEN_DLING, 0.9, 5, 1.0, 5);

	downloads_view_append_column (_("Last Complete"), cell_data_function_download_last_seen,
	                              GUINT_TO_POINTER(DL_SORT_ID_LAST_SEEN_CMPLT),
	                              DL_SORT_ID_LAST_SEEN_CMPLT, 0.9, 5, 1.0, 5);

	/* file type */

	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(UTF8_SHORT_PRINTF("%s",_("Type")),
	                                                  renderer, "text",
	                                                  DOWNLOADS_FILETYPE_COLUMN, NULL);
	gtk_tree_view_column_set_resizable(GTK_TREE_VIEW_COLUMN(column), TRUE);
	gtk_tree_view_column_set_sort_column_id(column, DOWNLOADS_FILETYPE_COLUMN);
	gtk_tree_view_append_column(GTK_TREE_VIEW(downloads_view), column);
	gtk_tree_view_column_set_alignment(column, 0.2);
	g_object_set (G_OBJECT(renderer), "xpad", 5, NULL);
	gtk_tree_view_column_set_min_width (column, 5);

	/* file extension */

	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(UTF8_SHORT_PRINTF("%s",_("Format")),
	                                                  renderer, "text", DOWNLOADS_FILEEXTENSION_COLUMN, NULL);
	gtk_tree_view_column_set_resizable(GTK_TREE_VIEW_COLUMN(column), TRUE);
	gtk_tree_view_column_set_sort_column_id(column, DOWNLOADS_FILEEXTENSION_COLUMN);
	gtk_tree_view_append_column(GTK_TREE_VIEW(downloads_view), column);
	gtk_tree_view_column_set_alignment(column, 0.2);
	g_object_set (G_OBJECT(renderer), "xpad", 5, NULL);
	gtk_tree_view_column_set_min_width (column, 5);

	scrollwin = gtk_scrolled_window_new (NULL,NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scrollwin), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER(scrollwin), downloads_view);

	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(downloads_view));

	/* set multi- or single- selection mode */
	if ( opt_get_bool(OPT_GUI_DOWNLOADS_SINGLE_SELECTION_MODE) )
		gtk_tree_selection_set_mode (selection, GTK_SELECTION_SINGLE);
	else
		gtk_tree_selection_set_mode (selection, GTK_SELECTION_MULTIPLE);


	for (sortid = 0; sortid < DL_SORT_ID_LAST; ++sortid)
	{
		gtk_tree_sortable_set_sort_func ( GTK_TREE_SORTABLE(downloads_store), sortid,
			             downloads_tree_sortable_sort_function, GUINT_TO_POINTER(sortid), NULL );
	}

	g_signal_connect (G_OBJECT(downloads_view), "row-activated", G_CALLBACK(downloads_onRowActivated), NULL);
	g_signal_connect (G_OBJECT(downloads_view), "button_press_event", G_CALLBACK(downloads_onButtonPress), NULL);
	g_signal_connect (G_OBJECT(downloads_view), "key_press_event", G_CALLBACK(downloads_onKeyPress), NULL);
	g_signal_connect (G_OBJECT(downloads_view), "popup_menu", G_CALLBACK(downloads_onPopUpMenu), NULL);

	/* hook up a download that checks every 2.0 secs whether
	 * the availability column has been resized, and if yes,
	 * redraws all the pixmaps */
	column_timeout_handle = g_timeout_add (2000, downloads_onTimeout, NULL);

	/* make sure those pointers are set to NULL when these objects are destroyed, otherwise
	 * the _onTimeout callback might be called and use invalid pointers shortly after
	 * the view has been destroyed. => segfault. We don't want that. */

	g_object_set_data_full (G_OBJECT(downloads_progressbar_column), "foo1",
	                        &downloads_progressbar_column, (GDestroyNotify)g_nullify_pointer);

	g_object_set_data_full (G_OBJECT(downloads_store), "foo1",
	                        &downloads_store, (GDestroyNotify)g_nullify_pointer);

	g_object_set_data_full (G_OBJECT(downloads_view), "foo1",
	                        &downloads_view, (GDestroyNotify)g_nullify_pointer);

	g_object_set_data_full (G_OBJECT(downloads_store_sort), "foo1",
	                        &downloads_store_sort, (GDestroyNotify)g_nullify_pointer);

	downloads_ht = g_hash_table_new_full ( misc_hash_hash, misc_hash_equal, g_free, NULL);

	g_object_set_data_full (G_OBJECT(downloads_view), "foo2",
	                        &downloads_ht, (GDestroyNotify)misc_hashtable_destroy);

	g_object_set_data_full(G_OBJECT(downloads_view), "baz3", NULL, (GDestroyNotify) downloads_save_attributes_before_destroy);


	downloads_update_rules_hint();

	gtk_widget_show_all (scrollwin);

	downloads_view_restore_settings();

	/* downloads_id_ht: used for status <-> dl mapping in cores >= v49.5 */
	downloads_id_ht = g_hash_table_new(g_direct_hash, g_direct_equal);

	/* create translated status strings, make sure they are freed when store object is destroyed */

	status_utf8 = g_new0 (gchar *, STATUS_LAST);

	g_object_set_data_full (G_OBJECT(downloads_store), "foo2",
	                        status_utf8, (GDestroyNotify)g_strfreev);

	status_utf8[STATUS_HASHING]      = TO_UTF8(_("(Hashing...)"));
	status_utf8[STATUS_QUEUED]       = TO_UTF8(_("Queued."));
	status_utf8[STATUS_LOOKING]      = TO_UTF8(_("Looking..."));
	status_utf8[STATUS_DOWNLOADING]  = TO_UTF8(_("Downloading..."));
	status_utf8[STATUS_PAUSED]       = TO_UTF8(_("Paused."));
	status_utf8[STATUS_IDS]          = TO_UTF8(_("Insufficient Disk Space."));
	status_utf8[STATUS_NOSRCS]       = TO_UTF8(_("No Sources."));
	status_utf8[STATUS_HASHING2]     = TO_UTF8(_("Hashing..."));
	status_utf8[STATUS_ERRLOADING]   = TO_UTF8(_("Error Loading."));
	status_utf8[STATUS_COMPLETING]   = TO_UTF8(_("Completing..."));
	status_utf8[STATUS_COMPLETE]     = TO_UTF8(_("Complete."));
	status_utf8[STATUS_CORRUPTED]    = TO_UTF8(_("Complete, but corrupted."));
	status_utf8[STATUS_TRANSFERRING] = TO_UTF8(_("Transferring..."));
	status_utf8[STATUS_RESUMING]     = TO_UTF8(_("Resuming..."));
	status_utf8[STATUS_PAUSING]      = TO_UTF8(_("Pausing..."));

	/* Get font size of downloads view, so we can draw the progress bar
	 *  with a height that corresponds to the font size */
	if ( dlview_font_height == 0 )
	{
		GtkStyle *style;

		style = gtk_widget_get_style(downloads_view);

		if ((style) && (style->font_desc))
			dlview_font_height = PANGO_PIXELS(pango_font_description_get_size(style->font_desc));
	}

	downloads_option_cache_hide_junk = opt_get_bool(OPT_GUI_DOWNLOADS_HIDE_JUNK);

	return (scrollwin);
}




/***************************************************************************
 *
 *   downloads_page_set_free_space
 *
 ***/

void
downloads_page_set_free_space (gfloat needed, gfloat temp, gfloat incoming)
{
	g_return_if_fail (downloads_page_stats_label!=NULL);

	if (needed == -1.0 && temp == -1.0 && incoming == -1.0)
	{
		gtk_label_set_text (GTK_LABEL(downloads_page_stats_label), UTF8_SHORT_PRINTF("%s",_(" Downloads ")));
	}
	else
	{
		gtk_label_set_text (GTK_LABEL(downloads_page_stats_label),
		       UTF8_SHORT_PRINTF(_(" Downloads - need %.1fMB - free in temp: %.1fMB - "
		                           "free in incoming: %.1fMB "), needed, temp, incoming));
	}
}


/***************************************************************************
 *
 *   downloads_page_set_file_descriptor_usage
 *
 ***************************************************************************/

void
downloads_page_set_file_descriptor_usage (glong core_open, glong user_max)
{
	g_return_if_fail ( downloads_page_fd_label != NULL );

	if (core_open < 0 || user_max < 0)
	{
		gtk_label_set_text(GTK_LABEL(downloads_page_fd_label), "");
		return;
	}

	if ( user_max > 65535 )
	{
		gtk_label_set_text (GTK_LABEL(downloads_page_fd_label), _(" No file descriptor limit "));
		return;
	}

	gtk_label_set_text (GTK_LABEL(downloads_page_fd_label),
	       UTF8_SHORT_PRINTF(_(" core fd usage: %lu/%lu "), core_open, user_max));
}



/***************************************************************************
 *
 *   downloads_page_create_stats_label
 *
 ***/

static GtkWidget *
downloads_page_create_stats_label (void)
{
	GtkWidget	*hbox, *emptylabel;

	hbox = gtk_hbox_new(FALSE,0);
	gtk_widget_show(hbox);

	emptylabel = gtk_label_new(" ");
	gtk_widget_show(emptylabel);

	downloads_page_stats_label = gtk_label_new (UTF8_SHORT_PRINTF("%s",_(" Downloads ")));
	gtk_label_set_justify (GTK_LABEL(downloads_page_stats_label), GTK_JUSTIFY_LEFT);
	gtk_widget_show(downloads_page_stats_label);

	downloads_page_fd_label = gtk_label_new(" ");
	gtk_label_set_justify (GTK_LABEL(downloads_page_stats_label), GTK_JUSTIFY_RIGHT);
	gtk_widget_show(downloads_page_fd_label);

	gtk_box_pack_start (GTK_BOX(hbox), downloads_page_stats_label, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX(hbox), emptylabel, TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX(hbox), downloads_page_fd_label, FALSE, FALSE, 0);

	if (!opt_get_bool(OPT_GUI_NO_GREEN))
		set_style_recursively (hbox, style_notebook_label);

	g_object_set_data_full (G_OBJECT(downloads_page_stats_label), "foo1",
	                        &downloads_page_stats_label, (GDestroyNotify)g_nullify_pointer);

	g_object_set_data_full (G_OBJECT(downloads_page_fd_label), "foo1",
	                        &downloads_page_fd_label, (GDestroyNotify)g_nullify_pointer);

	return hbox;
}



/***************************************************************************
 *
 *   downloads_page_create
 *
 ***/

GtkWidget *
downloads_page_create (void)
{
	GtkWidget	*vbox;
	GtkWidget	*frame;

	vbox = gtk_vbox_new (FALSE,0);

	g_return_val_if_fail (vbox!=NULL, NULL);

	gtk_box_pack_start (GTK_BOX (vbox), downloads_page_create_stats_label(), FALSE, FALSE, 0);

	frame = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
	gtk_container_add (GTK_CONTAINER (frame), downloads_create_store_model_and_view ());
	gtk_box_pack_start (GTK_BOX(vbox), frame, TRUE, TRUE, 0);

	gtk_widget_show_all (vbox);

	return vbox;
}

