/***************************************************************************
              downloads_attrib.c  -  downloads attributes stuff

	* sync attributes data in this list with downloads
	* save attributes data to .xml file
	* read attributes data from .xml file

                             -------------------
    begin                : Wed Jul 10 2002
    copyright            : (C) 2002 by Tim-Philipp Mller
    email                : t.i.m@orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "downloads.h"
#include "downloads_attrib.h"
#include "global.h"
#include "options.h"
#include "misc.h"
#include "status_page.h"

#include <stdio.h>
#include <unistd.h>
#include <gtk/gtk.h>

// separate list with records recently downloaded. Currently only
//	saving/syncing all the 'last seen XYZ' data
static GPtrArray		*downloadsAttributes = NULL;


// to remedy old sins we need to undefine these as
//	long as filelist_item records are used...
#undef lasttime_complete
#undef time_added


// --------- local functions

static downloadAttribRecord	*downloadsattriblist_get_from_hash (const guint8 *hash);
static downloadAttribRecord	*downloadsattriblist_add_record (const guint8 *hash);
static void			 downloadsattriblist_weed_out_unused (void);


// downloadsattriblist_add_record
//
// Adds a new record to the downloads attributes list with the given hash.
// Returns new record if it didn't exist, old record if it existed, or NULL on error.

static downloadAttribRecord *
downloadsattriblist_add_record (const guint8 *hash)
{
	downloadAttribRecord *rec;

	g_return_val_if_fail (hash!=NULL, NULL);

	if (!downloadsAttributes)
		downloadsAttributes = g_ptr_array_new();

	g_return_val_if_fail (downloadsAttributes!=NULL,NULL);

	rec = downloadsattriblist_get_from_hash (hash);

	if (rec)
		return rec;	/* exists already */

	rec = g_new0 (downloadAttribRecord, 1);

	g_return_val_if_fail ( rec != NULL, NULL);

	memcpy (rec->hash, hash, 16);
	rec->last_downloading = 0;
	rec->last_complete = 0;
	rec->last_used = time(NULL);
	rec->time_added = time(NULL);

	g_ptr_array_add (downloadsAttributes, (gpointer) rec);

	return rec;
}



/******************************************************************************
 *
 *   downloadsattriblist_get_from_hash
 *
 *   Find the attributes record corresponding to the given file hash
 *   Returns record or NULL if not found
 *
 ***/

static downloadAttribRecord *
downloadsattriblist_get_from_hash (const guint8 *hash)
{
	g_return_val_if_fail (hash!=NULL, NULL);

	if (!downloadsAttributes)
		return NULL;

	PROCESS_GPTRARRAY( downloadsAttributes, downloadAttribRecord,
		if (item)
		{
			if ( memcmp (item->hash, hash, 16) == 0 )
				return item;
		}
	);

	return NULL;
}


/******************************************************************************
 *
 *   update_dlstatslist_from_downloadlist
 *
 *   sync the attributes list with the downloads list (so that we always
 *   save the latest values when we save the attributes list)
 *
 ***/

static void
downloadsattriblist_update_from_downloadlist (void)
{
	downloadAttribRecord *rec;
	Download             *dl;
	GSList               *all, *node;

	all = downloads_get_all();

	for ( node = all;  node != NULL;  node = node->next )
	{
		dl = (Download*)node->data;
		if (dl)
		{
			rec = downloadsattriblist_get_from_hash (dl->hash);

			if (!rec)
				rec = downloadsattriblist_add_record (dl->hash);

			g_return_if_fail ( rec != NULL );

			rec->last_downloading = dl->last_downloading;
			rec->last_complete = dl->last_complete;
			rec->time_added = dl->time_added;

			rec->last_used = time(NULL);

			if ( rec->time_added == 0 )
				rec->time_added = time(NULL);
		}
	}

	g_slist_free(all);
}



/******************************************************************************
 *
 *   downloadsattriblist_retrieve_attributes
 *
 *   Retrieves the saved attributes for the download with hash X
 *   Returns TRUE if the download has been found and the attributes are set.
 *
 ***/

gboolean
downloadsattriblist_retrieve_attributes ( const gchar *hash,
                                          time_t *lasttime_complete,
                                          time_t *lasttime_downloading,
                                          time_t *time_added )
{
	downloadAttribRecord *rec;

	g_return_val_if_fail ( hash                 != NULL, FALSE);
	g_return_val_if_fail ( lasttime_complete    != NULL, FALSE);
	g_return_val_if_fail ( lasttime_downloading != NULL, FALSE);
	g_return_val_if_fail ( time_added           != NULL, FALSE);

	if (!downloadsAttributes)
		return FALSE;

	rec = downloadsattriblist_get_from_hash (hash);

	if (!rec)
		return FALSE;	/* not found */

	*lasttime_complete = rec->last_complete;

	*lasttime_downloading = rec->last_downloading;

	if (rec->time_added == 0)
		rec->time_added = time(NULL);

	*time_added = rec->time_added;

	rec->last_used = time(NULL);

  return TRUE;
}


/******************************************************************************
 *
 *   downloadsattriblist_weed_out_unused
 *
 *   weed out entries in the downloads attributes list
 *   that haven't been used for a long time.
 *
 ***/

static void
downloadsattriblist_weed_out_unused (void)
{
	if (!downloadsAttributes)
		return;

	PROCESS_GPTRARRAY_FROM_END ( downloadsAttributes, downloadAttribRecord,
		if ((item) && time(NULL)-item->last_used > 3600*24*31)
		{
			g_ptr_array_remove_fast (downloadsAttributes, (gpointer) item);
			g_free(item);
		}
	);
}


/******************************************************************************
 *
 *   downloadsattriblist_save
 *
 *   Save downloads attributes list to file
 *
 ***/

void
downloadsattriblist_save (void)
{
	FILE         *f;
	const gchar  *fn;

	fn = opt_get_opt_filename ("gui_downloadstats.txt");

	g_return_if_fail (fn != NULL);

	f = fopen (fn, "w");

	if ( f == NULL )
	{
		status_system_error_msg (_("Couldn't open gui_downloadsstats.txt for writing."));
	}

	fprintf (f, "# ed2k_gui --- gui_downloadsstats --- config format version 0.3.0\n");
	fprintf (f, "#\n%s\n", _("# This file is created by the ed2k gui. It contains information about your"));
	fprintf (f, "%s\n#\n", _("# downloads. DO NOT CHANGE THE ABOVE HEADER!"));
	fprintf (f, "%s\n#\n", _("# hash - last seen downloading - last seen fully available - last used - time added"));

	if (!downloadsAttributes)
	{
		fclose(f);
		return;
	}

	downloadsattriblist_update_from_downloadlist();
	downloadsattriblist_weed_out_unused();

	PROCESS_GPTRARRAY ( downloadsAttributes, downloadAttribRecord,
		if (item)
		{
			fprintf (f, "%32s %lu %lu %lu %lu\n",
				hash_to_hash_str(item->hash),
				item->last_downloading,
				item->last_complete,
				item->last_used,
				item->time_added);
		}
	);

	fclose(f);

}



/******************************************************************************
 *
 *   downloadsattriblist_load
 *
 *   Loads the downloads attributes list from a file.
 *
 ***/

void
downloadsattriblist_load (void)
{
	const gchar    *fn;
	gchar         **line, **linearray;
	guint           vermaj, vermin, verminmin;
	gint            ret;

	fn = opt_get_opt_filename("gui_downloadstats.txt");

	g_return_if_fail ( fn != NULL );

	linearray = misc_get_array_of_lines_from_textfile (fn, FALSE, TRUE);

	if (!linearray)
		return;

	line = linearray;
	ret = sscanf(*line, "# ed2k_gui --- gui_downloadsstats --- config format version %u.%u.%u\n", &vermaj, &vermin, &verminmin);
	if ( ret != 3)
	{
		g_printerr ("%s", _("gui_downloadsstats.txt: invalid header?\n"));
		g_strfreev (linearray);
		return;
	}
	else
	{
		gchar	hashstr[33];
		guint	last_downloading, last_fullyavailable, last_used, time_added;

		line++;
		while (*line)
		{
			if (**line!='#' && **line!=0x00)
			{
				if (5 == sscanf(*line, "%32s %u %u %u %u\n", hashstr, &last_downloading, &last_fullyavailable, &last_used, &time_added))
				{
					downloadAttribRecord *rec = downloadsattriblist_add_record (hash_str_to_hash(hashstr));
					if (rec)
					{
						rec->last_downloading = (time_t) last_downloading;
						rec->last_complete = (time_t) last_fullyavailable;
						rec->last_used = (time_t) last_used;
						rec->time_added = (time_t) time_added;
					}
				}
			}
			line++;
		}
	}
	g_strfreev (linearray);
}


