/***************************************************************************
                          guicoreconn.c  -  description
                             -------------------
    begin                : Wed Jun 26 2002
    copyright            : (C) 2002 by Tim-Philipp Mller
    email                : t.i.m@orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// TODO:
//
// * Only renew the connection timeout on packet complete if the packet is not a
//	status packet. If it's a status packet, renew to a long timeout, but only
//	if the last packet was not a status packet. This way we can detect core
//	screw-ups where the core keeps sending stats, but doesn't reply to our
//	commands any longer.
//

#include "connect_gui.h"
#include "dns_lookup.h"
#include "downloads.h"
#include "global.h"
#include "guicoreconn.h"
#include "gui_flags.h"
#include "mainwindow.h"
#include "misc.h"
#include "misc_gtk.h"
#include "misc_strings.h"
#include "notebook.h"
#include "options.h"
#include "packet.h"
#include "parse_message.h"
#include "search.h"
#include "send_command.h"
#include "servers.h"
#include "statusbar.h"
#include "status_page.h"
#include "toolbar.h"

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>

#define RECONNECT_TIMER_WAIT_SECONDS     15
#define RECONNECT_TIMER_PBAR_PULSES      30
#define RECONNECT_TIMER_INTERVAL_MS      ((RECONNECT_TIMER_WAIT_SECONDS*1000)/RECONNECT_TIMER_PBAR_PULSES)
#define RECONNECT_TIMER_PBAR_PULS_STEP   (1.0/(RECONNECT_TIMER_PBAR_PULSES))

/* global variables */

GConn		*guicoreconn = NULL;					// the GUI-core connection

gboolean	 guicore_conn_core_is_on_localhost = TRUE;		// for CORE_IS_LOCAL
guint		 guicoreconn_status = GUI_CORE_CONN_STATUS_NONE;	// status

gboolean	 option_always_try_to_reconnect_to_core = TRUE;

// This variable is set accordingly in a specific case of connection loss/timeout
//	(we don't want to overwrite the options variable for that)
gboolean	 guicoreconn_do_reconnect = FALSE;


static GQueue          *queue = NULL;


/* local functions */

static void         guicoreconn_callback_function (GConn *conn,
                                                   GConnEvent *event,
                                                   gpointer data);

static void         guicoreconn_connected (void);

static void         guicoreconn_close_and_cleanup (void);

static void         guicoreconn_parse_message_from_server (ramfile *packet);

static gint         guicoreconn_core_is_on_localhost (const GInetAddr *coreia);


/* local variables */

static gchar		*guicoreconn_user = NULL;
static gchar		*guicoreconn_pass = NULL;

// Have we ever received data on the current connection?
//	If we get a timeout on the connection and this is
//	false, we can show a message that the GUI is probably
//	connected to the wrong port (ie. the client port) on
//	the core.
static gboolean	 	 received_data = FALSE;

// We only need this glib mainloop in order to establish a
//	connection to the core to shut it down or send an
//	advanced command
static GMainLoop	*shutdown_mainloop = NULL;

// For the 'Reconnecting ...' dialog (progress bar stuff)
static	GtkWidget	*pbar = NULL;
static	guint		 pbar_timer_handle = 0;
static	guint		 guicoreconn_reconnect_counter = 0; // this is global because we need to reset it on 'connect now'

/* if we are trying to connect, then we don't need to replace the
 * 'reconnect' dialog with a 'connect to' dialog, because if the
 *  connection attempt fails it will show another dialog. On the
 *  other hand, if we deliberately close/destroy the 'reconnect' 
 *  dialog, then we want a 'connect to' dialog popping up instead.
 */
static gboolean		 guicoreconn_reconnect_dialog_new_dialog_needed = FALSE;

#define ED2K_BYTE	0xe3

#define GUICORECONN_READ_TIMEOUT	0
#define GUICORECONN_WRITE_TIMEOUT	0
#define GUICORECONN_CONNECT_TIMEOUT	10*1000		// does gnet lib use this now? (see docs)

// if we haven't received data from the core for 90s, a timeout function is called
#define GUICORECONN_TIMEOUT_INITIAL	15*1000
#define GUICORECONN_TIMEOUT		90*1000


static gboolean
guicoreconn_onTimeoutRequestRegularStats (gpointer data)
{
	send_start_sending_dl_status();
	send_start_sending_ul_status();

	return FALSE; /* only call once */
}

static void
guicoreconn_connected (void)
{
	send_login (guicoreconn_user, guicoreconn_pass);

	if (!window)
		return;

	//servers_init();

	first_time_we_get_options = 1; 	// so core version gets checked after every connect

	connection_closed_deliberately = FALSE;

	local_donkey_pid = (CORE_IS_LOCAL) ? donkey_process_running_check () : 0;	// update value, just in case

	// set our options to some default values (will be overridden once options are received)

	toolbar_set_server_data (NULL,0,0);
	option_max_dl_speed = 9999.0;
	option_max_ul_speed = 9999.0;

	G_FREE(option_incoming);
	option_incoming = g_strdup("");
	G_FREE(option_temporary);
	option_temporary = g_strdup("");
	G_FREE(option_nickname);
	option_nickname = g_strdup(DUMMY_NICKNAME);

	notebook_set_page(NOTEBOOK_PAGE_DOWNLOADS);

	toolbar_set_not_connected_to_core ();
	toolbar_set_server_data (NULL,0,0);

	/* set stuff back to donkey mode */
	toolbar_set_overnet_mode(FALSE);
	servers_set_overnet_mode(FALSE);
	search_set_overnet_mode (FALSE);
	notebook_set_overnet_mode (FALSE);
	mainwindow_set_title ("eDonkey2000");
	core_is_overnet = FALSE;
	core_is_hybrid  = FALSE;

	/* send this first thing, so we can find out whether it's eDonkey or overnet */
	send_command("g");

	send_get_options();	// do this as first thing, in case the lookup takes longer than 15secs and we get message

	// read in server names to look up, then look them up
	dns_lookup_read_lookuplist();
	dns_lookup_do_lookup();

	send_start_sending_dl_status();
	send_start_sending_ul_status();

	/* send the above 2 messages again after 3 seconds to give the core some time
	 * to do its house-keeping (otherwise it seems to not always start sending the
	 * status stuff for some reason). */
	g_timeout_add (1000, guicoreconn_onTimeoutRequestRegularStats, NULL);
	g_timeout_add (3000, guicoreconn_onTimeoutRequestRegularStats, NULL);
	g_timeout_add (6000, guicoreconn_onTimeoutRequestRegularStats, NULL);

	send_get_serverlist();
	send_get_shared_dirs();

	if ( !opt_get_bool(OPT_GUI_SLOW_GUI_CORE_CONNECTION) )
	{
		send_get_shared_files();
	}

	if ( !opt_get_bool(OPT_GUI_CHANGES_FROM_040_TO_050_SHOWN) && (window))
	{
		misc_gtk_ok_dialog (GTK_MESSAGE_INFO,
			_("Note: there are some interface changes\n"
			"since the last GUI version:\n"
			"\n"
			"* Click on the donkey logo to access the functions that\n"
			"  used to be on the about page (disconnect from core etc.)\n"
			"\n"
			"* Start a search and do much more via the toolbar\n"
			"  (type 'help' in the toolbar to get more information)\n"
			"\n"
			"* Add servers and send advanced commands via the toolbar\n"
			"\n"
			"* The pipe has moved from /tmp/.ed2k_gui_socket to ~/.ed2k_gui/pipe\n"
			"\n"));

		opt_set_bool (OPT_GUI_CHANGES_FROM_040_TO_050_SHOWN, TRUE);
	}

	/* let's check no one is stupid enough to run us as root, and nag if yes
	 * XXX - dialog may be overridden (transient + visibility) by connect to dialog!
	 */

	if ( geteuid() == (uid_t) 0 )
	{
		misc_gtk_ok_dialog (GTK_MESSAGE_WARNING,
			_("You are running this GUI as user 'root'.\n\n"
			  "You really should not do this!\n") );
	}

}


/* idle timeout callback */
static gboolean
guicoreconn_parse_complete_packet (ramfile *packet)
{
	g_return_val_if_fail ( packet != NULL, FALSE );
	g_return_val_if_fail ( ramfile_now_active == NULL, FALSE );

	if ( !CORE_IS_ALIVE )
	{
		ramfile_free(packet);

		while ((queue) && (!g_queue_is_empty(queue)))
			ramfile_free((ramfile*)g_queue_pop_head(queue));

		return FALSE;	/* discard packet from previous gui-core connection */
	}

	guicoreconn_parse_message_from_server (packet);

	ramfile_free (packet);

	if ((queue) && !g_queue_is_empty(queue) )
	{
		g_idle_add( (GSourceFunc) guicoreconn_parse_complete_packet, g_queue_pop_head(queue) );
	}

	return FALSE;	/* remove */
}


/******************************************************************************
 *
 *   guicoreconn_read_data
 *
 *   we received data from the core
 *
 *   returns TRUE if we want more data in the future
 *
 ***/

static gboolean
guicoreconn_read_data (ramfile **packet, const gchar *buffer, gint length)
{
	g_return_val_if_fail ( packet != NULL, FALSE );

	if (length == 0 )
		return FALSE;

	g_return_val_if_fail ( buffer != NULL, FALSE );

	if ( !packet_append (packet, buffer, length) )
	{
		guicoreconn_close_and_cleanup();
		return FALSE;
	}

	if ( *packet == NULL )
		return FALSE;

	if (!packet_header_is_valid (*packet, NULL))
	{
		g_printerr (   "-----\n");
		g_printerr ( _("Invalid packet on GUI-core connection - closing connection.\n") );
		g_printerr (   "-----\n");
		ramfile_dump( *packet, "invalid packet" );
		guicoreconn_close_and_cleanup();	// XXX = treat it as a crash?
		return FALSE;
	}

	while ((*packet) && packet_is_complete(*packet))
	{
		if ( ramfile_now_active != NULL )
		{
			if (!queue)
				queue = g_queue_new();

			g_queue_push_tail ( queue, *packet );
		}
		else
		{
			/* only start passing the packet after we've given back control to the main loop  */
			/* this way we can use gtk_main_iteration_do() while parsing packets and still    */
			/* continue receiving incoming data from the core (otherwise gnet read will fail) */
			g_idle_add ( (GSourceFunc) guicoreconn_parse_complete_packet, *packet );
		}

		*packet = packet_get_next_packet_data(*packet);

		/* only send the command once */
		if ( ( only_send_shutdown || only_send_advcmd ) && guicoreconn_status != GUI_CORE_CONN_STATUS_COMMUNICATING)
		{
			if (!only_be_quiet)
				g_printerr ( _(">>>>> Logged in okay - sending %s command\n"), (only_send_shutdown==1) ? _("shutdown") : _("advanced"));
			if (only_send_shutdown==1)
				send_logout();
			if (only_send_advcmd==1)
				send_command (shutdown_advcmd);

			return FALSE;
		}

		if ( guicoreconn_status == GUI_CORE_CONN_STATUS_COMMUNICATING )
			continue;

		if ( guicoreconn_status == GUI_CORE_CONN_STATUS_CLOSED_DELIBERATELY )
			continue; /* XXX - shouldn't that be return FALSE here? */

		if ( guicoreconn_status == GUI_CORE_CONN_STATUS_AUTHENTICATING )
			status_msg ( _("GUI: we are LOGGED IN to the core. Enjoy.\n"));

		guicoreconn_status = GUI_CORE_CONN_STATUS_COMMUNICATING;
	}

	return TRUE;
}


/******************************************************************************
 *
 *   guicoreconn_callback_function
 *
 *   The callback function for the GUI-core connection
 *
 ***/

static void
guicoreconn_callback_function (GConn *conn, GConnEvent *event, gpointer destroywidget)
{
	static ramfile    *packet = NULL;

	g_return_if_fail ( event != NULL );

	if ( guicoreconn_status == GUI_CORE_CONN_STATUS_CONNECTING && (destroywidget))
	{
		gtk_grab_remove (GTK_WIDGET(destroywidget));
		gtk_widget_destroy (GTK_WIDGET(destroywidget));
		pbar = NULL;
	}

	switch ( event->type )
	{
		/* We are connected to the core! */
		case GNET_CONN_CONNECT:
		{
			status_msg ( _("GUI: we are CONNECTED to the core. Logging in...\n") );
			guicoreconn_status = GUI_CORE_CONN_STATUS_AUTHENTICATING;

			if (packet)
				ramfile_free (packet);
			packet = NULL;
			received_data = FALSE;

			gnet_conn_readn(conn,5);	/* request header */
			gnet_conn_timeout (conn, GUICORECONN_TIMEOUT_INITIAL);

			guicoreconn_connected();

			if (only_send_shutdown==1 || only_send_advcmd==1)
			{
				send_get_options();
				gnet_conn_timeout (conn, 5*1000);
			}
		}
		return;

		case GNET_CONN_CLOSE:
		{
			guicoreconn_close_and_cleanup ();
		}
		return;

		case GNET_CONN_READABLE:
			g_return_if_reached();

		/* data has been read and is handed over to us here */
		case GNET_CONN_READ:
		{
			if ( guicoreconn_read_data (&packet, event->buffer, event->length) )
			{
				/* so we're called for more data that arrives as well */
				if ( !packet )
				{
					gnet_conn_readn(conn,5);	/* get header of next packet */
				}
				else
				{
					guint32 datalen = 0;

					if ( ramfile_get_size(packet) < 5 )
					{
						datalen = 5 - ramfile_get_size(packet);  /* get rest of header */
					}
					else
					{
						g_return_if_fail ( packet_header_is_valid(packet,&datalen) == TRUE );

#if 0
						if (datalen == 0  || datalen > (4*1024*1024) )
						{
							ramfile_dump (packet, "too big of a packet!?\n");
							g_return_if_fail ( datalen > 0  && datalen < (8*1024*1024) );
						}
#endif
					}

					gnet_conn_readn(conn, datalen); /* got packet, get payload */
				}
				gnet_conn_timeout (conn, GUICORECONN_TIMEOUT); /* Valid data received. Renew timeout. */
			}
			else packet = NULL;
		}
		return;

		case GNET_CONN_WRITE:			/* data has been successfully written */
		{
			if (only_send_shutdown==1 || only_send_advcmd==1)
			{
				if (!only_be_quiet)
					g_printerr ( _(">>>>> Sent command.\n") );

				if (only_send_advcmd==1)
					gnet_conn_timeout (conn, 5*1000);
			}
		}
		return;

		case GNET_CONN_TIMEOUT:
		{
			if (guicoreconn_status == GUI_CORE_CONN_STATUS_AUTHENTICATING)
				guicoreconn_status = GUI_CORE_CONN_STATUS_CLOSED_TIMEOUT_NO_DATA;
			else
				guicoreconn_status = GUI_CORE_CONN_STATUS_CLOSED_TIMEOUT;

			guicoreconn_close_and_cleanup();
		}
		return;

		case GNET_CONN_ERROR:
		{
			guicoreconn_close_and_cleanup();
		}
		return;

		case GNET_CONN_WRITABLE:
			g_return_if_reached();
	}


	g_return_if_reached();
}


// guicoreconn_close_connection
//
// Close the connection 'deliberately'

void
guicoreconn_close_connection (void)
{
	guicoreconn_status = GUI_CORE_CONN_STATUS_CLOSED_DELIBERATELY;
	if (guicoreconn)
	{
		gnet_conn_disconnect (guicoreconn);
		gnet_conn_delete (guicoreconn);
		guicoreconn = NULL;
	}
	guicoreconn_close_and_cleanup ();
}



// guicoreconn_close_and_cleanup
//
// Clean up stuff after the connection has been closed or close it 'not deliberately'
//	if not already closed.
// Then take the appropriate action (ie. 'connect to..'-dialog, try to reconnect, etc.)

static void
guicoreconn_close_and_cleanup (void)
{
	toolbar_set_not_connected_to_core ();

	// Not reconnecting, unless set otherwise
	guicoreconn_do_reconnect = FALSE;

	// Do not change status if we've closed the connection deliberately. Deliberately stays deliberately.
	if (guicoreconn_status != GUI_CORE_CONN_STATUS_CLOSED_DELIBERATELY)
	{
		switch (guicoreconn_status)
		{
			case GUI_CORE_CONN_STATUS_CONNECTING:
			{
				if (only_send_shutdown==1 || only_send_advcmd==1)
				{
					if (!only_be_quiet) g_printerr ( _(">>>>> Connect to core failed - could not connect!?\n"));
					exit (EXIT_FAILURE);
				}
				status_msg ( _("GUI: Could not establish GUI-core connection. (Core running? Core listening on 'aport' 4663? Core firewalled?)\n"));
				guicoreconn_do_reconnect = ( option_always_try_to_reconnect_to_core && !connection_closed_deliberately );
			}
			break;

			case GUI_CORE_CONN_STATUS_AUTHENTICATING:
			{
				if (only_send_shutdown==1 || only_send_advcmd==1)
				{
					if (!only_be_quiet) g_printerr ( _(">>>>> Connect to core failed - couldn't log in.\n"));
					exit (EXIT_FAILURE);
				}

				status_msg (_("GUI: Have you set up a user and password with the core? (see FAQ)\n"));
				status_msg (_("GUI: Established GUI-core connection, but authentification failed.\n"));
				if (guicoreconn)
				{
					gnet_conn_disconnect (guicoreconn);
					gnet_conn_delete (guicoreconn);
					guicoreconn = NULL;
				}
				/* not trying to reconnect in this case, what's the point? */
			}
			break;

			case GUI_CORE_CONN_STATUS_CLOSED_TIMEOUT:
			{
				if (only_send_shutdown==1 || only_send_advcmd==1)
				{
					// we've waited a couple of seconds
					exit (EXIT_SUCCESS);
				}
				// Houston, we've got a problem - haven't gotten reply from core for 90secs
				g_printerr ( _("Is core dead? Haven't received any data from core for %u seconds now...\n"), GUICORECONN_TIMEOUT/1000);
				if (CORE_IS_LOCAL)
				{
					g_printerr ( _("(currently I won't do anything. option to restart dead core will be implemented soon)\n"));
					g_printerr ( _("dead core PID = %d\n"), local_donkey_pid);
				}
				else
				{
					g_printerr ( _("core is on another computer?!\n"));
					// TODO: have command in options to execute when dead remote core is detected,
					// so one could set it up to start a remote script via ssh or something
				}
				guicoreconn_do_reconnect = option_always_try_to_reconnect_to_core;
			}
			break;

			case GUI_CORE_CONN_STATUS_CLOSED_TIMEOUT_NO_DATA:
			{
				const gchar	*msg[] = {
						"---\n",
						_("I haven't gotten a response from the core for 15 seconds now,\n"),
						_("although we are obviously connected to somewhere.\n"),
						_("Probably you are connecting to the core on the wrong port\n"),
						_("(ie the port for clients, 4662).\n\n"),
						_("You have to connect to the 'admin port', which is\n"),
						_("usually port 4663 (unless you have changed it with the\n"),
						_("'aport' command within the command line client).\n\n"),
						_("If you can't connect on that port, make sure you have started a core - \n"),
						_("either by clicking on the 'spawn local donkey' button (the dialog should\n"),
						_("display 'core already running' after that) or manually from the command\n"),
						_("line by using the '-' option (ie. './donkey - ! &amp;').\n\n"),
						_("In short: Try port 4663, otherwise please consult the FAQ\n"),
						"---\n", NULL };
				const gchar		**line = msg;

				if (only_send_shutdown==1 || only_send_advcmd==1)
				{
					if (!only_be_quiet) g_printerr (_(">>>>> Connect to core failed - no data received within 5 seconds!?\n"));
					exit (EXIT_FAILURE);
				}

				while (*line) { g_printerr ("%s", *line); line++; }
				status_msg (_("GUI-core connection to %s:%u timed out and we have never received any data.\n"), guicoreconn->hostname, guicoreconn->port);
				// not trying to reconnect here - what's the point?
			}
			break;

			default:
			{
				if (CORE_IS_LOCAL)
				{
					gint checkpid = donkey_process_running_check();	// double-check (debug)
					g_printerr (_("I think the core crashed or was killed.\n"));
					if (checkpid>0)
					{
						g_printerr (_("Hmm, but it seems like there is still a local donkey running?! (PID=%d)\n"), checkpid);
					}
					g_printerr (_("(core restart on crash option is disabled for now until it works properly.)\n"));
				}
				status_msg (_("GUI: GUI-core connection broke unexpectedly! (core crashed/killed?)\n"));
				guicoreconn_do_reconnect = option_always_try_to_reconnect_to_core;
			}
			break;
		}
		guicoreconn_status = GUI_CORE_CONN_STATUS_CLOSED_UNEXPECTEDLY;
	} else status_msg (_("GUI: GUI-core connection closed deliberately.\n"));


	if (only_send_shutdown==1 || only_send_advcmd==1)
	{
		if (!only_be_quiet) g_printerr (_(">>>>> Seems like we've shut down the core alright\n"));
		exit (EXIT_SUCCESS);
	}

	/* switch to status page */
	if (window)
		notebook_set_page(NOTEBOOK_PAGE_STATUS);

	/* gnet_conn_disconnect does nothing if connection is already closed */
	if (guicoreconn)
		gnet_conn_disconnect (guicoreconn);

	cleanup_after_disconnect();
}

static void
guicoreconn_parse_message_from_server (ramfile *packet)
{
	g_assert(ramfile_now_active == NULL);
	ramfile_now_active = packet;

	parse_message_from_client (packet);

	ramfile_now_active = NULL;
}


// guicoreconn_set_connection
//
// Sets up a connection structure to connect to the core
//
// destroywidget: a GtkWidget that will be destroyed if connect is successful (e.g. connect to dialog)
// Returns FALSE if we couldn't create the connection structure, otherwise TRUE

gboolean
guicoreconn_set_connection (const gchar *hostname, guint port, const gchar *username, const gchar *password, GtkWidget *destroywidget)
{
	GInetAddr	*ia;
	gint		 ret;

	g_return_val_if_fail (hostname!=NULL || *hostname==0x00, FALSE);
	g_return_val_if_fail (username!=NULL || *username==0x00, FALSE);
	g_return_val_if_fail (password!=NULL || *password==0x00, FALSE);

	if (guicoreconn)
		gnet_conn_delete (guicoreconn);

	ia = gnet_inetaddr_new (hostname, port);
	g_return_val_if_fail (ia!=NULL, FALSE);
	guicoreconn = gnet_conn_new_inetaddr (ia, guicoreconn_callback_function, (gpointer) destroywidget);
	g_return_val_if_fail (guicoreconn!=NULL, FALSE);
	if (guicoreconn_user) g_free (guicoreconn_user);
	if (guicoreconn_pass) g_free (guicoreconn_pass);
	guicoreconn_user = g_strdup (username);
	guicoreconn_pass = g_strdup (password);

	ret = guicoreconn_core_is_on_localhost (ia);
	if (ret<0) g_printerr ( _("Could not determine from hostname whether core runs on local machine or not.\n"));
	guicore_conn_core_is_on_localhost = (ret==1);
	if (ia) gnet_inetaddr_delete (ia);
	return TRUE;
}


/******************************************************************************
 *
 *  guicoreconn_connect
 *
 *  Initiate connection to core. Returns FALSE if GConn *
 *    structure not available, otherwise TRUE.
 *
 */

gboolean
guicoreconn_connect (void)
{
	if (!guicoreconn)
		return FALSE;

	guicoreconn_status = GUI_CORE_CONN_STATUS_CONNECTING;

	/* Switch to status page */
	if (window)
		notebook_set_page(NOTEBOOK_PAGE_STATUS);

	gnet_conn_timeout (guicoreconn, GUICORECONN_CONNECT_TIMEOUT);
	gnet_conn_connect (guicoreconn);

	/* the callback will take it from here... */
	return TRUE;
}


/******************************************************************************
 *
 *  guicoreconn_send_packet
 *
 *  TAKES OWNERSHIP OF PACKET PASSED AND FREES IT!
 *
 ***/

void
guicoreconn_send_packet (ramfile *packet)
{
	guint8   *np;
	guint32   data_size, intel_size;

	if (!guicoreconn)
		return;

	if (!gnet_conn_is_connected(guicoreconn))
		return;

	data_size = ramfile_get_size(packet);
	intel_size = GUINT32_TO_LE(data_size);		// to little endian

	np = g_new (guint8, data_size+5);

	np[0] = ED2K_BYTE;
	memcpy (np+1, &intel_size, 4);
	memcpy (np+1+4, ramfile_get_buffer(packet), data_size);

	gnet_conn_write (guicoreconn, np, data_size+5);

	/* data is copied by gnet_conn_write, so we can delete the buffer here right away */
	g_free(np);

	ramfile_free(packet);
}



/******************************************************************************
 *
 *   guicoreconn_core_is_on_localhost
 *
 *   Checks whether a given GInetAddr (which will be the inetaddr of the GUI-core
 *     connection here) is on the local host or on another host.
 *
 *   Note: we only check for localhost or not, and not for local interfaces any
 *         more, so people who want to do nifty things can do nifty things ;)
 *         (SSH-forwarding etc.)
 *
 * returns
 *   -1: error
 *    0: no, core inetaddr is not an interface on the local host
 *    1: yes, core inetaddr is an interface on the local host
 */

static gint
guicoreconn_core_is_on_localhost (const GInetAddr *coreia)
{
	gchar		*core_ipstr;

	g_return_val_if_fail (coreia!=NULL, -1);	// error
	
	core_ipstr = gnet_inetaddr_get_canonical_name(coreia);

	if (g_ascii_strncasecmp(core_ipstr, "127.0.0.", 8)==0)
	{
		g_free(core_ipstr);
		return 1;
	}

	g_free(core_ipstr);
	return 0;
}


// guicoreconn_do_shutdown_connection
//
// Establishes a connection with the purpose to either shut
//	down the core or send a single advanced command
//
// will EXIT either with EXIT_FAILURE or EXIT_SUCCESS!

void
guicoreconn_do_shutdown_connection (void)
{
	GMainContext	*context;	/* whatever that is for */

	if (only_send_advcmd==1)
	{
		g_print ( _("-=-=-=-=-   Sorry, this command line option currently doesn't work!   -=-=-=-=-\n"));
		exit (EXIT_FAILURE);
	}

	if (!guicoreconn_set_connection (shutdown_host, shutdown_port, shutdown_user, shutdown_pass, NULL))
	{
		if (!only_be_quiet)
			g_printerr ( _(">>>>> Connect to core failed - couldn't initiate valid connection structure with default values.\n"));

		exit (EXIT_FAILURE);
	}

	if (!guicoreconn_connect())
	{
		if (!only_be_quiet)
			g_printerr ( _(">>>>> Connect to core failed - connection structure hasn't been set up?!\n"));

		exit (EXIT_FAILURE);
	}

	context = g_main_context_new();
	shutdown_mainloop = g_main_loop_new (context, TRUE);

	if (!shutdown_mainloop)
		exit(EXIT_FAILURE);

	if (!only_be_quiet)
		g_print ( _(">>>>> Connecting to core...\n"));

	g_main_loop_run (shutdown_mainloop);

	exit (EXIT_SUCCESS);
}


/******************************************************************************
 *
 *   guicoreconn_reconnect_connect_now_button_clicked
 *
 *   The 'connect now' button has been pressed => connect now ;)
 *   NB: widget may be NULL (when called from timer)
 *
 *   If the connect attempt is succesful, we don't
 *     need another dialog ('reconnect' or 'connect to'),
 *     and the connection callback will destroy the
 *     existing one for us.
 *
 *   If the connect fails, the destroy handler will
 *     bring up a new dialog.
 */

static void
guicoreconn_reconnect_connect_now_button_clicked (GtkWidget *widget, gpointer pbdialogwidget)
{
	if (pbar_timer_handle>0)
		g_source_remove (pbar_timer_handle);
	pbar_timer_handle = 0;

	status_msg (_("GUI: Trying to reconnect to core ...\n"));

	if (!guicoreconn_connect())
	{
		status_msg (_("GUI: Reconnect not possible for some reason (bug!) ...\n"));
		// The dialog box is usually destroyed the first time
		//	the connection calback is called. If the connect()
		//	function fails, however, we have to do that ourself
		if ((pbdialogwidget) && GTK_IS_WIDGET(pbdialogwidget))
		{
			gtk_grab_remove (GTK_WIDGET(pbdialogwidget));
			gtk_widget_destroy (GTK_WIDGET(pbdialogwidget));
		}
		pbar = NULL;
	}
}



/******************************************************************************
 *
 *  guicoreconn_reconnect_stop_button_clicked
 *
 *  Stop reconnecting and show the 'connect to...' dialogue again.
 *  NB: widget may be NULL, data is the pbar_dialog widget
 *
 */

static void
guicoreconn_reconnect_stop_button_clicked (GtkWidget *widget, gpointer pbdialogwidget)
{
	if (pbar_timer_handle>0)
		g_source_remove (pbar_timer_handle);
	pbar_timer_handle = 0;

	if ((pbdialogwidget) && GTK_IS_WIDGET(pbdialogwidget))
	{
		guicoreconn_reconnect_dialog_new_dialog_needed = TRUE;
		gtk_grab_remove (GTK_WIDGET(pbdialogwidget));
		gtk_widget_destroy (GTK_WIDGET(pbdialogwidget));
	}
	pbar = NULL;
}


/******************************************************************************
 *
 *  guicoreconn_reconnect_timeout_callback
 *
 */

static gboolean
guicoreconn_reconnect_timeout_callback (gpointer data)
{
	guint	counter = ++guicoreconn_reconnect_counter;

	if (CORE_IS_ALIVE)
		return FALSE;

	gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR(pbar), (counter%(RECONNECT_TIMER_WAIT_SECONDS*2))*1.0/30.0);

	if (counter%(RECONNECT_TIMER_WAIT_SECONDS*2)==0)
	{
		guicoreconn_reconnect_connect_now_button_clicked (NULL, data);
		return FALSE;	// we don't want to be called again
	}
	return TRUE;	// we want to be called again
}


/******************************************************************************
 *
 *  guicoreconn_reconnect_dialog_delete_event
 *
 *  ('widget' is the pbdialog here)
 *
 */

static gint
guicoreconn_reconnect_dialog_delete_event (GtkWidget *widget, GdkEvent *event, gpointer data)
{
	guicoreconn_reconnect_dialog_new_dialog_needed = TRUE;
	
	if (pbar_timer_handle>0)
		g_source_remove (pbar_timer_handle);
	pbar_timer_handle = 0;

	if (widget)
		gtk_grab_remove (widget);

	pbar = NULL;

	return FALSE; /* dialog can be destroyed */
}


/******************************************************************************
 *
 *  guicoreconn_destroy_callback
 *
 */

static void
guicoreconn_destroy_callback (GtkWidget *widget, gpointer data)
{
	if (guicoreconn_reconnect_dialog_new_dialog_needed)
	{
		gui_connect_dialog();
	}
}


/******************************************************************************
 *
 *  guicoreconn_reconnect
 *
 *  Displays a dialog with a timeline and two buttons ('connect now', 'exit')
 *	 and reconnects after a certain amount of time (30secs)
 *
 */

void
guicoreconn_reconnect (void)	// XXX => make this static after putting cleanup stuff in this file!
{
	GtkWidget	*pbdialog, *cnow_button, *stop_button, *reconnect_label;

	if (!guicoreconn)
		gui_connect_dialog();

	if ((!window) || CORE_IS_ALIVE || (!guicoreconn))
		return;

	pbdialog = gtk_dialog_new ();
	gtk_window_set_title (GTK_WINDOW (pbdialog), _("Re-connecting to core ..."));
	gtk_window_set_position (GTK_WINDOW(pbdialog), GTK_WIN_POS_CENTER_ALWAYS);

	guicoreconn->user_data = pbdialog;

	new_button_pack_and_signal_connect (&cnow_button,
	                                    GTK_DIALOG(pbdialog)->action_area,
	                                    _(" Reconnect now "),
	                                    guicoreconn_reconnect_connect_now_button_clicked,
	                                    pbdialog, 1);

	new_button_pack_and_signal_connect (&stop_button,
	                                    GTK_DIALOG(pbdialog)->action_area,
	                                    _(" Stop "),
	                                    guicoreconn_reconnect_stop_button_clicked,
	                                    pbdialog, 1);

	// connect the destroy event to the dialogue, so people can't just click in the corner of
	// the connect dialog and close it and run the gui without being connected to the core
	// (NB: for some reason I didn't manage to pass the user_data to the callback)
	g_signal_connect (G_OBJECT (pbdialog), "delete_event",
	                  G_CALLBACK(guicoreconn_reconnect_dialog_delete_event), NULL);
	g_signal_connect (GTK_OBJECT (pbdialog), "destroy",
	                  G_CALLBACK(guicoreconn_destroy_callback), NULL);

	reconnect_label = gtk_label_new( UTF8_SHORT_PRINTF("%s",
	               _("\nThe connection to the core broke unexpectedly\nor auto-connect on start-up failed.\n\n"
	                 "  I will try to reconnect to the core in a couple of seconds...  ")));
	gtk_box_pack_start (GTK_BOX(GTK_DIALOG(pbdialog)->vbox),reconnect_label,FALSE,TRUE,0);
	gtk_widget_show (reconnect_label);

	pbar = gtk_progress_bar_new ();
	gtk_progress_bar_set_orientation (GTK_PROGRESS_BAR(pbar), GTK_PROGRESS_LEFT_TO_RIGHT);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG(pbdialog)->vbox), pbar, TRUE, TRUE, 20);
	gtk_widget_show (pbar);

	gtk_grab_add (pbdialog);
	gtk_window_set_transient_for (GTK_WINDOW(pbdialog), GTK_WINDOW(window));
	gtk_widget_show (pbdialog);

	guicoreconn_reconnect_counter = 0;
	guicoreconn_reconnect_dialog_new_dialog_needed = FALSE;
	pbar_timer_handle = g_timeout_add (500, guicoreconn_reconnect_timeout_callback, pbdialog);
}




