/*
 * main.c
 *
 * ed2k_gui - a nice gui for the donkey
 *
 * (c) 2001-2002 Tim-Philipp Muller <t.i.m@orange.net>
 *
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "blacklist.h"
#include "colors.h"
#include "connect_gui.h"
#include "downloads_attrib.h"
#include "dns_lookup.h"
#include "filter.h"
#include "global.h"
#include "gui_flags.h"
#include "guicoreconn.h"
#include "http_get.h"	/* for no_of_servers_fetched */
#include "icons.h"
#include "linkhandler.h"
#include "logfile.h"
#include "mainwindow.h"
#include "misc.h"
#include "misc_sys.h"
#include "notebook.h"
#include "options.h"
#include "packet.h"
#include "parse_message.h"
#include "pipe.h"
#include "search.h"
#include "send_command.h"
#include "servers.h"
#include "shared_page.h"
#include "signal_handler.h"
#include "spawn_core.h"
#include "servers.h"
#include "stats.h"
#include "status_page.h"
#include "statusbar.h"
#include "toolbar.h"
#include "uploads_page.h"
#include "uploadstats.h"

/* standard stuff */

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <limits.h>
#include <stdlib.h>
#include <ctype.h>
#include <signal.h>

/* global variables */

#if defined(__FreeBSD__) && defined(WITH_MALLOC_DEBUG)
  const gchar *_malloc_options = "AJVX";
#endif

/* local variables */

static time_t	 last_emergency_server_fetch = 0;
static gint		 general_update_timer_id = -1;

static gboolean	 use_options_data_for_shutdown = FALSE;
static gchar	*advcmd_to_execute = "";

static gchar	*main_ed2k_link_passed_on_commandline = NULL;

/* local functions */
static void		 main_parse_commandline_args (gint argc, gchar **argv);



/* functions */

// FIXME: remove -c option msg (wait for next release because of translations)

static void
print_helpscreen (const gchar *extramsg, gint exitcode)
{
	fprintf (stderr, "%s\n\n", extramsg);
	printf (_("linux gui for eDonkey2000 (www.eDonkey2000.com) - build %s %s\n"
		"(c) 2001 Tim-Philipp Muller <tim@edonkey2000.com>\n\n"
		"There are NO command line arguments you HAVE to specifiy.\n\n"
		"Optional command line options:\n"
		"-c\tconnect right away to the core without bringing up the 'connect GUI'\n"
		"\tdialog, using the default host,port,username and password specified on\n"
		"\tthe GUI's options page (and saved in gui_options on program exit).\n"
		"\tThis will only work if there is already a gui_options file.\n"
		"-i1,-i0\t\tThis will determine whether an application icon is set (-i1) or not (-i0)\n"
		"-p path\t\tspecifies the path which shall be used for the gui config files\n"
		"-n num\t\tspecifies 'instance' of GUI if you want to run multiple instances\n"
		"\t\t\tof the GUI at the same time and keep separate config and stats files\n"
		"\t\t\tfor them. e.g.: if num=2, then 'gui_options.2' is used\n"
		"\t\t\t(filter+blacklist files will be shared unless you specify different dirs)\n"
		"-q\t\tbe quiet when doing '-s' or '-a' commands (see below)\n"
		"-s host port username password\t\t just connect and shutdown core\n"
		"-a host port username password cmd\t just connect and send advanced command\n\n"),
		__DATE__, __TIME__);
	exit(exitcode);
}


/******************************************************************************
 *
 *   fill_in_special_commandline_args
 *
 *
 */

static void
fill_in_special_commandline_args (const gchar *a0, gint a1, const gchar *a2, const gchar *a3, const gchar *a4)
{
	shutdown_host   = g_strdup(a0);
	shutdown_port   = a1;
	shutdown_user   = g_strdup(a2);
	shutdown_pass   = g_strdup(a3);
	shutdown_advcmd = g_strdup(a4);
}


/******************************************************************************
 *
 *   main_parse_commandline_args
 *
 *
 */

static void
main_parse_commandline_args (gint argc, gchar **argv)
{
	gchar *unrecognised = _("UNRECOGNISED COMMAND-LINE OPTION");
	guint i;

	if (argc<=1) return;	/* any args at all? */

	for (i=1; i<argc; i++)
	{
		char *token=argv[i];
		if (!token) return;
		if (token[0]==0x00) return;
		if (token[0]=='-')
		{
			token++;
			if (token[0]=='-') token++;
			switch (token[0])
			{
				case 'h':	print_helpscreen("", 0);
						break;

				case 'c':
					g_printerr ("The '-c' option is no longer supported.\n");
				break;

				case 'q':	only_be_quiet	   = 1;
						break;

				case 's':	/* just shutdown core */
				{
					if (argc>=(1+i+4))
					{
						only_send_shutdown = 1;
						fill_in_special_commandline_args (argv[i+1], atoi(argv[i+2]), argv[i+3], argv[i+4], "");
						if (!only_be_quiet)
						{
							printf (_(">>>>> shutdown core @ %s:%u (%s:%s)\n"),
									shutdown_host, shutdown_port, shutdown_user, shutdown_pass);
						}
						i=i+4;
					} else {
						only_send_shutdown = 1;
						use_options_data_for_shutdown = TRUE;
						if (!only_be_quiet)
							printf (_(">>>>> doing core shutdown (with saved parameters)\n"));
					}
				}
				break;

				case 'a':	/* just send advanced command */
				{
					if (argc>=(1+i+5))
					{
						only_send_advcmd = 1;
						fill_in_special_commandline_args (argv[i+1], atoi(argv[i+2]), argv[i+3], argv[i+4], argv[i+5]);
						if (!only_be_quiet) printf (_(">>>>> send advanced command to core @ %s:%u (%s:%s) - '%s'\n"),
									shutdown_host, shutdown_port, shutdown_user, shutdown_pass, shutdown_advcmd);
						i=i+5;
					} else if (argc>=(1+i+1))
					{
						only_send_advcmd = 1;
						use_options_data_for_shutdown = TRUE;
						advcmd_to_execute = argv[i+1];
						i++;
					} else print_helpscreen(_("The '-a' option requires ONE or FIVE extra parameters"),1);
				}
				break;

				case 'i':	// don't set application icon (for Jaroslav)
				{
					g_printerr ("The '-i' option is no longer supported. Please edit the ~/.ed2k_gui/options file manually.\n");
				}
				break;

				case 'p':	/* specify path where config files are stored */
				{
					if (token[1]==0x00)	/* -p /home/joe/ed2k/ variant */
					{
						if (argc>=(1+i+1))
						{
							opt_set_prefpath(argv[i+1]);

							i++;
						}
						else print_helpscreen (_("The '-p' option requires an extra parameter"),1);
					}
					else opt_set_prefpath(token+1);
				}
				break;

				case 'n':	/* specify 'instance' of GUI */
				{
					if (token[1]==0x00)	// case '-n 2'
					{
						if (argc>=(1+i+1))
						{
							option_gui_instance = atoi(argv[i+1]);
							i++;
						} else print_helpscreen (_("The '-n' option requires an extra parameter"),1);
					} else option_gui_instance = atoi(token+1);
				}
				break;
				default:	fprintf (stderr, "---> '%s' <--- ", token);
						print_helpscreen(unrecognised,1);
			}
		}
		else
		{
			if (g_ascii_strncasecmp(token,"ed2k:",5)!=0)
			{
				fprintf (stderr, "---> '%s' <--- ", token);
				print_helpscreen(unrecognised,1);
			} else main_ed2k_link_passed_on_commandline = g_strdup(token);
		}
	}
}



/******************************************************************************
 *
 *   general_update_timer
 *
 *   do stuff that needs to be done every x minutes or so
 */

gint
general_update_timer (gpointer data)
{
	static int 	minutes = 0;

	/* every minute */

	/* update info in the statusbar (=switch to current page) */
	notebook_set_page(-1);

	if (CORE_IS_ALIVE)
		send_get_peer_stats ();


	downloads_page_set_file_descriptor_usage (-1,-1); /* clear label */

	if ( CORE_IS_ALIVE &&  CORE_IS_LOCAL  &&  local_donkey_pid > 0 )
	{
		glong  max_open, num_open;

		if (misc_sys_get_max_open_files(&max_open) && misc_sys_get_pid_open_files(local_donkey_pid,&num_open))
			downloads_page_set_file_descriptor_usage(num_open, max_open);
	}


	send_get_space_stats();

	if ( IS_OVERNET || IS_HYBRID )
	{
		send_get_overnet_contacts();
		send_command("g");	/* get current ID and firewalled status */
	}

	/* every 5 minutes */
	if (minutes%5 == 0)
	{
		/* we don't want to totally run out of servers, do we? */
		if ( opt_get_bool(OPT_GUI_SERVERS_FETCH_NEW_LIST_WHEN_TOO_FEW)
			&& ( IS_OVERNET == FALSE || IS_HYBRID == TRUE )
			&& ( servers_get_count () < 10)
			&& ((option_nickname) && strcmp(option_nickname, DUMMY_NICKNAME)!=0) )
		{
			if (CORE_IS_ALIVE)
			{
				static guint whichlist = SERVER_LIST_NUM;

				whichlist = ( (whichlist+1) % SERVER_LIST_NUM );

				/* if fetch unsuccessful, we don't want to re-fetch
				 * unsuccessfully every 5 minutes, do we?
				 *
				 *   max. is every 30 minutes
				 */
				if (last_emergency_server_fetch==0 || (time(NULL)-last_emergency_server_fetch)>1800)
				{
					last_emergency_server_fetch = time(NULL);
					status_message_blue (_("GUI: less than 10 servers in serverlist - trying to fetch more via http.\n"));

					servers_fetch_serverlist ( whichlist );
				}
			}
		}

		if ((CORE_IS_ALIVE) && ( IS_OVERNET || IS_HYBRID ))
		{
			if ( option_autoremove == 1 )	/* in case core is too lazy */
				servers_remove_dead();

			send_get_serverlist();
		}

		mainwindow_retrieve_some_options_before_main_window_is_destroyed();
		opt_write_to_disk();
		write_gui_uploadstats();
		stats_save_to_disk ();
		downloadsattriblist_save();
	}

	/* every 30 minutes */
	if (minutes%30 == 0 && CORE_IS_ALIVE)
	{
		send_get_shared_dirs();
		if ( !opt_get_bool(OPT_GUI_SLOW_GUI_CORE_CONNECTION))	// if slow do it only every 90 mins.
			send_get_shared_files();
	}

	if (minutes%90 == 0)
	{
		if (opt_get_bool(OPT_GUI_SLOW_GUI_CORE_CONNECTION))
			send_get_shared_files();
	}

	/* server rotate feature */
	if ( opt_get_int(OPT_GUI_SERVERS_ROTATE_INTERVAL) > 0
	        && ( IS_OVERNET == FALSE || IS_HYBRID == TRUE ))
	{
		if ( minutes > 0 && minutes >= opt_get_int(OPT_GUI_SERVERS_ROTATE_INTERVAL)
			 && (minutes % opt_get_int(OPT_GUI_SERVERS_ROTATE_INTERVAL) )==0 && CORE_IS_ALIVE)
		{
			/* will set current server to low priority */
			servers_register_connect_status (NULL, SERVERS_CONNECT_STATUS_DO_SERVER_ROTATE);
		}
	}

	/* clear status log if user has set the option accordingly */
	if ( opt_get_int(OPT_GUI_STATUS_WINDOW_CLEAR_INTERVAL) > 0
	     && minutes >= opt_get_int(OPT_GUI_STATUS_WINDOW_CLEAR_INTERVAL)
	     && minutes >  0
		 && (minutes % opt_get_int(OPT_GUI_STATUS_WINDOW_CLEAR_INTERVAL))==0 && CORE_IS_ALIVE)
	{
		status_page_clear();
		status_msg (_("GUI: cleared status window. You will find all previous output in the gui_statuslog file.\n"));
	}


	/* check if the blacklistfile has been changed by hand */
	check_if_blacklistfile_changed ();
	check_if_filterlistfile_changed ();

	minutes++;
	return(1);	/* call us in the future */
}


/******************************************************************************
 *
 *   main_clean_up_connections_and_callbacks_before_quit
 *
 *   Remove timeouts and send some stuff to core before we go away.
 *   Now that all this stuff is done via glib/gnet, we need to do it
 *    before we call gtk_main_quit()
 *
 */

void
main_clean_up_connections_and_callbacks_before_quit (void)
{
	time_t	now;

	g_source_remove(general_update_timer_id);

	/* will not be sent if core is not alive */
	send_stop_sending_dl_status();
	send_stop_sending_ul_status();

	if (opt_get_bool(OPT_GUI_SHUTDOWN_CORE_ON_EXIT))
	{
		dns_lookup_remove_dns_servers ();
		send_logout();
	}

	/* We do this so core tries to send sth after we closed the
	 * socket and recognises that we're gone (Another Control
	 * Attempt problem on second connect...)
	 *
	 */
	send_get_shared_files();

	/* make sure our stuff gets sent out (but wait max. 1 second) */
	now = time(NULL);
	while ((gtk_events_pending()) && (time(NULL)-now)<1)
		(void) gtk_main_iteration();

	guicoreconn_close_connection ();
}


/******************************************************************************
 *
 *   main_send_shutdown_or_advanced_command_to_core
 *
 *   Only sends a shutdown message or an advanced
 *   command message to the core, nothing else.
 *
 *   Does not return.
 *
 */

static void
main_send_shutdown_or_advanced_command_to_core (void)
{
	opt_read_from_disk();

	if (use_options_data_for_shutdown==1)
	{
		/* XXX - TODO: check if default options are set?! */
		fill_in_special_commandline_args( opt_get_str(OPT_GUI_DEFAULT_CORE_HOST),
		                                  opt_get_int(OPT_GUI_DEFAULT_CORE_PORT),
		                                  opt_get_str(OPT_GUI_DEFAULT_CORE_USER),
		                                  opt_get_str(OPT_GUI_DEFAULT_CORE_PASS),
		                                  advcmd_to_execute);

		if (!only_be_quiet)
		{
			if (only_send_shutdown)
			{
				printf (_(">>>>> shutdown core @ %s:%u (%s:%s)\n"),
					shutdown_host, shutdown_port,
					shutdown_user, shutdown_pass);
			}

			if (only_send_advcmd)
			{
				printf (_(">>>>> send advanced command to core @ %s:%u (%s:%s) - '%s'\n"),
					shutdown_host, shutdown_port,
					shutdown_user, shutdown_pass,
					shutdown_advcmd);
			}
		}
	}

	guicoreconn_do_shutdown_connection ();
	g_assert_not_reached();
}


/**************************************************************************
 *
 *   main_onIdleBringUpDialogWhenInMainLoop
 *
 **************************************************************************/

static gboolean
main_onIdleBringUpDialogWhenInMainLoop (gpointer data)
{
	gui_connect_dialog();

	return FALSE; /* never call us again */
}


/**************************************************************************
 *
 *   main_initiate_auto_connect_if_requested
 *
 **************************************************************************/

static void
main_initiate_auto_connect_if_requested(void)
{
	connection_closed_deliberately = TRUE; /* to avoid reconnect + connect dialog at same time */

	/* bring up dialog once we're in gtk main loop, in case auto-connect on start-up is switched off */
	if (!opt_get_bool(OPT_GUI_AUTO_CONNECT_TO_CORE_ON_STARTUP))
	{
		gtk_idle_add (main_onIdleBringUpDialogWhenInMainLoop, NULL);
		return;
	}

	if (!guicoreconn_set_connection ( opt_get_str(OPT_GUI_DEFAULT_CORE_HOST),
	                                 opt_get_int(OPT_GUI_DEFAULT_CORE_PORT),
	                                 opt_get_str(OPT_GUI_DEFAULT_CORE_USER),
	                                 opt_get_str(OPT_GUI_DEFAULT_CORE_PASS),
	                                 NULL) )
	{
		/* bring up dialog once in gtk main loop in case connection structure could not be set up */
		g_printerr (_("Auto-connect on startup failed - couldn't initiate valid connection structure with default values.\n"));
		gtk_idle_add (main_onIdleBringUpDialogWhenInMainLoop, NULL); /* bring up dialog once we're in gtk main loop */
		return;
	}

	if (!guicoreconn_connect())
	{
		/* bring up dialog once in gtk main loop in case connection connect failed for some reason */
		g_printerr (_("Auto-connect on startup failed - connection structure hasn't been set up?!\n"));
		gtk_idle_add (main_onIdleBringUpDialogWhenInMainLoop, NULL); /* bring up dialog once we're in gtk main loop */
		return;
	}

	/* Everything seems to have gone smoothly so far. The actual connection will
	 *  go ahead once we are in the gtk main loop. If it fails, the connect-to
	 *  dialog will be brought up anyway, so we don't have to worry about that. */

	return;
}


/******************************************************************************
 *
 *   main
 *
 ***/

int
main (int argc, char *argv[])
{
	/* just being hyper-careful... */
	g_return_val_if_fail (sizeof(gfloat)==4, EXIT_FAILURE);
	g_return_val_if_fail (sizeof(gint  )==4, EXIT_FAILURE);
	g_return_val_if_fail (sizeof(guint )==4, EXIT_FAILURE);
	g_return_val_if_fail (sizeof(gchar )==1, EXIT_FAILURE);
	g_return_val_if_fail (sizeof(gshort)==2, EXIT_FAILURE);

	/* set our own error handler, so we can use g_printerr() our way */
	g_set_printerr_handler ((GPrintFunc)gui_print_error_function);

	/* init gettext:
	 *
	 * Set locale according to environment variables
	 * Set $LANG before executing the GUI to set your preferred language,
	 *   e.g. with "export LANG=fr" (sh)(if it is not already set that is)
	 *
	 */
	(void) setlocale (LC_ALL, "");
	(void) bindtextdomain (PACKAGE, LOCALEPATH);
	(void) textdomain (PACKAGE);

	/* signal handling
	 *
	 * by Mario Sergio Fujikawa Ferreira <lioux@FreeBSD.org>
	 *
	 * XXX - [Tim:] Does SIGCHLD interfere with libgnet in any way?
	 * XXX - [Tim:] should we block SIGTRAP? (I get this sometimes when doing http fetch while running GUI in gdb)
	 *
	 */
	signal(SIGCHLD, sig_chld_handler);       	/* Prevents zombies */
	signal(SIGPIPE, sig_pipe_ignore_handler);	/* Handles post mortem conversations */
	/*sig_pipe_block(); */                   	/* Blocks signals from post morten conversations */


	/* GNet initialisation. Needs to be called _BEFORE_ gtk_init() */
	gnet_init();

	gtk_rc_add_default_file(opt_get_opt_filename_without_instance("custom.gtkrc-2.0"));
	gtk_rc_add_default_file(opt_get_opt_filename_without_instance("font.gtkrc-2.0"));

	/* GTK initialization (strips GTK commandline args if there are any) */
	gtk_init (&argc, &argv);

	main_parse_commandline_args(argc, argv);

	if (only_send_shutdown==1 || only_send_advcmd==1)
	{
		main_send_shutdown_or_advanced_command_to_core();
		g_assert_not_reached();
	}

	create_default_styles ();	/* create some styles */

	opt_read_from_disk();
	stats_read_from_disk ();
	downloadsattriblist_load();

	/* if we have been passed an ed2k-link as argument,
	 *  the linkhandler window might take over and not
	 *  the normal main GUI window.
	 */
	if (linkhandler_main(main_ed2k_link_passed_on_commandline))
	{
		g_free(main_ed2k_link_passed_on_commandline);
		exit(0);
	}

	logfile_open();

	(void) pipe_create();           	/* create named pipe   */

	read_filters();                   	/* read search filters */
	check_if_blacklistfile_changed ();	/* read blacklist file */

	// add timer that takes care of stuff that needs to be done regularly but is
	// not important. called every minute
	general_update_timer_id = g_timeout_add (1000*60, general_update_timer, NULL);

	icons_init ();

	/* create main window */
	mainwindow_create();

	//icons_init_pixmaps ();

	toolbar_set_not_connected_to_core();
	search_page_set_initial_extra_search_options_visibility();


	// if GUI option is set so, try to spawn a local donkey if it is not already running

	if (opt_get_bool(OPT_GUI_SPAWN_CORE_ON_STARTUP_IF_NEEDED))
	{
		if ( donkey_process_running_check() <=0)
		{
			status_msg (_("GUI: auto-spawn: no local core running, we need to spawn one.\n"));
			(void) spawn_core_and_print_spawn_result_message(opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH));
		}
		else
			status_msg (_("GUI: auto-spawn: local core already running.\n"));
	}

	clean_up_stuff();

	/* Auto-connect to core on start-up if GUI option is set so  */
	main_initiate_auto_connect_if_requested();

	gtk_main();

	// note: we saved the window options (position, width etc) with main win delete event
	opt_write_to_disk();
	write_gui_uploadstats();
	stats_save_to_disk();
	write_filters();
	free_filters ();
	cleanup_after_disconnect();

	logfile_close();

	G_FREE(main_ed2k_link_passed_on_commandline);

	pipe_unlink();

	return(0);
}
