/***************************************************************************
         parse_message.c  -  parses the message packets from the core

                             -------------------
    begin                : June 2001
    copyright            : (C) 2001-2002 by Tim-Philipp Mller
    email                : t.i.m at orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// TODO:
//
// * If 'download corrupted' message comes up, find the file record from the filename
//	and set the 'corrupted' flag instead of the way we do it now.

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <stdlib.h>

#include "guicoreconn.h"
#include "parse_message.h"
#include "packet.h"
#include "gui_core_protocol.h"
#include "send_command.h"
#include "search.h"
#include "http_get.h"
#include "tag.h"

#include "blacklist.h"
#include "connect_gui.h"
#include "linkhandler.h"
#include "mainwindow.h"
#include "misc_gtk.h"
#include "notebook.h"
#include "progressbar.h"
#include "search.h"

#include "servers.h"
#include "shared_dirs.h"
#include "shared_files.h"
#include "shared_page.h"
#include "stats.h"
#include "statusbar.h"
#include "status_page.h"
#include "toolbar.h"
#include "uploads.h"
#include "uploads_page.h"
#include "downloads.h"
#include "options.h"
#include "options_page.h"
#include "global.h"
#include "misc.h"
#include "notebook.h"
#include "gui_flags.h"

#include "shared_dirs.h"


/* global variables */

/* we set this whenever the user sends an advanced command,
   so we know when we let output pass or when to suppress it */
time_t            user_just_sent_advanced_command = 0;

/* local variables */

static guint      print_next_clientID; /* 0 */ /* only the first after connect to status log */

static time_t     last_connecting; /* 0 */

/* last time we RECEIVED a client status packet with the space free/needed status */
static time_t     space_stats_last_updated; /* 0 */

/* The core doesn't tell us other than by sending a log message (grrrrrr @ Jed) */
static gboolean   last_status_message_was_download_corrupted; /* FALSE */

/* We need this outside of the function, so the p_m_receive_dl_status_list()
 *  routine can call the stats register function with the up-to-date UL speed
 *  value at hand. We need to pass that value to the DL status function somehow.
 */
static gfloat     total_ul_speed = 0.0;


/* a flag that indicates that we've received options correctly
 * after detecting that we are dealing with overnet */
static gboolean   parse_message_received_overnet_options;  /* FALSE */

static gboolean   FLAG_v495_style_download_status_mapping; /* FALSE */


/* local functions */

static void         parse_message_receive_dl_status_list (void);

static void         parse_message_receive_ul_status_list (gboolean newslotsystem);

static gfloat       parse_message_receive_dl_status_items (guint items);

static void         parse_message_read_shared_files (ramfile *packet);

static void         parse_message_read_serverlist (ramfile *packet);

static void         parse_message_read_new_download (ramfile *packet);

static void         parse_message_receive_options (void);


/******************************************************************************
 *
 *  parse_message_receive_dl_status_items
 *
 *  returns total download speed
 *
 *  Note:
 *   - core versions up to incl. 48.2 use a 16-bit index number to
 *     map the status info to the DL (e.g. = number in list)
 *   - core versions from 49.5 onwards will send a unique 32-bit ID
 *     (a pointer essentially, so always non-NULL) to map the status
 *     info to a download. That unique ID will also be sent at the end
 *     of the NEW_DOWNLOAD message, so we know which ID goes with which
 *     download.
 *
 ***/

static gfloat
parse_message_receive_dl_status_items (guint items)
{
	guint32   trans, id, i;
	guint8    status, avail, sources;
	gfloat    speed, total_dl_speed = 0.0;

	for (i=0; i<items; i++)
	{
		if (FLAG_v495_style_download_status_mapping)
			id = read_32();
		else
			id = (guint32) (read_16() + 1); /* need to increment by one, because          */
			                                /* we can't store NULL pointers in hash table */

		status  = read_8();
		speed   = read_float();
		trans   = read_32();
		avail   = read_8();
		sources = read_8();

		downloads_set_download_status (id, status, speed, trans, avail, sources);

		/* Note: when a corruption occurs, we might get astronomically high
		 * speed values for a couple of seconds. ignore those for the total */

		if ( speed > 2048.0 || speed < 0.0 )
			g_printerr (_("ignored out-of-bounds total download speed (>2MB/s or <0kB/s) for statistics\n"));
		else
			total_dl_speed += speed;
	}

	return total_dl_speed;
}


/******************************************************************************
 *
 *  parse_message_receive_dl_status
 *
 *  Receives an parses a download status packet from the core.
 *
 *  Note: we have to keep track of any file priority changes
 *        ourselves. With CORE_NEW_DOWNLOAD we only receive the
 *        'initial' priority state, but no updates after that.
 *
 *  Note: It is possible that the core sends a status packet on connect by
 *        the GUI before we have received the CORE_NEW_DOWNLOAD messages.
 *        (those are being ignored then)
 ***/

#define STAT_PACKET_INTERVAL_IN_S 3

static void
parse_message_receive_dl_status_list (void)
{
	static time_t  last_status_packet = 0;
	gint           timediff = STAT_PACKET_INTERVAL_IN_S;
	gfloat         total_dl_speed = 0.0;

	if (last_status_packet > 0)
		timediff = CLAMP ((time(NULL)-last_status_packet), 2, 4);

	last_status_packet = time(NULL);

	total_dl_speed = parse_message_receive_dl_status_items( read_16() );

	if (total_dl_speed > 0.0)
	{
		seconds_download_was_nonzero += timediff;
		total_downloaded +=  total_dl_speed * ((gfloat) timediff);
	}

	/* Upload stat packet is sent by core before download
	 *  stat packet, that's why we call the function here */
	stats_register_upload_download_stats_packet ( time(NULL),
	                                              total_ul_speed,
	                                              total_dl_speed,
	                                              0, 0);

	toolbar_set_dl_speed (total_dl_speed);
}



/******************************************************************************
 *
 * parse_message_receive_ul_status_list
 *
 ****/

static void
parse_message_receive_ul_status_list (gboolean newslotsystem)
{
	static time_t   lastpacket = 0;
	gfloat          total_ul_inc;
	guint           timediff, i, num;
	gfloat          ulspeed;

	timediff = (guint)(time(NULL)-lastpacket);

	if (timediff>5)
			timediff = 3;

	total_ul_speed = 0.0;

	num = read_16();

	for (i=0; i<num; i++)
	{
		if (!newslotsystem)
		{
			guint indexpos = read_16();	/* upload index number */
			ulspeed = read_float();

			uploads_set_speed (ulspeed, indexpos, timediff);
		}
		else
		{
			guint32 slotid = read_32();
			ulspeed = read_float();

			uploads_set_slot_speed (ulspeed, slotid, timediff);
		}

		total_ul_speed += ulspeed;
	}

	if (total_ul_speed > 0.0)
		seconds_upload_was_nonzero += timediff;

	lastpacket = time(NULL);

	toolbar_set_ul_speed (total_ul_speed);

	total_ul_inc = (timediff*1.0) * total_ul_speed;

	/* more than 10000 kB/s = ca. 10 MB/s upload? unlikely */
	g_return_if_fail (total_ul_inc < 10000.0);

	total_uploaded += total_ul_inc;
}


/******************************************************************************
 *
 *  PM_show_no_server_flags_msg_timeout
 *
 ****/

static gint PM_show_no_server_flags_msg_handle; /* 0 */

gboolean
PM_show_no_server_flags_msg_timeout (gpointer data)
{
	/* static, so we won't show this message after disconnecting
	 *  and reconnecting to core, which is strictly speaking a bug,
	 *  but who cares? ;) */

	static gboolean msgflag = FALSE;

	if (!msgflag && !IS_OVERNET && !IS_HYBRID)
	{
		status_message_blue(_("GUI: For you information: this core does not tell us the\n"
		                      "'always stay connected' and 'automatically remove dead servers'\n"
		                      "options properly (core bug), therefore there will be harmless\n"
		                      "GUI-core inconsistencies.\n"));
		status_message_blue(_("GUI: Please ignore the previous message if this is an overnet core ;)\n"));
	}

	msgflag = TRUE;

//	PM_show_no_server_flags_msg_handle = 0; // leave it set, so we don't set up a timeout again

	return FALSE; /* Don't call again */
}


/******************************************************************************
 *
 *  parse_message_receive_options
 *
 ****/

static void
parse_message_receive_options (void)
{
	gint   dummy;

	/* we set this to 0 again here, because we set it
	 * to 1 when we try to do autoconnect at startup,
	 * so we don't get two donkeys if initial spawn
	 * worked, but not the connect. */
	connection_closed_deliberately = 0;

	coreversion = read_16();  /* core client version */

	/* chances are that we have a test version numbering that
	 * doesn't fit into 16 bits, e.g. we get 39376 instead of 104912 = test v49.12 */
	if (coreversion > 20000)
		coreversion = (guint16)(((guint32) coreversion + 65536)%10000);

	/* e.g. 10454 = test version 45.4 */
	if (coreversion > 10000)
		coreversion = coreversion/10;

	/* e.g. 454 = version 45.4 */
	if (coreversion > 100 && coreversion < 1000)
		coreversion = coreversion/10;

	option_max_dl_speed = 1.0 * int_from_float_read();
	option_max_ul_speed = 1.0 * int_from_float_read();

	option_doorport = read_16();

	dummy = read_16();              /* max connections (ooops, double)*/

	G_FREE(option_nickname);
	option_nickname = read_i4str();

	G_FREE(option_temporary);
	option_temporary = read_i4str();

	G_FREE(option_incoming);
	option_incoming = read_i4str();

	if ( IS_OVERNET == FALSE  && IS_HYBRID == FALSE )
	{
		option_always_connected = read_8();
		dummy                   = read_8(); /* autoReconnectUIFlag (not used)         */
		option_autoremove       = read_8();

		/* new cores (>= v43) will always give us 0 for the above options (bug in guicontroller.cpp)
		 * let's assume this stays with us for a while... ( no <= test any more)
		 * If we don't get them from the core, set them to the most useful values */
		option_always_connected = 1; /* *** FIXME *** */
		option_autoremove = 1;

		/* Give us 5 secs to detect overnet and suppress the message, otherwise show message */
		if (!PM_show_no_server_flags_msg_handle)
			PM_show_no_server_flags_msg_handle = g_timeout_add(5000, PM_show_no_server_flags_msg_timeout, NULL);
	}
	else
	{
		dummy = read_8(); /* donkey 'always connected' flag         */
		dummy = read_8(); /* autoReconnectUIFlag (not used)         */
		dummy = read_8(); /* donkey 'auto-remove dead servers' flag */
		parse_message_received_overnet_options = TRUE;
	}

	option_allowpm         = read_8();
	option_save_corrupted  = read_8();
	dummy                  = read_8();  /* verify cancel - we ignore this and use our own */
	option_adminport       = read_16();
	option_max_connections = read_32();

	core_build_date = (gint) read_32(); /* core build date - all newer cores have that    */

	option_line_down_speed = int_from_float_read();

	/* core PID - all newer cores have that (but a 'package sent too short bug'
	 *  does not always give us this data) */
	dummy = read_32();
	if ( dummy != -1 )
		local_donkey_pid = dummy;

	/* Return here if this is not the very first time that we got the options.
	 *  After all, the core version won't change in one and the same GUI-core session, etc. */
	if (!first_time_we_get_options)
		return;

	if ( core_build_date != 0 && core_build_date >= CORE_TOO_FAR_IN_THE_FUTURE)
	{
		misc_gtk_ok_dialog (GTK_MESSAGE_WARNING,
			 _("This version of the core claims to have been built in\n"
			   "the future :) In other words: There is something wrong.\n"
			   "Please get a recent core from the GUI website:\n\n"
			   "http://ed2k-gtk-gui.sf.net/index.shtml\n"));
	}

	if (1==1)
	{
		gchar  cbdstr[128];      /* core build date string            */
		gchar *verstr = NULL;    /* all version stuff together        */
		gchar *cpidstr = NULL;   /* core pid string (if on localhost) */

		g_snprintf (cbdstr, sizeof(cbdstr), " (build %02u%02u%04u)",
		                                     core_build_date&0xff,                     /* day   */
		                                     (core_build_date&0xff00)>>8,              /* month */
		                                     ((core_build_date&0xff0000)>>16)+2000);   /* year  */

		if (local_donkey_pid>0)
		{
			gchar  *bn, *binname;

			bn = (CORE_IS_LOCAL) ? misc_get_basename_from_pid(local_donkey_pid,NULL) : g_strdup("");

			binname = (CORE_IS_LOCAL) ? g_strdup_printf (" - '%s'", (bn!=NULL) ? bn : "?") : g_strdup("");

			cpidstr = g_strdup_printf (" (%spid=%u%s)", (CORE_IS_LOCAL) ? "" : _("remote "), local_donkey_pid, binname);

			g_free(binname);
			G_FREE(bn);
		}
		else cpidstr = g_strdup_printf (_(" (%s, pid not available)"), (CORE_IS_LOCAL) ? _("local") : _("remote"));

		verstr = g_strdup_printf ("%s: %u%s%s", _("Core version"), coreversion, cbdstr, cpidstr);

		toolbar_set_core_version_string (verstr);

		first_time_we_get_options = 0;

		linkhandler_check_for_pending_links();

		G_FREE(cpidstr);
		G_FREE(verstr);
	}
}

int timer_get_new_serverlist_10s_after_connect (gpointer data)
{
        send_get_serverlist();
        return 0;       /* we want to be called only once */
}


// we want to move all error items to the back, because the core doesn't send stats for them (changing the item
// numbering!)
// hmm, I'm not sure if this is true, let's not do it for now.

void downloadlist_move_error_items_to_the_back(void)
{
	;
}



// parse_status_message_from_core
//
// Often the core only tells us certain things via a status text message, in other
//	cases we want to explain some messages or act on certain messages. Do all
//	that here.
//
// Returns TRUE if the message is to be shown to the user, otherwise FALSE

enum
{
CORE_STATUS_MSG_CANTCONNECT,
CORE_STATUS_MSG_CONNECTINGTO,
CORE_STATUS_MSG_PLACEDONCONNQ,
CORE_STATUS_MSG_DONELOADINGFILES,
CORE_STATUS_MSG_CORRUPTEDDOWNLOAD,
CORE_STATUS_MSG_IMPORTFAILED,
CORE_STATUS_MSG_OUTOFDATE,
CORE_STATUS_MSG_OVERNETID,			/* this must come after the client ID */
CORE_STATUS_MSG_USERS,
CORE_STATUS_MSG_NOTALLWRITTEN,
CORE_STATUS_MSG_NOMORETOEXTENDTO,
CORE_STATUS_MSG_YOURIDHASCHANGED,
CORE_STATUS_MSG_LOADINGSHAREDFILES,
CORE_STATUS_MSG_NOTCONNECTEDTOASERVER,
CORE_STATUS_MSG_CONNECTED,
CORE_STATUS_MSG_ALREADYCONNECTED,
CORE_STATUS_MSG_ERRORLOWID,
CORE_STATUS_MSG_OVERNETOPEN,
CORE_STATUS_MSG_OVERNETFW,
CORE_STATUS_MSG_OVERNETUNCLEAR,
CORE_STATUS_MSG_HYBRID,
CORE_STATUS_MSG_CONTACTLIST,
CORE_STATUS_MSG_CONTACTS,
CORE_STATUS_MSG_BRACKET, // should be LAST!
CORE_STATUS_MSG_LAST,
} ;

static const gchar *coreStatMsgs[CORE_STATUS_MSG_LAST] =
{
"Can't connect to",
"Connecting to",
"Placed on connection Queue",
"Done loading shared",
"Corrupted Download:",
"Import failed",
"server is out of date",	/* Roman Hodek: my core says "your client" not "the client" */
"ID: ",						/* this must come after the client ID (substring) */
"Users:",	// from 'g' command - we want to suppress output of that (only need client ID)
"ot all written",
"No more servers to extend to",
"Your ID has changed",
"Loading shared files",
"Not connected to a server",
"Connected to:",
"Already Connected",
"not reachable. You have a LOWID",
"(open)",
"(firewalled)",
"(unclear)",
"Shibboleth: ",
"Contact List",
"contacts",
"("		// the opening bracket of a peer from 'vp' or of individual DL from 'space'
} ;


/* parse_status_message_from_core
 *
 * returns FALSE if message should NOT be shown to user/put in statuslog
 *
 */

static gboolean
parse_status_message_from_core (const gchar *msg)
{
	guint			 msgnum;
	static gchar	*last_overnet_status = NULL;
	static gboolean	 lastMsgWasLoadingSharedFiles = FALSE;

	g_return_val_if_fail (msg!=NULL, FALSE);

	for (msgnum=0; msgnum < (sizeof(coreStatMsgs)/sizeof(gchar*)); msgnum++)
	{
		if (strstr(msg, coreStatMsgs[msgnum])!=NULL)
			break;
	}

	if (user_just_sent_advanced_command > 0  &&  (time(NULL) - user_just_sent_advanced_command) > 15 )
		user_just_sent_advanced_command = 0;

	switch (msgnum)
	{
		case CORE_STATUS_MSG_BRACKET:
		{
			guint  num;
			gchar  letter;

			if ( user_just_sent_advanced_command > 0 )
				return TRUE;

			if ( sscanf(msg, "(%u) %c", &num, &letter) == 2 )
				return FALSE;

			if ( sscanf(msg, "(%u)\t%c", &num, &letter) == 2 )
				return FALSE;

			/* a contact list entry */
			if ( sscanf(msg, "%u)", &num) == 1 )
				return FALSE;

			/* Roman Hodek: Recognize indiv. space messages by '(' _and_ ":\t" */
			if ( *msg == '('  &&  strstr(msg,":\t") != NULL )
				return FALSE;
		}
		break;

		case CORE_STATUS_MSG_CANTCONNECT:
		{
			servers_register_connect_status (NULL, SERVERS_CONNECT_STATUS_CANTCONNECT);
			toolbar_set_server_data (NULL, 0, 0);
			last_connecting = 0;
		}
		break;

		case CORE_STATUS_MSG_CONNECTINGTO:
		{
			// Find server name or IP from whole line
			gchar *ipOrName = g_strdup(msg+14);
			gchar *lastbit = strstr(ipOrName, " ...");
			if (lastbit) *lastbit=0x00;
			servers_register_connect_status (ipOrName, SERVERS_CONNECT_STATUS_CONNECTING);
			toolbar_set_server_data (_(" [connecting] "), -1, -1);
			g_free(ipOrName);
			last_connecting = time(NULL);
		}
		break;
		case CORE_STATUS_MSG_PLACEDONCONNQ:
		{
			/* FIX ME: why no servers_register_connect_status() here? */
			toolbar_set_server_data (_(" [on connection queue] "), -1, -1);
		}
		break;
		case CORE_STATUS_MSG_DONELOADINGFILES:
			send_get_shared_files();
			send_get_shared_dirs();
			shared_page_set_stats_label_message (_("Done loading shared files."));
		break;
		case CORE_STATUS_MSG_CORRUPTEDDOWNLOAD:
			last_status_message_was_download_corrupted = TRUE;
		break;
		case CORE_STATUS_MSG_IMPORTFAILED:
			status_msg ("%s.\n", _("GUI: This might be because you only have read, but no write permission on that file"));
		break;
		case CORE_STATUS_MSG_OUTOFDATE:
			status_msg ("%s.\n", _("GUI: Server is rejecting us. *pah*"));
			servers_register_connect_status (NULL, SERVERS_CONNECT_STATUS_THINKS_WE_ARE_FIREWALLED);
		break;

		case CORE_STATUS_MSG_OVERNETID:
		{
			gchar overnetid[33];

			if (sscanf(msg, "ID: %32[0-9a-fA-F]", overnetid))
			{
				gchar *fullmsg = g_strdup_printf ("%s%s", msg, (last_overnet_status!=NULL) ? last_overnet_status : "");

				servers_set_client_id(fullmsg);
				status_page_print_clientID_info (fullmsg);

				if (print_next_clientID==1)
					status_msg(fullmsg);

				if (!core_is_overnet)
				{
					core_is_overnet = TRUE;

					status_message_blue(_("GUI: This seems to be an OVERNET core.\n"));
					mainwindow_set_title ("overnet");
					toolbar_set_overnet_mode(TRUE);
					servers_set_overnet_mode(TRUE);
					search_set_overnet_mode (TRUE);	/* XXX - DO WE NEED TO CHECK FOR IS_HYBRID HERE? */
					notebook_set_overnet_mode (TRUE);
					send_get_overnet_contacts();
				}


				G_FREE(fullmsg);
			}
			print_next_clientID=0;
		}
		return ( user_just_sent_advanced_command > 0 );

		case CORE_STATUS_MSG_YOURIDHASCHANGED:	/* just supress this */

			user_just_sent_advanced_command = 0;
			return FALSE;

		case CORE_STATUS_MSG_USERS:
		{
			user_just_sent_advanced_command = 0;

			if (IS_OVERNET)
			{
				guint users, files;
				if (sscanf(msg, "Users: %u   Files: %u", &users, &files)==2)
				{
					toolbar_set_server_data ("overnet", files, users);
				}
			}
		}
		return FALSE;

		/* note: no more support for parsing old cores' space stats. we've got
		 *       a GUI message for that now in recent cores                    */

		case CORE_STATUS_MSG_NOTALLWRITTEN:
			g_printerr (_("The above error _might_ be due to a full harddisk.\n"));
		break;

		case CORE_STATUS_MSG_NOMORETOEXTENDTO:
			statusbar_extra_message (" %s", msg);
		return FALSE;

		case CORE_STATUS_MSG_LOADINGSHAREDFILES:
		{
			if (lastMsgWasLoadingSharedFiles)
				return FALSE;
			lastMsgWasLoadingSharedFiles = TRUE;
			shared_page_set_stats_label_message (msg);
		}
		return TRUE;

		case CORE_STATUS_MSG_NOTCONNECTEDTOASERVER:
		{
			/* we must set the 'search' button sensitive again if we couldn't search
			 * because we are not connected to a server
			 */
			gtk_widget_set_sensitive(button_extend, FALSE);
			gtk_widget_set_sensitive(button_more, FALSE);
//			gtk_widget_set_sensitive(button_addsearch, TRUE);
		}
		return TRUE;

		case CORE_STATUS_MSG_CONNECTED:
		{
			static gchar server[33];	/* 255.255.255.255 + '\0' */
			memset (server, 0x00, sizeof(server));
			sscanf (msg, "Connected to: %30s", server);
			toolbar_set_server_data (server, -1, -1);
			send_get_client_stats();
		}
		return TRUE;

		case CORE_STATUS_MSG_ALREADYCONNECTED:
		{
			return FALSE;
		}

		case CORE_STATUS_MSG_ERRORLOWID:
		{
			/* XXX - do this even if OPT_GUI_SERVERS_DISCONNECT_IF_FIREWALLED is not set?
			 * (Problem is: whether server closes connection on lowid,
			 *  depends on server, not on core)
			 */
			if (opt_get_bool(OPT_GUI_SERVERS_DISCONNECT_IF_FIREWALLED))
			{
				servers_register_connect_status (NULL, SERVERS_CONNECT_STATUS_CANTCONNECT);
				toolbar_set_server_data (NULL, 0, 0);
			}
		}
		return TRUE;

		case CORE_STATUS_MSG_OVERNETOPEN:
		{
			if (g_ascii_strcasecmp(msg,"(open)")==0)
				last_overnet_status = _(" (open) ");
		}
		return ( user_just_sent_advanced_command > 0 );

		case CORE_STATUS_MSG_OVERNETFW:
		{
			if (g_ascii_strcasecmp(msg,"(firewalled)")==0)
				last_overnet_status = _(" (firewalled) ");
		}
		return ( user_just_sent_advanced_command > 0 );

		case CORE_STATUS_MSG_OVERNETUNCLEAR:
		{
			if (g_ascii_strcasecmp(msg,"(unclear)")==0)
				last_overnet_status = _(" (unclear) ");
		}
		return ( user_just_sent_advanced_command > 0 );

		case CORE_STATUS_MSG_HYBRID:
		{
			core_is_hybrid = TRUE;
		}
		return FALSE;

		case CORE_STATUS_MSG_CONTACTLIST:

			return ( user_just_sent_advanced_command > 0 );

		case CORE_STATUS_MSG_CONTACTS:
		{
			gboolean  ret = ( user_just_sent_advanced_command > 0 );

/* the 'boot' command in overnet doesn't seem to work properly.
 *  No reply from Jed as to whether it should work or not, so I can't
 *  do much about it for now apart from leaving it. Apparently the
 *  contacts somehow get to the core, because they are saved in
 *  contacts.dat, but the booting just doesn't happen :-( */
#if 0
			if ( g_ascii_strcasecmp (msg, "0 contacts") == 0
			      ||  g_ascii_strcasecmp (msg, "0 (0) contacts") == 0 )
			{
				misc_boot_overnet();
				user_just_sent_advanced_command = 0;
			}
			else
			{
				misc_boot_overnet_stop_booting();
			}
#endif

			return ret;
		}
		break;

		default:
		break;
	}

	lastMsgWasLoadingSharedFiles = FALSE;

	user_just_sent_advanced_command = 0;

	return TRUE;
}



static void
parse_message_extend_5x (void)
{
	send_extended_search();
	send_extended_search();
	send_extended_search();
	send_extended_search();
	send_extended_search();
}


static void
parse_message_get_results (guint num, gboolean use_overnet)
{
	guint i;

	for (i=0; i<num; i++)
	{
		GPtrArray	*taglist;
		guint        size        = 0;
		guint        avail       = 0;
		guint        bitrate     = 0;
		guint8		*servicehash = NULL;
		guint8		*hash        = NULL;
		guint8		*unused      = NULL;
		const gchar *format      = NULL;
		const gchar *type        = NULL;
		const gchar *artist      = NULL;
		const gchar *title       = NULL;
		const gchar *album       = NULL;
		const gchar *length      = NULL;
		const gchar *codec       = NULL;
		const gchar *name        = "(no name)";

		if ( use_overnet )
		{
			servicehash = read_x(16);
			hash        = read_x(16);
		}
		else
		{
			const guint8 zero[4+2] = { 0x00,0x00,0x00,0x00,0x00,0x00 };

			hash   = read_x(16);
			unused = read_x(4+2);

			if ( (unused) && memcmp(unused,zero,6)!=0 )
			{
				g_printerr ("ATTENTION: Jed has changed FileMeta->toNetwork()! This needs to be fixed in the GUI.\n");
				return;
			}
		}

		taglist = taglist_read_from_ramfile (ramfile_now_active);

		if ( taglist_get_string_tag ( taglist, TAGNAME_FILENAME_SHORT, &name ) == TRUE )
		{
			/* remove bad characters (like a slash) from the filename */
			(void) g_strdelimit((gchar*)name, "/\\&", ' ');
		}

		(void) taglist_get_int_tag ( taglist, TAGNAME_FILESIZE_SHORT, &size );
		(void) taglist_get_int_tag ( taglist, TAGNAME_AVAILABILITY_SHORT, &avail );
		(void) taglist_get_int_tag ( taglist, "bitrate", &bitrate );

		(void) taglist_get_string_tag ( taglist, "Artist", &artist );
		(void) taglist_get_string_tag ( taglist, "Title",  &title );
		(void) taglist_get_string_tag ( taglist, "Album",  &album );
		(void) taglist_get_string_tag ( taglist, "length", &length );
		(void) taglist_get_string_tag ( taglist, "codec",  &codec );

		(void) taglist_get_string_tag ( taglist, TAGNAME_FILETYPE_SHORT,    &type );
		(void) taglist_get_string_tag ( taglist, TAGNAME_FILEFORMAT_SHORT,  &format );

		search_add_or_update_record ( search_get_current_search(),
		                              hash, servicehash, size, avail, name, type, format,
		                              artist, album, title, length, bitrate, codec);

		if (taglist)
			taglist_free(taglist);

		G_FREE(hash);
		G_FREE(servicehash);
		G_FREE(unused);
	}
}


/******************************************************************************
 *
 *   parse_message_read_shared_files
 *
 ***/
static void
parse_message_read_shared_files (ramfile *packet)
{
	gboolean         manyfiles;
	guint            i, num;

	if (!shared_files_no_delayed_adds_in_progess())
		return;

	shared_files_mark_all_for_removal();

	core_supports_upload_priority = FALSE;

	num = read_32();

	manyfiles = ( num > 3000 );

	if ( manyfiles )
	{
		static gboolean shwn = FALSE;

		shared_files_disable_sorting();
		opt_set_bool(OPT_GUI_SHOW_ICONS_IN_SHARED_FILES_LIST, FALSE);

		if (!shwn)
		{
			status_warning (_("GUI: Disabled sorting in the shared files list. Too many files ( > 3000 ).\n"));
			shwn = TRUE;
		}
	}
	else
		shared_files_enable_sorting();

	for ( i = 0;  i < num;  i++ )
	{
		guint8       *hash    = read_x(16+4+2);	/* last six bytes are unused */
		GPtrArray    *taglist = taglist_read_from_ramfile (packet);
		guint         size    = 0xFFFFFFFF;
		guint         prio    = 1;	/* normal by default */
		const gchar  *name    = "(no name)";
		const gchar  *type    = NULL;
		const gchar  *format  = NULL;

		g_return_if_fail ( taglist != NULL );

		if ( taglist_get_int_tag ( taglist, TAGNAME_PRIORITY_SHORT, &prio) == TRUE )
			core_supports_upload_priority = TRUE;

		taglist_get_int_tag      ( taglist, TAGNAME_FILESIZE_SHORT,   &size);
		taglist_get_string_tag   ( taglist, TAGNAME_FILENAME_SHORT,   &name );

		taglist_get_string_tag   ( taglist, TAGNAME_FILEFORMAT_SHORT, &format );
		taglist_get_string_tag   ( taglist, TAGNAME_FILETYPE_SHORT,   &type );

		if (!format)
			format = misc_get_extension_from_filename (name);

		if (!type)
			type = misc_get_filetype_from_extension (format);

		shared_files_add_or_update_record (hash, name, (guint32) size, format, type, prio);

		if (taglist)
			taglist_free(taglist);

		G_FREE(hash);
	}

	shared_files_remove_old ();
}


/******************************************************************************
 *
 *   parse_message_read_serverlist
 *
 ***/

static void
parse_message_read_serverlist (ramfile *packet)
{
	guint i, num;

	servers_mark_all_for_removal();

	num = read_32();

	g_return_if_fail ( num < 2000 );	/* more than 2000 servers is unlikely */

	for (i=0; i < num; i++)
	{
		guint8       *notused;
		guint32       ip, users = 0, files = 0, ping = 0, pref = 1;
		guint16       port;
		GPtrArray    *taglist;
		const gchar  *name = NULL;
		const gchar  *desc = NULL;

		notused = read_x(16);
		ip      = read_IP();
		port    = read_16();

		taglist = taglist_read_from_ramfile(packet);

		taglist_get_string_tag   ( taglist, TAGNAME_FILENAME_SHORT,          &name );
		taglist_get_string_tag   ( taglist, TAGNAME_SERVERDESCRIPTION_SHORT, &desc );

		taglist_get_int_tag      ( taglist, "users",                    &users);
		taglist_get_int_tag      ( taglist, "files",                    &files);
		taglist_get_int_tag      ( taglist, TAGNAME_PING_SHORT,         &ping );
		taglist_get_int_tag      ( taglist, TAGNAME_PREFERENCE_SHORT,   &pref );

		servers_add_or_update_record ( ip, port, name, desc, users, files, ping, (gint8)(pref&0xff) );

		if (taglist)
			taglist_free(taglist);

		G_FREE(notused);
	}

	servers_remove_all_old();

	statusbar_extra_message (" ");
}



/******************************************************************************
 *
 *   parse_message_read_new_download_get_tempfilename
 *
 *   Different core version unfortunately give us different things as
 *    the temp file string. This function returns a full path temp file
 *    string (the one used for previews) based on the core version detected.
 *
 *   Returns NULL or temp file string.
 *
 ***/

static gchar *
parse_message_read_new_download_get_tempfilename (const gchar *name, const gchar *tfn)
{
	gchar *tempfn = NULL;

	g_return_val_if_fail ( tfn != NULL, NULL );

	/* NOTE: core versions < 48 do not send us the options first thing,
	 *       so we don't know the core version at that point. */

	switch (CORE_VERSION_BASE)
	{
		/* v49: temp file is usually /tmpdir/filename/1.1.part, but can be
		 *      any x.1.part in case of old v48 imports. (TODO: ignore in future releases)
		 *      Therefore: check if 1.1.part exists, and if not read dir and find base number.
		 *      Can also be /tmpdir/x.part in case of old (pre-v48) partstyle, if configured in pref.met */
		case 49:
		case 50:
		{
			if (tfn[strlen(tfn)-1] != '/' && tfn[strlen(tfn)-1] != '\\')   /* not a directory path */
			{
				if ( !CORE_IS_LOCAL || g_file_test(tfn,G_FILE_TEST_EXISTS) )
				{
					tempfn = g_strdup(tfn);
					break; /* TODO: what happens with imported v48 downloads here? */
				}
			}

			/* okay, we did get a directory path as temp filename string */

			tempfn = g_strdup_printf("%s%s%c1.1.part", tfn, name, G_DIR_SEPARATOR);
			if ( CORE_IS_LOCAL && !g_file_test(tempfn,G_FILE_TEST_EXISTS))
			{
				gchar *dname;
				GDir  *dir;

				G_FREE(tempfn); /* sets it to NULL as well */

				dname = g_strconcat(tfn, name, NULL);

				dir = g_dir_open(dname, 0, NULL);
				if (dir)
				{
					const gchar *dentry;
					while ( (dentry = g_dir_read_name(dir)) && (!tempfn) )
					{
						if ( atoi(dentry) > 0)
							tempfn = g_strdup_printf("%s%s%c%d.1.part", tfn, name, G_DIR_SEPARATOR, atoi(dentry));
					}
					g_dir_close(dir);
				}
				G_FREE(dname);
			}
			/* ELSE: either no local core - then we can't check the constructed
			 *       temp file name anyway (but don't need it anyway), or the name
			 *       checked out fine. */
		}
		break;

		/* v48: Gives us /tmpdir/x.part.met, whereas we really need /tmpdir/x.1.part */
		case 48:
		{
			guint  num;
			gchar *dirbasename;

			if (strlen(tfn) < 10) /* strlen("x.part.met") = 10 */
				break;

			if (sscanf(tfn+strlen(tfn)-10, "%u.part.met", &num) != 1)
			{
				/* maybe very old partstyle imported? if it exists, take it */
				if ( (!CORE_IS_LOCAL) || g_file_test(tfn,G_FILE_TEST_EXISTS))
					tempfn = g_strdup(tfn);
				break;
			}

			dirbasename = g_path_get_dirname(tfn);
			tempfn = g_strdup_printf("%s%c%u.1.part",dirbasename, G_DIR_SEPARATOR, num);
			G_FREE(dirbasename);
		}
		break;

		/* Versions before v48: we can just hope they send us the right string. */
		default:
		{
			static gboolean msgshown; /* FALSE */
			if (!msgshown)
			{
				status_warning ("%s ******\n", _("GUI: This core seems to be old. You should really get a newer version."));
				msgshown = TRUE;
			}
			tempfn = g_strdup(tfn);
		}
		break;
	}

	if (!tempfn)
	{
		g_warning("Unexpected v%u temp file string in NEW_DOWNLOAD msg: '%s'."
		          "Please send this to tim@edonkey2000.com, with info whether you are"
		          "using a local core or a remote core (if you know it).\n", CORE_VERSION_BASE, tfn);
		tempfn = g_strdup(tfn);
	}

	return tempfn;
}

/******************************************************************************
 *
 *   parse_message_read_new_download
 *
 ***/

static void
parse_message_read_new_download (ramfile *packet)
{
	const gchar  *name = NULL;
	GPtrArray    *taglist;
	guint32       size = 0, id;
	guint8       *hash, priority;
	gchar        *tempfn = NULL;

	hash = read_x(16+4+2);  /* read unused fields (4+2) too (they do no harm) */

	g_return_if_fail ( hash != NULL );

	taglist = taglist_read_from_ramfile (ramfile_now_active);
	g_return_if_fail ( taglist != NULL );

	taglist_get_string_tag   ( taglist, TAGNAME_FILENAME_SHORT, &name );
	taglist_get_int_tag      ( taglist, TAGNAME_FILESIZE_SHORT, &size );

	priority = read_8();    /* get initial priority state */

	if ( !IS_OVERNET && (hash[16]|hash[17]|hash[18]|hash[19]|hash[20]|hash[21]) != 0 )
	{
		g_printerr ("ATTENTION: Jed has changed FileMeta->toNetwork()! This needs to be fixed in the GUI.\n");
		return;
	}

	/* Note: very early cores have no part number information at all. */
	/* At least 3 more bytes... should be non-ancient core            */
	if (packet_left_from_pos(packet) > 2)
	{
		gchar *tfn;
		tfn = read_i4str();
		if (tfn)
		{
			tempfn = parse_message_read_new_download_get_tempfilename(name, tfn);
			g_free(tfn);
		}
	}

	if (!tempfn)
		status_warning("%s\n", _("This core seems to be old: no temp file info available."));

	/* Cores >= v49.5 have a unique 32-bit ID here now, which
	 *  is used to later map the status msg to the right download. */
	if (packet_left_from_pos(packet) >= 4)
		FLAG_v495_style_download_status_mapping = TRUE;

	if (FLAG_v495_style_download_status_mapping)
		id = read_32();
	else
		id = (downloads_get_num() + 1); /* old style <= v48.2: ID = index number in DL List   */
		/* need to increment by one ^----, because we can't store NULL pointers in hash table */

	downloads_add_download ( hash, name, size, tempfn, priority, id );

	if (CORE_IS_GAP_ENABLED)
		send_get_download_gaps (hash);

	if (taglist)
		taglist_free(taglist);

	G_FREE(hash);
	G_FREE(tempfn);
}


/******************************************************************************
 *
 *   parse_message_read_core_client_stats
 *
 ***/

static void
parse_message_read_core_client_stats (ramfile *packet)
{
	gfloat     tmp, in, needed;
	guint16    on_queue, cons_now;
	guint      clientid;
	gchar     *stattxt = NULL;

	tmp      = float_read();
	in       = float_read();
	needed   = float_read();
	clientid = read_32();
	cons_now = read_16();
	on_queue = read_16();

	if ( !IS_OVERNET && !IS_HYBRID )
	{
		stattxt = g_strdup_printf ("ClientID: %u (%s)\n", clientid, (clientid>0xFFFFFF) ? _("open") : _("firewalled"));
		g_return_if_fail ( stattxt != NULL );

		servers_set_client_id(stattxt);
		status_page_print_clientID_info(stattxt);

		if ( opt_get_bool(OPT_GUI_SERVERS_DISCONNECT_IF_FIREWALLED)  &&  strlen(stattxt) > 10  )
		{
			gint id = atoi(stattxt+10);
			if ( id > 0  &&  id < 0xffffff )
			{
				status_msg ("%s", stattxt);
				status_msg ("%s.\n", _("GUI: Server thinks you are firewalled. Disconnecting."));
				servers_register_connect_status (NULL, SERVERS_CONNECT_STATUS_THINKS_WE_ARE_FIREWALLED);
			}
		}
	}

	downloads_page_set_free_space (needed, tmp, in);

	if (tmp <= (opt_get_int(OPT_GUI_DOWNLOADS_MIN_FREE_DISKSPACE)*1.0))
		downloads_pause_all(tmp);

	uploads_page_set_people_on_queue (on_queue);

	(void) toolbar_set_cons ((guint)cons_now);

	space_stats_last_updated = time(NULL);

	G_FREE(stattxt);
}


// parse_message_from_client
//
// Reads a single message from the GUI and acts upon it

void
parse_message_from_client (ramfile *packet)
{
	static time_t          first_time_not_connecting = 0;
	static gboolean        we_are_connected = FALSE;
	gboolean               use_overnet = IS_OVERNET;
	guint8                 msg;

	g_return_if_fail (packet!=NULL);

	if (!window)
		return;


/* let's not do this: seems to provoke core segfaults through
 * double-connects at the same time (=> core bug):
 */
#ifdef HUMBUG

	/* Make sure we connect to a server if the option is set */
	/*  (the core is a bit sluggish in this respect)         */
	if (last_connecting==0)
	{
		if (first_time_not_connecting==0)
			first_time_not_connecting = time(NULL);
	} else first_time_not_connecting = 0;

	if ((option_always_connected) && (!we_are_connected))
	{
		if (first_time_not_connecting>0 && (time(NULL)-first_time_not_connecting)>45)
		{
			status_message_blue ("GUI DEBUG: connecting. NOW-first_time_not_connecting>45\n");
			send_disconnect ();
			send_connect_to_server(0,0);
			first_time_not_connecting = 0;
		}
		else if (last_connecting>0 && (time(NULL)-last_connecting)>45)
		{
			status_message_blue ("GUI DEBUG: connecting. NOW-last_connecting>45\n");
			/* XXX - should we send a disconnect here (possibly hung at 'placed on conn queue')? */
			send_connect_to_server(0,0);
			last_connecting = 0;
		}
	}

#endif
	// if we've just killed and restarted the core, we want a success message if it
	// worked - here it is
	if (show_restart_success_message_on_next_packet == 1)
	{
		status_message_blue ("%s.\n", _("GUI: restart and reconnect successful"));
		show_restart_success_message_on_next_packet = 0;
	}

	// get and parse message
	ramfile_seek (packet, 5, RAMFILE_SEEK_FROM_START);	// skip header

	msg = read_8();

	switch (msg)
	{
		case -1:		return;	// error (packet short)

		case CORE_STATUS_MSG:
		{
			gchar *msg = read_i4str();
			last_status_message_was_download_corrupted = FALSE;
			if (msg)
			{
				if (parse_status_message_from_core(msg))
					status_msg ("%s\n", msg);
				g_free(msg);
			}
		}
		break;

     		case CORE_ERROR_MSG:
		{
			gchar *msg = read_i4str();
			last_status_message_was_download_corrupted = FALSE;
			if (msg)
			{
				if (parse_status_message_from_core(msg))
					status_warning ("%s", msg);
				g_free(msg);
			}
		}
		break;

		case CORE_HASHING:
		{
			gchar	*name = read_i4str();
			gchar	*txt = g_strdup_printf (_("Hashing %s ..."), name);
			g_return_if_fail (txt!=NULL);
			shared_page_set_stats_label_message(txt);
			g_free(txt);
			g_free(name);
		}
		break;

		case CORE_CONNECTED_TO:
		{
			gchar *servername = read_i4str();
			if (servername)
			{
				toolbar_set_server_data (servername, 0, 0);
				status_msg (_(" --- connected to server '%s' ---\n"), servername);
				g_free(servername);
				servername = NULL;
			}
			servers_reinstate_alwaysconnected_option();
			//servers_connect_queue_clear ();
			// add timer to get a new serverlist 10s after the connect
			(void) g_timeout_add (10*1000, timer_get_new_serverlist_10s_after_connect, NULL);
			print_next_clientID=1;
			send_get_client_stats();
			gtk_widget_set_sensitive(button_extend, FALSE);
			gtk_widget_set_sensitive(button_more, FALSE);
			last_connecting = 0;
			we_are_connected = TRUE;
		}
		break;

		case CORE_SET_SERVER_STATS:
		{
			guint32 files = read_32();
			guint32 users = read_32();
			toolbar_set_server_data (NULL, (guint)files, (guint)users);
		}
		break;

		case CORE_DISCONNECTED:
		{
			status_msg (_(" --- disconnected from server ---\n"));
			toolbar_set_server_data (NULL,0,0);
			gtk_widget_set_sensitive(button_extend, FALSE);
			gtk_widget_set_sensitive(button_more, FALSE);
/*			can't do this, because 'placed on queue' is followed by disconnect as well
			servers_reinstate_alwaysconnected_option();
			servers_clear_connect_to_list();
*/
			last_connecting = 0;
			we_are_connected = FALSE;
			first_time_not_connecting = 0;
		}
		break;

		case CORE_ED_NEW_SEARCH_RESULT:

			use_overnet = FALSE;
			/* note: no break here - go on */

		case CORE_NEW_SEARCH_RESULTS:
		{
			Search *search; 

			// counter, so we don't go on forever asking for more results
			static gint	requested_more_results = 0;

			guint		results, num=read_32();
			gboolean	more; // more results available?
			GString		*actstr;

//			gtk_widget_set_sensitive (button_addsearch, TRUE);

			if (num==0)
			{
				statusbar_extra_message(_(" No results. Automatically extending search to five more servers..."));
				parse_message_extend_5x ();
				gtk_widget_set_sensitive (button_extend, TRUE);
				break;
			}

			search = search_get_current_search();

			g_return_if_fail ( search        != NULL );
			g_return_if_fail ( search->store != NULL );

			parse_message_get_results (num, use_overnet);

			results = (search->store)->length;

			more = (read_8()!=0);
			gtk_widget_set_sensitive (button_more, more);
	
			actstr = g_string_sized_new (256);
			g_string_printf (actstr, _("Search results received. (%u results filtered) "), search->filtered);

			if (results<10 && (!more))
			{
				actstr = g_string_append (actstr, _("Not very many though - extending search..."));
// XXX
				if ( search->isUserList )
					g_printerr ("FIX ME: different extend for user list! (bug)\n");

				parse_message_extend_5x ();
			}

			// make sure we don't request forever
			if ((more) && requested_more_results<20)
			{
				actstr = g_string_append (actstr, _("More results available..."));
				// only ask automatically for more results if
				// we didn't get a single one that we don't
				// have yet - otherwise let user do it manually
				if (results==0)
				{
					gchar numstr[16];
					sprintf (numstr, "(%d)", requested_more_results);
					actstr = g_string_append (actstr, numstr);
//					gtk_widget_set_sensitive (button_addsearch, FALSE);

					// more search for users not available!
					if ( !search->isUserList )
						send_more_search();
					else
						g_print ("FIX ME: different more search for user list! (bug)\n");

					requested_more_results++;
				}
			}
			 else
			{
				actstr = g_string_append (actstr, _("Search completed."));
				requested_more_results=0;
			}
			statusbar_extra_message (" %s", actstr->str);
			gtk_widget_set_sensitive (button_extend, TRUE);
			g_string_free(actstr,TRUE);
		}
		break;

		case CORE_SERVER_LIST:
		{
			parse_message_read_serverlist(packet);
		}
		break;

		case CORE_EXTENDING_SEARCH:
		{
			gchar *servername = read_i4str();
			if (servername)
			{
				statusbar_extra_message (_(" Extending search to server %s..."), servername);
				g_free(servername);
			}
		}
		break;


 		case CORE_ED_SEARCH_RESULT:

 			use_overnet = FALSE;
			/* note: no break here - go on */

		case CORE_SEARCH_RESULT:
		{
			parse_message_get_results (1, use_overnet);
		}
		break;

		case CORE_NEW_DOWNLOAD:
		{
			parse_message_read_new_download (packet);
		}
		break;

		case CORE_REMOVE_DOWNLOAD:
		{
			guint8     remove_hash[16];
			gchar     *name = NULL;
			gboolean   cancelled, corrupted;
			guint      size = 0;

			ramfile_read(ramfile_now_active,remove_hash,16);

			if (!downloads_remove_download (remove_hash, &name, &corrupted, &cancelled, &size))
			{
				g_printerr ("downloads_remove_download() failed in %s:%d\n", __FUNCTION__, __LINE__);
				break;
			}

			if (cancelled)
			{
				status_msg ("%s: ed2k://|file|%s|%u|%s|\n", _("GUI: download cancelled"), name, size, hash_to_hash_str(remove_hash));
				G_FREE(name);
				break;
			}

			status_message_blue (_("----- DOWNLOAD COMPLETE ---------------\n"));
			status_message_blue ("%s\n", name);

			if ((corrupted) || (last_status_message_was_download_corrupted))
			{
				if (option_save_corrupted == 0)
				{
					g_printerr (_("has been corrupted - file has been deleted from disk.\n"));
				}
				else
				{
					g_printerr (_("has been corrupted - file has been saved and is in your temp folder.\n"));
				}

				G_FREE(name);
				break;
			}

			status_message_blue (_("should have been moved to your incoming folder.\n"));

			exec_command_on_complete_download(name);

			send_get_download_gaps(remove_hash);	/* update gaps one last time */

			status_message_blue ("---------------------------------------\n");
			statusbar_extra_message (_(" %s has been completed."), choplast(name));
			G_FREE(name);
		}
		break;

		case CORE_DOWNLOAD_STATUS:
		{
			parse_message_receive_dl_status_list();
		}
		break;

		case CORE_NEW_UPLOAD:
		{
			gchar		*filename = read_i4str();
			guint8		*userhash = read_x(16+4+2);	// read unused fields too (they do no harm)
			GPtrArray	*taglist = taglist_read_from_ramfile (ramfile_now_active);
			NGS_tag		*usernametag = taglist_find_tag_from_tagname (taglist, TAGNAME_FILENAME_SHORT);

			/* Note: in this function, the file hash will be set to zero. With each
			 *	upload status packet we get, we'll check whether the filehash is zero
			 *	and try to find it via the filename (from downloads or shared files) if
			 *	it is. Once we have the filehash, we can do upload statistics...        */
			uploads_add_new_upload (userhash, filename,
				(usernametag) ? (gchar*)usernametag->data : "(?)");

			if (taglist) taglist_free(taglist);
			G_FREE(userhash);
			G_FREE(filename);
		}
		break;

		case CORE_REMOVE_UPLOAD:
		{
			guint8 *userhash = read_x(16);
			if (userhash)
			{
				uploads_remove_upload (userhash);
				g_free(userhash);
			}
		}
		break;

		/* CORE_UPLOAD_STATUS: this message is sent by cores before v45 (v45 sends nothing) */
		case CORE_UPLOAD_STATUS:
		{
			parse_message_receive_ul_status_list(FALSE);
		}
		break;

		/* CORE_NEW_UPLOAD_SLOT: this message is sent by core before v45 (v45 sends nothing) */
		case CORE_NEW_UPLOAD_SLOT:
		{
			g_return_if_fail ( !CORE_HAS_NEW_UPLOAD_SLOT_MANAGEMENT );
		}
		break;

		/* CORE_REMOVE_UPLOAD_SLOT: this message is sent by core versions after v46.0 */
		case CORE_REMOVE_UPLOAD_SLOT:
		{
			guint32 slotid;

			if ( !NO_CORE_VERSION_YET && !CORE_HAS_NEW_UPLOAD_SLOT_MANAGEMENT )
				break;

			slotid = read_32();

			uploads_remove_upload_slot(slotid);
		}
		break;

		/* CORE_UPLOAD_SLOT_STATUS: this message is sent by core versions before v45 */
		case CORE_UPLOAD_SLOT_STATUS:
		{
			g_return_if_fail ( NO_CORE_VERSION_YET || CORE_HAS_NEW_UPLOAD_SLOT_MANAGEMENT );

			parse_message_receive_ul_status_list(TRUE);
		}
		break;

		/* CORE_CHANGE_UPLOAD_SLOT: this message is sent by core versions after v46.0 */
		case CORE_CHANGE_UPLOAD_SLOT:
		{
			const gchar  *username = NULL;
			GPtrArray    *usertaglist;
			guint32       slotid;
			guint8       *filehash, *userhash, *junk;
			gchar        *filename;

			g_return_if_fail ( NO_CORE_VERSION_YET || CORE_HAS_NEW_UPLOAD_SLOT_MANAGEMENT );

			slotid   = read_32();
			filename = read_i4str();
			userhash = read_x(16);
			junk     = read_x(4+2);
			usertaglist  = taglist_read_from_ramfile (ramfile_now_active);
			filehash = read_x(16);

			g_return_if_fail ( filename != NULL );
			g_return_if_fail ( filehash != NULL );
			g_return_if_fail ( userhash != NULL );

			if (slotid == G_MAXUINT) /* TODO: remove debug msg */
				g_print ("CHANGE_UPLOAD_SLOT: slotid == G_MAXUINT\n");

			taglist_get_string_tag   ( usertaglist, TAGNAME_FILENAME_SHORT, &username );

			if (!username)
				username = "(no name)";

			/* uploads_change_upload_slot() takes ownership of filehash and userhash! */
			uploads_change_upload_slot ( slotid, filename, filehash, username, userhash );

			G_FREE(junk);
			G_FREE(filename);
		}
		break;

		case CORE_SHARED_FILES:
		{
			parse_message_read_shared_files(packet);
		}
		break;

		case CORE_SHARED_DIRS:
		{
			guint i, num;

			shared_dirs_set_remove_all ();

			num = read_16();

			for (i=0; i<num; i++)
			{
				gchar	*dirname = read_i4str();

				/* TODO: we could avoid some mem alloc if shared_... took ownership of dirname */
				shared_dirs_add_or_update_record (dirname, i + 1);

				G_FREE(dirname);
			}

			shared_dirs_remove_old ();

		}
		break;

		case CORE_OPTIONS:
		{
			parse_message_receive_options();
			options_update_option_widgets();
		}
		break;

		case CORE_DOWNLOAD_GAP_DETAILS:
		{
			progressbarGap *pbgaps = NULL;
			guint32         id = 0;
			guint16         gaps, i;
			guint8          hash[16];

			if (!CORE_SENDS_DL_ID_INSTEAD_OF_HASH_WITH_GAPS)
			{
				ramfile_read(ramfile_now_active,hash,16);
			}
			else
			{
				id = read_32();
			}

			gaps = read_16();
			g_return_if_fail ( gaps != 0xffff );

			/* better make sure this is a reasonable number */
			if ( gaps > 2000 )
				break;

			/* allocate NULL-terminated array of progressbarGap structures */
			pbgaps = g_new0 (progressbarGap, (gaps+1) );
			g_return_if_fail ( pbgaps != NULL );

			for (i=0; i < gaps; i++)
			{
				pbgaps[i].start  = read_32();
				pbgaps[i].end    = read_32();
				pbgaps[i].status = read_16();
			}

			/* note: downloads_set_download_gaplist takes ownership of pbgaps. Don't free it */
			if ( id == 0 )
			{
				if ( downloads_set_download_gaplist (hash, pbgaps) == FALSE )
					g_free(pbgaps);
			}
			else
			{
				if ( downloads_set_download_gaplist_from_id (id, pbgaps) == FALSE )
					g_free(pbgaps);
			}

		}
		break;

		case CORE_CLIENT_STATS:
		{
			parse_message_read_core_client_stats(packet);
		}
		break;

		default:
		{
			ramfile_dump (packet, _(" Received unknown message from core "));
		}
	}
}
