/* pipe.c for ed2k_gui
*
* creates and checks the named pipe opened for receiving ed2k:// links
*
* (c) 2001-2002 Tim-Philipp Muller <t.i.m@orange.net>
*/


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "global.h"
#include "downloads.h"
#include "pipe.h"
#include "http_get.h"
#include "send_command.h"
#include "options.h"
#include "status_page.h"
#include "statusbar.h"
#include "misc.h"
#include "misc_strings.h"

#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>	// XXX - what's this for???
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#define PIPE_BUF_SIZE   256U  /* # of bytes to increase the buffer by each iteration */

static gchar        *pipename = NULL;
static GIOChannel   *ioc = NULL;
static guint         ioc_handler = 0;

static gboolean      pipe_iochannel_callback (GIOChannel *ioc, GIOCondition status, gpointer data);



/******************************************************************************
 *
 *   pipe_create
 *
 *   create a named pipe to receive ed2k-links etc.
 *
 *   returns TRUE if okay, or FALSE if failed
 *
 ***/

gboolean
pipe_create (void)
{
	gint	pipe_flags;
	gint	npipe;

	if (!pipename)
	{
		pipename = g_strdup(opt_get_opt_filename("pipe"));
	}

	g_return_val_if_fail (pipename!=NULL, FALSE);

	/* if pipe exists, delete and create new one to make sure it's a fifo and not a normal file! */
	if ( unlink (pipename) != 0  &&  errno!= ENOENT )
	{
		const gchar *msg1 = _("GUI: couldn't remove and re-create named pipe to listen for ed2k:// links\n");
		status_msg ("%s", msg1);
		status_msg (_("GUI: (possibly owned by other user?) well, it might work anyway...\n"));
		statusbar_extra_message (" %s", msg1);
		return FALSE;
	}

	if (mkfifo(pipename, 0700) < 0)
	{
		status_msg (_("GUI: Could not create named pipe '%s' - %s\n"), pipename, g_strerror(errno));
		return FALSE;
	} /* else status_msg ("GUI: Okay, created named pipe\n"); */

#if !defined(__FreeBSD__)
	pipe_flags = O_RDONLY | O_NONBLOCK;
#else
	pipe_flags = O_RDONLY;	/* solves bug #656923 apparently */
#endif

	if ((npipe = open (pipename, pipe_flags)) < 0)
	{
		status_msg (_("GUI: Could not create named pipe '%s' - %s\n"), pipename, g_strerror(errno));
		return FALSE;
	}

	if ((ioc = g_io_channel_unix_new (npipe)))
	{
		static gboolean	 showed_pipefilename = FALSE;
		GIOStatus        ret;
		gchar           *msg;

		ret = g_io_channel_set_encoding (ioc, NULL, NULL);
		g_return_val_if_fail ( ret == G_IO_STATUS_NORMAL, FALSE );

		g_io_channel_set_buffered (ioc, FALSE);

		g_io_channel_set_close_on_unref (ioc, FALSE);

//		ret = g_io_channel_set_flags (ioc, , NULL);
//		g_return_val_if_fail ( ret == G_IO_STATUS_NORMAL, FALSE );

		ioc_handler = g_io_add_watch (ioc, G_IO_IN|G_IO_PRI|G_IO_ERR|G_IO_HUP|G_IO_NVAL, pipe_iochannel_callback, NULL);
		g_io_channel_unref (ioc);	/* io_channel_add_watch() ads a reference */

		if (!showed_pipefilename)
		{
			msg = g_strdup_printf (_("GUI: Listening on named pipe '%s' for ed2k:// links.\n"), pipename);
			g_return_val_if_fail (msg!=NULL, FALSE);
			status_msg (msg);
			g_free(msg);
			showed_pipefilename = TRUE;	/* we don't want this message after every received link */
		}
	} else g_printerr (_("Could not set up GIOChannel on named pipe for ed2k:// links.\n"));
	return (ioc!=NULL);
}


/******************************************************************************
 *
 *   pipe_parse_message
 *
 *   parses message sent to pipe or message dropped into main window via
 *      drag'n'drop
 *
 ***/

void
pipe_parse_message (gchar *msg)
{
	gchar	*start, *msg_utf8;

	g_return_if_fail (msg!=NULL);

	/* replace '%xx' unescapes by real characters */
	misc_strings_unescape_url (msg);

	/* Check if it's an ed2k://|command|<advanced command>| link
	 *
	 * XXX - This is 'exploitable' in theory - people could set up
	 *	html-refresh with an ed2k://|command| link - thus the 'q'
	 *	and the 'm # X' commands are not allowed
	 */

	if ((start = strstr (msg, "|ipchanged|")))
	{
		downloads_pause_and_resume_all_active_downloads ();
		status_msg (_("received IP CHANGED indicator\n"));
		return;
	}

	if ((start = strstr (msg, "|command|")))
	{
		gchar *end = strrchr (msg, '|');
		start+=9;
		if (end>start)
		{
			end[0]=0x00;
			while (*start==' ' && start<end) start++;	/* skip initial spaces */
			if (strncmp(start,"m ",2)!=0 && start[0]!='q')
			{
				send_command (start);
				status_msg (_("received ed2k:// command link ('%s')\n"), start);
			} else status_msg (_("received ed2k:// command link with not permitted command ('m x y' or 'q')!\n"));
		} else status_msg (_("received invalid ed2k:// command link.\n"));
		return;
	}

	msg_utf8 = TO_UTF8(msg);
	if (extract_ed2k_links (msg_utf8) <= 0 )	/* ... and don't clear search results list when extracting the links */
	{
		if (extract_server_links(msg, FALSE) > 0)	/* if not a file link, maybe server links? */
		{
			send_get_serverlist();
			status_msg (_("received ed2k:// server link(s)\n"));
		} else g_printerr (_("received something not an ed2k:// link (via pipe or drag'n'drop)\n"));
	} else status_msg (_("received ed2k:// file link\n"));

	FREE_UTF8(msg_utf8);
}

/******************************************************************************
 *
 *   pipe_unlink
 *
 ***/

void
pipe_unlink (void)
{
	if (ioc)
	{
		GError *err = NULL;

// XXX - we SHOULD unref the channel, so it's freed, but if we do, we get segfaults.
//		g_io_channel_unref (ioc);	/* initial ref count of 1 on io channel creation */
		g_io_channel_shutdown (ioc, FALSE, &err);

		if (err)
		{
			g_print ("Error in g_io_channel_shutdown()/%s: %s\n", __FUNCTION__, err->message);
			g_error_free(err);
		}
	}

	if (ioc_handler>0)
		g_source_remove (ioc_handler);

	ioc = NULL;
	ioc_handler = 0;

	if (pipename)
		unlink (pipename);
}


/******************************************************************************
 *
 *   pipe_iochannel_callback
 *
 ***/

static gboolean
pipe_iochannel_callback (GIOChannel *ioc, GIOCondition status, gpointer data)
{
	static GIOStatus  ret = G_IO_STATUS_NORMAL;
	static guint      len = 999, size = 0;
	static gchar     *buf = NULL;

	g_return_val_if_fail (ioc!=NULL, FALSE);
/*
	g_print ("ioc = %p --- status = %u --- data = %p\n", ioc, status, data);
	g_print ("\tG_IO_IN   = %u\n", (status & G_IO_IN ));
	g_print ("\tG_IO_PRI  = %u\n", (status & G_IO_PRI ));
	g_print ("\tG_IO_OUT  = %u\n", (status & G_IO_OUT ));
	g_print ("\tG_IO_NVAL = %u\n", (status & G_IO_NVAL ));
	g_print ("\tG_IO_HUP  = %u\n", (status & G_IO_HUP ));
*/
	if ( (status&G_IO_IN) == G_IO_IN || (status&G_IO_PRI) == G_IO_PRI )
	{
		GError *err = NULL;

		while (ret == G_IO_STATUS_NORMAL && len > 0 && err == NULL)
		{
			buf = g_renew (gchar, buf, size+PIPE_BUF_SIZE);
			ret = g_io_channel_read_chars (ioc, buf+size, PIPE_BUF_SIZE, &len, &err);
			size += len;
		}

		// Do not return TRUE here, because status is bit_wise and it could
		//	be G_IO_IN|G_IO_HUP, for instance.
	}

	/* G_IO_OUT should never happen on a read-only pipe */
	g_return_val_if_fail ( (status&G_IO_OUT) == 0, FALSE);


	if (  ((status&G_IO_ERR) == G_IO_ERR || (status&G_IO_HUP) == G_IO_HUP)
		  ||  (((status&G_IO_IN)==G_IO_IN || (status&G_IO_PRI)==G_IO_PRI) && (len==0 || ret == G_IO_STATUS_EOF) ))
	{
		pipe_unlink();

		pipe_create();

		if (buf)
		{
			pipe_parse_message(buf);
			g_free (buf);
			len = 999;
			size = 0;
			ret = G_IO_STATUS_NORMAL;
			buf = NULL;
		}

		return FALSE;	/* remove event source */
	}

	if ((status&G_IO_NVAL) == G_IO_NVAL)
	{
		g_printerr (_("Named pipe '%s' couldn't be set up (G_IO_NVAL).\n"), pipename);
		pipe_unlink();
		return FALSE;	/* remove event source */
	}

	return TRUE;
}

