/*
*   send_command.c     (part of the eDonkey2000 linux GUI)
*
*   Functions to send commands to the core/command line client
*
*   (c) 2001 Tim-Philipp Mueller <t.i.m@orange.net>
*
*/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "connect_gui.h"
#include "global.h"
#include "gui_core_protocol.h"
#include "gui_flags.h"
#include "guicoreconn.h"
#include "misc.h"
#include "options.h"
#include "packet.h"
#include "parse_message.h"
#include "ramfile.h"
#include "send_command.h"
#include "shared_page.h"
#include "status_page.h"
#include "tag.h"

#include <time.h>

// local functions
static void	send_message_without_data (guint8 msg);
static void	send_server_command (guint8 cmd, guint32 ip, guint16 port);
static void	send_simple_search (guint8 searchcmd, const gchar *searchstr);
static void	send_command_with_hash (guint8 cmd, const guint8 *hash);


// send_message_without_data
//
// Sends a message that only has a message byte and no data

static void
send_message_without_data (guint8 msg)
{
	ramfile	*packet = ramfile_new(1);
//	g_print ("%s - msg=%u\n", __FUNCTION__, msg);
	ramfile_write8 (packet,msg);
	guicoreconn_send_packet (packet);
}

/* login / logout ************************************************************/


// send_login
//
// Send login (username/password)

void
send_login (const gchar *user, const gchar *pass)
{
	ramfile	*packet = ramfile_new(1+2+1+2+1);
	g_return_if_fail (user!=NULL);
	g_return_if_fail (pass!=NULL);
	g_return_if_fail (packet!=NULL);

	ramfile_write8 (packet, CONT_LOGIN);
	ramfile_write_i4_string (packet, user);
	ramfile_write_i4_string (packet, pass);
	guicoreconn_send_packet (packet);
}

// send_logout
//
// Shutdown core

void
send_logout (void)
{
	send_message_without_data (CONT_STOP);
	guicoreconn_status = GUI_CORE_CONN_STATUS_CLOSED_DELIBERATELY;
}


/* servers *******************************************************************/

void
send_get_serverlist (void)
{
	send_message_without_data (GUI_GET_SERVERLIST);
}

void
send_disconnect (void)
{
	send_message_without_data (GUI_DISCONNECT);
}

// send_server_command
//
// Most server commands are commonly structured (ie. CMD - IP - Port), so why
//	not simplify this in one function..

static void
send_server_command (guint8 cmd, guint32 ip, guint16 port)
{
	ramfile	*packet = ramfile_new (1+4+2);
	g_return_if_fail (packet!=NULL);
	ramfile_write8  (packet, cmd);
	ramfile_writeIP (packet, ip);
	ramfile_write16 (packet, port);
	guicoreconn_send_packet (packet);
}

// send_connect_to_server
//
// Connect to a specific server (0/0 = any)

void
send_connect_to_server (guint32 ip, guint16 port)
{
	send_server_command (GUI_CONNECT, ip, port);
}

// send_add_server
//
// Add a server to the server list

void
send_add_server (guint32 ip, guint16 port)
{
	send_server_command (GUI_ADD_SERVER, ip, port);
}


// send_remove_server
//
// Remove a server from the server list

void
send_remove_server (guint32 ip, guint16 port)
{
	send_server_command (GUI_DELETE_SERVER, ip, port);
}


// send_set_server_priority
//
// Sets the priority of a certain server

void
send_set_server_priority (guint32 ip, guint16 port, ServerPriority priority)
{
	ramfile	*packet = ramfile_new (1+4+2+1);
	g_return_if_fail (packet!=NULL);
	ramfile_write8  (packet, GUI_SETPRI_SERVER);
	ramfile_writeIP (packet, ip);
	ramfile_write16 (packet, port);
	ramfile_write8  (packet, (guint8)priority);
	guicoreconn_send_packet (packet);
}


/* search ******************************************************************/

// send_simple_search
//
// Search for only one (untyped) string (can be multiple words?) or user name
// searchcmd: GUI_SEARCH or GUI_SEARCH_USER

static void
send_simple_search (guint8 searchcmd, const gchar *searchstr)
{
	ramfile	*packet;

#warning *** FIX overnet search support
	g_return_if_fail (IS_OVERNET==FALSE && IS_HYBRID==FALSE);
	
	packet = ramfile_new (1+1+2+0);
	g_return_if_fail (packet!=NULL);
	ramfile_write8  (packet, searchcmd);	// GUI_SEARCH or GUI_SEARCH_USER
	ramfile_write8  (packet, 0x01);		// simple search. 0x01 = FIND_STR search node type
	ramfile_write_i4_string (packet, searchstr);
	guicoreconn_send_packet (packet);
}


// send_search
//
// Search for string w/o file type etc.

void
send_search (const gchar *str)
{

#warning *** FIX overnet search support
	if ( IS_OVERNET || IS_HYBRID )
	{
		gchar *s = g_strdup_printf ("s %s", str);
		send_command(s);
		g_free(s);
		return;
	}

	send_simple_search(GUI_SEARCH, str);
}


// send_search_with_extra_options
//
// Search for simple string w/ extra options

void
send_search_with_extra_options (const gchar *str, guint32 i_minsize, guint32 i_maxsize, guint32 i_minavail)
{
	ramfile	*packet = ramfile_new (2+2+9+9+2+1+9);
	g_return_if_fail (packet!=NULL);

#warning *** FIX overnet search support
	if ( IS_OVERNET == TRUE  &&  IS_HYBRID == FALSE )
	{
		status_message_blue ("GUI: Overnet + extra search options not possible with this version of the GUI.\n");
		status_message_blue ("GUI: (using normal simple search)\n");
		send_search(str);
		ramfile_free(packet);
		return;
	}

	ramfile_write8  (packet, GUI_SEARCH);

	// left node first level (2 bytes)
	ramfile_write8  (packet, 0x00);		// compound search node
	ramfile_write8  (packet, 0x00);		// AND operation

		// left node second level (2 bytes)
		ramfile_write8  (packet, 0x00);		// compound search node
		ramfile_write8  (packet, 0x00);		// AND operation

			// left node third level (9 bytes)
			ramfile_write8  (packet, 0x03);								// match int
			ramfile_write32 (packet, i_maxsize);						// maximum size
			ramfile_write8  (packet, 0x04);								// less or equal
			ramfile_write_i4_string (packet, TAGNAME_FILESIZE_SHORT);	// in filesize tag

			// right node third level (9 bytes)
			ramfile_write8  (packet, 0x03);								// match int
			ramfile_write32 (packet, i_minsize);						// minimum size
			ramfile_write8  (packet, 0x03);								// greater or equal
			ramfile_write_i4_string (packet, TAGNAME_FILESIZE_SHORT);	// in filesize tag

		// right node second level (2 bytes)
		ramfile_write8  (packet, 0x00);		// compound search node
		ramfile_write8  (packet, 0x00);		// AND operation

			// left node third level (>=1 bytes)
			if ((str) && (*str))
			{
				ramfile_write8  (packet, 0x01);					// simple string search
				ramfile_write_i4_string (packet, str);
			} else ramfile_write8  (packet, 0x04);				// match all if no string given

			// right node third level (9 bytes)
			ramfile_write8 (packet, 0x03);									// match int
			ramfile_write32 (packet, i_minavail);							// min avail
			ramfile_write8 (packet, 0x03);									// greater or equal
			ramfile_write_i4_string (packet, TAGNAME_AVAILABILITY_SHORT);	// in availability tag

	guicoreconn_send_packet (packet);
}

// send_typed_search_with_extra_options
//
// Search for string w/ extra options + file type

void
send_type_search_with_extra_options (const gchar *str, const gchar *type, guint32 i_minsize, guint32 i_maxsize, guint32 i_minavail, guint32 i_minbitrate)
{
	ramfile	*packet = ramfile_new (2+2+2+9+9+2+1+9);
	g_return_if_fail (packet!=NULL);
	ramfile_write8  (packet, GUI_SEARCH);

#warning *** FIX overnet search support
	if ( IS_OVERNET == TRUE  &&  IS_HYBRID == FALSE )
	{
		status_message_blue ("GUI: Overnet + extra search options/type not possible with this version of the GUI.\n");
		status_message_blue ("GUI: (using normal simple search)\n");
		send_search(str);
		ramfile_free(packet);
		return;
	}

	// first node -> compound node (2 bytes)
	ramfile_write8  (packet, 0x00);		// compound search node
	ramfile_write8  (packet, 0x00);		// AND operation

		// left node second level (2 bytes)
		ramfile_write8  (packet, 0x00);		// compound search node
		ramfile_write8  (packet, 0x00);		// AND operation

			// left node third level (2 bytes)
			ramfile_write8  (packet, 0x00);		// compound search node
			ramfile_write8  (packet, 0x00);		// AND operation

				// left node 4th level (9 bytes)
				ramfile_write8  (packet, 0x03);								// match int
				ramfile_write32 (packet, i_maxsize);						// maximum size
				ramfile_write8  (packet, 0x04);								// less or equal
				ramfile_write_i4_string (packet, TAGNAME_FILESIZE_SHORT);	// in filesize tag

				// right node 4th  level (9 bytes)
				ramfile_write8  (packet, 0x03);								// match int
				ramfile_write32 (packet, i_minsize);						// minimum size
				ramfile_write8  (packet, 0x03);								// greater or equal
				ramfile_write_i4_string (packet, TAGNAME_FILESIZE_SHORT);	// in filesize tag

			// right node third level (2 bytes)
			if ((type) && (g_ascii_strncasecmp(type,"any",3)!=0))
			{
				ramfile_write8 (packet, 0x02);					// find string in tag
				ramfile_write_i4_string (packet, type);
				ramfile_write_i4_string (packet, TAGNAME_FILETYPE_SHORT);
			} else ramfile_write8 (packet, 0x04);	// match any if no type specified

		// right node second level (2 bytes)
		ramfile_write8  (packet, 0x00);		// compound search node
		ramfile_write8  (packet, 0x00);		// AND operation

			// left node third level (>=1 bytes)
			if ((str) && (*str))
			{
				ramfile_write8  (packet, 0x01);					// simple string search
				ramfile_write_i4_string (packet, str);
			} else ramfile_write8  (packet, 0x04);				// match all if no string given

			// right node third level (2 bytes)
			ramfile_write8  (packet, 0x00);		// compound search node
			ramfile_write8  (packet, 0x00);		// AND operation

				// left node 4th level (9 bytes)
				ramfile_write8 (packet, 0x03);									// match int
				ramfile_write32 (packet, i_minavail);							// min avail
				ramfile_write8 (packet, 0x03);									// greater or equal
				ramfile_write_i4_string (packet, TAGNAME_AVAILABILITY_SHORT);	// in availability tag

				// right node 4th level (only match bitrate if we have type 'audio')
				if ((type) && (g_ascii_strncasecmp(type,"audio",5)==0))
				{
					ramfile_write8  (packet, 0x03);								// match int
					ramfile_write32 (packet, i_minbitrate);						// minimum bitrate
					ramfile_write8  (packet, 0x03);								// greater or equal
					ramfile_write_i4_string (packet, "bitrate");				// in bitrate tag
				} else ramfile_write8 (packet, 0x04);	// match any if no type specified
				
	guicoreconn_send_packet (packet);
}


// send_user_search
//
// Search for a user by name

void
send_search_user (const gchar *str)
{
	send_simple_search(GUI_SEARCH_USER, str);
}


// send_type_search
//
// Search for files with filetype
//

void
send_type_search (const gchar *str, const gchar *type, guint32 i_minbitrate)
{
	ramfile		*packet = ramfile_new (1+ 3+2+0+ 4+2+0);
	gboolean	 audiosearch;

	g_return_if_fail (packet!=NULL);

#warning *** FIX overnet search support
	if ( IS_OVERNET == TRUE  &&  IS_HYBRID == FALSE )
	{
		status_message_blue ("GUI: Overnet + type search not possible with this version of the GUI.\n");
		status_message_blue ("GUI: (using normal simple search)\n");
		send_search(str);
		ramfile_free(packet);
		return;
	}


	audiosearch = ((type) && (*type) && g_ascii_strcasecmp(type,"audio")==0);
	ramfile_write8  (packet, GUI_SEARCH);

	if (audiosearch)
	{
		ramfile_write8  (packet, 0x00);		// compound
		ramfile_write8  (packet, 0x00);		// AND
	} 

	if ((type) && (*type))
	{
		ramfile_write8  (packet, 0x00);		// compound
		ramfile_write8  (packet, 0x00);		// AND

		ramfile_write8  (packet, 0x02);		// left node
		ramfile_write_i4_string (packet, type);
		ramfile_write_i4_string (packet, TAGNAME_FILETYPE_SHORT);
	}

	if ((str) && (*str))
	{
		ramfile_write8  (packet, 0x01);		// right node
		ramfile_write_i4_string (packet, str);
	} else ramfile_write8 (packet, 0x04);	// match any

	if (audiosearch)
	{
		ramfile_write8  (packet, 0x03);								// match int
		ramfile_write32 (packet, i_minbitrate);						// minimum bitrate
		ramfile_write8  (packet, 0x03);								// greater or equal
		ramfile_write_i4_string (packet, "bitrate");				// in bitrate tag
	}

	guicoreconn_send_packet (packet);
}

void
send_extended_search (void)
{
	send_message_without_data (GUI_EXSEARCH);
}

void
send_extended_search_user (void)
{
	send_message_without_data (GUI_EXSEARCH_USER);
}

// send_more_search
//
// Get the next 200 results

void
send_more_search (void)
{
	send_message_without_data (GUI_MORESEARCH);
}


/* downloads ******************************************************************/

// send_command_with_hash
//
// Sends a command that is followed by a hash only

static void
send_command_with_hash (guint8 cmd, const guint8 *hash)
{
	ramfile	*packet = ramfile_new (1+16);
	g_return_if_fail (packet!=NULL);
	g_return_if_fail (hash!=NULL);
	ramfile_write8 (packet, cmd);
	ramfile_write (packet, hash, 16);
	guicoreconn_send_packet (packet);
}

// send_download
//
// Start new download from given file hash that is WITHIN SEARCH RESULTS

void
send_download (const guint8 *hash)
{
	send_command_with_hash (GUI_DOWNLOAD, hash);
}

// send_new_download
//
// Download from ed2k-link ie NOT FROM THE SEARCH RESULTS

void
send_new_download (const guint8 *hash, guint32 size, const gchar *fn)
{
	ramfile	*packet = ramfile_new (1+  16+4+2+  4+  1+2+1+4+  1+2+1+2+0);
	g_return_if_fail (packet!=NULL);
	g_return_if_fail (hash!=NULL);
	ramfile_write8 (packet, GUI_NEW_DOWNLOAD);
	// Now, we send a FileMeta record, in principle...
	//	(a) hash + ip + port
	ramfile_write (packet, hash, 16);
	ramfile_writeIP (packet, 0);	// unused
	ramfile_write16 (packet, 0);	// unused
	//	(b) metatags
	ramfile_write32 (packet, 2);	// 2 tags

	//	Tag 1 - filesize
	ramfile_write8  (packet, 0x03);	// Tag type 3 = integer tag
	ramfile_write16 (packet, 1);	// Tag name len = 1 => special tag (abbreviation)
	ramfile_write8  (packet, 0x02);	// Tag name = 0x02 => filesize
	ramfile_write32 (packet, size);	// filesize integer value

	//	Tag 2 - filename
	ramfile_write8  (packet, 0x02);		// Tag type 2 = string tag
	ramfile_write16 (packet, 1);		// Tag name len = 1 => special tag (abbreviation)
	ramfile_write8  (packet, 0x01);		// Tag name = 0x01 => file name
	ramfile_write_i4_string (packet, fn);	// file name string
	guicoreconn_send_packet (packet);
}


void
send_start_sending_dl_status (void)
{
	send_message_without_data (GUI_START_DL_STATUS);
}

void
send_stop_sending_dl_status (void)
{
	send_message_without_data (GUI_STOP_DL_STATUS);
}


void
send_set_download_priority (const guint8 *hash, guint8 priority)
{
	ramfile	*packet = ramfile_new (1+16+1);
	g_return_if_fail (packet!=NULL);
	g_return_if_fail (hash!=NULL);
	ramfile_write8 (packet, GUI_SETPRI_DOWNLOAD);
	ramfile_write (packet, hash, 16);
	ramfile_write8 (packet, priority);
	guicoreconn_send_packet (packet);
}

void
send_pause_download (const guint8 *hash)
{
	send_command_with_hash (GUI_PAUSE_DOWNLOAD, hash);
	// send get download gaps after that, so the pixmap is refreshed
	// (if the download is paused, we wont' update the gaplist, so the pixmap
	// is never redrawn in 'paused mode' with greyish gaps instead of red ones
	if (CORE_IS_GAP_ENABLED)
		send_get_download_gaps (hash);
}


void
send_resume_download (const guint8 *hash)
{
	send_command_with_hash (GUI_RESUME_DOWNLOAD, hash);
	// send get download gaps after that, so the pixmap is refreshed
	// (if the download is paused, we wont' update the gaplist, so the pixmap
	// is never redrawn in 'paused mode' with greyish gaps instead of red ones
	if (CORE_IS_GAP_ENABLED) send_get_download_gaps (hash);
}

void
send_cancel_download (const guint8 *hash)
{
	send_command_with_hash (GUI_CANCEL_DOWNLOAD, hash);
}

void
send_get_download_gaps (const guint8 *hash)
{
	ramfile	*packet = ramfile_new (1+16+1);
	g_return_if_fail (packet!=NULL);
	g_return_if_fail (hash!=NULL);
	ramfile_write8 (packet, GUI_GET_DOWNLOAD_GAP_DETAILS);
	ramfile_write (packet, hash, 16);
	ramfile_write8 (packet, 0); // flag for future purposes (e.g. send additional source infos)
	guicoreconn_send_packet (packet);
}

// ------ uploads -----

void
send_start_sending_ul_status (void)
{
	send_message_without_data (GUI_START_UL_STATUS);
}

void
send_stop_sending_ul_status (void)
{
	send_message_without_data (GUI_STOP_UL_STATUS);
}

/* shared stuff */

void
send_get_shared_dirs (void)
{
	send_message_without_data (GUI_GET_SHARE_DIRS);
//  gtk_clist_clear  (GTK_CLIST (clist_shared));
}

void
send_get_shared_files (void)
{
	send_message_without_data (GUI_GET_SHARE_FILES);
}

void
send_unshare_dir_no (guint dir_no)
{
	gchar	*buf = g_strdup_printf ("u %u", dir_no);
	g_return_if_fail (buf!=NULL);
	send_command (buf);
	g_free (buf);
}

void
send_add_shared_dir (const gchar *dir)
{
	gchar	*buf;
	g_return_if_fail (dir!=NULL);
	buf = g_strdup_printf ("a %s", dir);
	g_return_if_fail (buf!=NULL);
	send_command (buf);
	g_free (buf);
}

void
send_add_shared_dir_structure (const gchar *dir)
{
	gchar	*buf;
	g_return_if_fail (dir!=NULL);
	buf = g_strdup_printf ("a+ %s", dir);
	g_return_if_fail (buf!=NULL);
	send_command (buf);
	g_free (buf);
}


/* other */

// send_command
//
// send 'advanced' command, ie. pass on command in command line client syntax to CL controller

void
send_command (const gchar *cmdstr)
{
	ramfile	*packet = ramfile_new(1+2+0);
	g_return_if_fail (cmdstr!=NULL);
	ramfile_write8 (packet, CONT_CMD);
	ramfile_write_i4_string (packet, cmdstr);
	guicoreconn_send_packet (packet);
}

/* options */

void
send_get_options (void)
{
	send_message_without_data(GUI_GET_OPTIONS);
}

void
send_set_port (guint16 port)
{
	gchar	*buf;
	buf = g_strdup_printf ("port %u", port);
	g_return_if_fail (buf!=NULL);
	send_command (buf);
	g_free (buf);
}

void
send_set_admin_port (guint16 aport)
{
	gchar	*buf;
	buf = g_strdup_printf ("aport %u", aport);
	g_return_if_fail (buf!=NULL);
	send_command (buf);
	g_free (buf);
}


void
send_set_max_speeds (guint dlmax, guint ulmax)
{
	gchar	*buf;
	buf = g_strdup_printf ("dumax %u %u", dlmax, ulmax);
	g_return_if_fail (buf!=NULL);
	send_command (buf);
	g_free (buf);
}

void
send_set_max_connections (guint maxcon)
{
	gchar	*buf;
	buf = g_strdup_printf ("mcon %u", maxcon);
	g_return_if_fail (buf!=NULL);
	send_command (buf);
	g_free (buf);
}


void
send_auto_connect (gboolean autoconnectON)
{
	if (autoconnectON)
	{
		send_command ("auto");
	} else send_command ("auto -");
}

void
send_save_corrupted (gboolean savecorruptedON)
{
	if (savecorruptedON)
	{
		send_command ("save");
	} else send_command ("save -");
}


void
send_auto_server_remove (gboolean autoserverremoveON)
{
	if (autoserverremoveON)
	{
		send_command ("asr");
	} else send_command ("asr -");
}


void
send_verbose (gboolean verboseON)
{
	if (verboseON)
	{
		send_command ("verbose");
	} else send_command ("verbose -");
}

void
send_set_nickname (const gchar *nick)
{
	gchar	*buf;
	g_return_if_fail (nick!=NULL);
	buf = g_strdup_printf ("name %s", nick);
	g_return_if_fail (buf!=NULL);
	send_command (buf);
	g_free (buf);
}


void
send_set_incoming (const gchar *dir)
{
	gchar	*buf;
	g_return_if_fail (dir!=NULL);
	buf = g_strdup_printf ("in %s", dir);
	g_return_if_fail (buf!=NULL);
	send_command (buf);
	g_free (buf);
}


void
send_set_temp (const gchar *dir)
{
	gchar	*buf;
	g_return_if_fail (dir!=NULL);
	buf = g_strdup_printf ("temp %s", dir);
	g_return_if_fail (buf!=NULL);
	send_command (buf);
	g_free (buf);
}


// stats

void
send_get_peer_stats (void)
{
	if ( !CORE_IS_ALIVE )
		return;

	if (core_build_date >= CORE_BUILD_1)
	{
		send_message_without_data (GUI_GET_CLIENT_STATS);
	}
	else
	{
		g_printerr (_("%s: please get a newer core.\n"), __FUNCTION__);
	}
}


void
send_get_space_stats (void)
{
	if ( !CORE_IS_ALIVE )
		return;

	if (core_build_date >= CORE_BUILD_1)
	{
		send_message_without_data (GUI_GET_CLIENT_STATS);
	}
	else
	{
		g_printerr (_("%s: please get a newer core.\n"), __FUNCTION__);
	}
}


void
send_get_client_stats (void)
{
	if ( !CORE_IS_ALIVE )
		return;

	if (core_build_date >= CORE_BUILD_1)
	{
		send_message_without_data (GUI_GET_CLIENT_STATS);
	}
	else
	{
		g_printerr (_("%s: please get a newer core.\n"), __FUNCTION__);
	}
}



/******************************************************************************
 *
 *   send_set_upload_priority
 *
 ***/

void
send_set_upload_priority (const guint8 *hash, guint newprio)
{
	ramfile	*packet;

	g_return_if_fail ( hash != NULL );
	g_return_if_fail ( newprio >= 0 && newprio <= 2 );

	/* core must be newer than Feb 21, 2003 */
	if (core_build_date < 0x030215)	
	{
		g_printerr (_("This core version does not support setting the upload priority..\n"));
		return;
	}

	packet = ramfile_new (1+16+1);

	g_return_if_fail (packet!=NULL);

	ramfile_write8 (packet, GUI_SET_UPLOAD_PRIORITY);
	ramfile_write  (packet, hash, 16);
	ramfile_write8 (packet, newprio);

	guicoreconn_send_packet (packet);
}


/******************************************************************************
 *
 *   send_get_overnet_contacts
 *
 ***/

void
send_get_overnet_contacts (void)
{
	static time_t lastasked;       /* 0 */
	static guint  evil_happenings; /* 0 */

	g_return_if_fail ( IS_OVERNET == TRUE  ||  IS_HYBRID == TRUE );

	if ( lastasked > 0 && ++evil_happenings > 5 )
		g_return_if_fail( (time(NULL)-lastasked) >= 45 );

	send_command ("vc");

	lastasked = time(NULL);
}





