/***************************************************************************
                          shared_dirs.c  -  description
                             -------------------
    begin                : Sat Jul 27 2002
    copyright            : (C) 2002 by Tim-Philipp Mller
    email                : t.i.m@orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "global.h"
#include "guicoreconn.h"
#include "options.h"
#include "misc_gtk.h"
#include "misc_strings.h"
#include "send_command.h"
#include "statusbar.h"
#include "shared_dirs.h"
#include "shared_page.h"
#include "status_page.h"

enum
{
	SHARED_DIRS_ICON_COLUMN,
	SHARED_DIRS_NAME_COLUMN,
	SHARED_DIRS_REMOVE_COLUMN,
	SHARED_DIRS_IDXNUM_COLUMN,
	SHARED_DIRS_UNSHARE_FLAG_COLUMN,
	SHARED_DIRS_ORIGINAL_STRING_COLUMN,
	SHARED_DIRS_N_COLUMNS
};

#define SHARED_DIRS_TYPES_PLACEHOLDER \
	GDK_TYPE_PIXBUF,\
	G_TYPE_STRING,\
	G_TYPE_BOOLEAN,\
	G_TYPE_INT,\
	G_TYPE_BOOLEAN, \
	G_TYPE_STRING


static guint          shared_dirs_count = 0;

static GtkWidget     *shared_dirs_view = NULL;

static GtkListStore  *shared_dirs_store = NULL;

static gchar         *shared_dirs_fileselector_lastdir = NULL;

/* dirsht: non-utf8 dirname as key, GtkTreeRowReferences as values */
static GHashTable    *dirsht = NULL;



/* functions */

static gint     helper_send_remove_records_sort_func (gconstpointer p1, gconstpointer p2);

static void     shared_dirs_send_request (guint millisecs);

static gint     shared_dirs_send_request_timeout (gpointer data);

static void     shared_dirs_add_dir (const gchar *path, int add_whole_structure);

static gboolean check_if_path_is_allowed (const gchar *addpath);

static void     shared_dirs_menu_onRefresh (GtkWidget *widget, gpointer data);

static void     shared_dirs_menu_onUnshareSelectedDirectoryTrees (GtkWidget *widget, gpointer data);

static void     shared_dirs_menu_onUnshareSelectedDirectories (GtkWidget *widget, gpointer data);

static void     shared_dirs_menu_onAddDirectoryTreeToShare (GtkWidget *widget, gpointer data);

static void     shared_dirs_popup_menu (guint button, guint32 activate_time);

static gint     shared_dirs_onButtonPress (GtkWidget *widget, GdkEventButton *event, gpointer data);

static gint     shared_dirs_onPopupMenu (GtkWidget *widget, gpointer data);



/******************************************************************************
 *
 *    shared_dirs_get_count
 *
 ***/

guint
shared_dirs_get_count (void)
{
	return shared_dirs_count;
}



/******************************************************************************
 *
 *   shared_dirs_clear_list
 *
 ***/

void
shared_dirs_clear_list (void)
{
	if (shared_dirs_store)
		gtk_list_store_clear (shared_dirs_store);

	if (dirsht)
		g_hash_table_destroy(dirsht);

	dirsht = g_hash_table_new_full ( g_str_hash, g_str_equal, g_free, (GDestroyNotify) gtk_tree_row_reference_free );
}



/******************************************************************************
 *
 *   shared_dirs_add_or_update_record
 *
 *   Inserts a record for DIRNAME into the store. If record already exists,
 *     IDXNUM is updated.
 *
 ***/

void
shared_dirs_add_or_update_record (const gchar *dirname, gint idxnum)
{
	GtkTreeRowReference    *ref;
	GtkTreeIter             iter;
	const gchar            *name_utf8;

	g_return_if_fail ( dirname != NULL );
	g_return_if_fail ( dirsht  != NULL );

	ref = g_hash_table_lookup (dirsht, dirname);

	/* if we already have this in our list, just update the values (index num etc.) */

	if ( ref != NULL )
	{
		if (misc_gtk_tree_model_get_iter_from_row_reference ( GTK_TREE_MODEL(shared_dirs_store), ref, &iter ))
		{
			gtk_list_store_set ( shared_dirs_store, &iter,
			                     SHARED_DIRS_IDXNUM_COLUMN, idxnum,
			                     SHARED_DIRS_REMOVE_COLUMN, FALSE,
			                     SHARED_DIRS_UNSHARE_FLAG_COLUMN, FALSE,
			                     -1);
			return;
		}
		g_return_if_reached();
	}

	/* if we don't have it yet, add it */

	gtk_list_store_append (shared_dirs_store, &iter);

	ref = misc_gtk_tree_model_get_row_reference_from_iter (GTK_TREE_MODEL(shared_dirs_store), &iter);

	g_return_if_fail ( ref != NULL );

	name_utf8 = TO_CONST_UTF8_USE_TABLE(dirname);

	g_return_if_fail ( name_utf8 != NULL );

	gtk_list_store_set ( shared_dirs_store, &iter,
	                     SHARED_DIRS_ICON_COLUMN, get_icon (ICON_FILETYPE_FOLDER),
	                     SHARED_DIRS_NAME_COLUMN, name_utf8,
	                     SHARED_DIRS_IDXNUM_COLUMN, idxnum,
	                     SHARED_DIRS_REMOVE_COLUMN, FALSE,
	                     SHARED_DIRS_UNSHARE_FLAG_COLUMN, FALSE,
	                     SHARED_DIRS_ORIGINAL_STRING_COLUMN, dirname,
	                     -1 );

	g_hash_table_insert (dirsht, g_strdup(dirname), ref);

	shared_dirs_count++;
}



/******************************************************************************
 *
 *   shared_dirs_set_remove_all
 *
 *   Flags all dirs in the shared dirs list for removal. Called from
 *    parse_message when a new list arrives from the core.
 *
 ***/

void
shared_dirs_set_remove_all (void)
{
	GtkTreeIter     iter;
	gboolean        val;

	val = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(shared_dirs_store), &iter);

	while (val)
	{
		gtk_list_store_set ( shared_dirs_store, &iter, SHARED_DIRS_REMOVE_COLUMN, TRUE, -1);

		val = gtk_tree_model_iter_next (GTK_TREE_MODEL(shared_dirs_store), &iter);
	}
}



/******************************************************************************
 *
 *   shared_dirs_remove_old
 *
 *   Removes entries flagged for removal from the store.
 *
 ***/
static gboolean
shared_dirs_remove_old_htcallback (gpointer key, GtkTreeRowReference *ref, gpointer userdata)
{
	GtkTreeIter     iter;
	gboolean        remove;

	if (misc_gtk_tree_model_get_iter_from_row_reference ( GTK_TREE_MODEL(shared_dirs_store), ref, &iter))
	{
			gtk_tree_model_get ( GTK_TREE_MODEL(shared_dirs_store), &iter,
			                     SHARED_DIRS_REMOVE_COLUMN, &remove,
			                     -1);

			if ( remove == TRUE )
			{
				shared_dirs_count--;

				gtk_list_store_remove (shared_dirs_store, &iter);

				return TRUE;
			}

			return FALSE;
	}

	g_return_val_if_reached(FALSE);
}


void
shared_dirs_remove_old (void)
{
	g_return_if_fail ( dirsht != NULL );

	g_hash_table_foreach_remove (dirsht, (GHRFunc) shared_dirs_remove_old_htcallback, NULL);
}




/******************************************************************************
 *
 *  helper_send_remove_records_sort_func
 *
 *   callback for g_slist_sort() to sort a linked list of shareddirRecords
 *   in reverse order of their index numbers (highest numbers first)
 *
 ***/

static gint
helper_send_remove_records_sort_func (gconstpointer p1, gconstpointer p2)
{
	guint n1, n2;

	g_return_val_if_fail ( p1 != NULL, 0 );
	g_return_val_if_fail ( p2 != NULL, 0 );

	n1 = GPOINTER_TO_UINT(p1);
	n2 = GPOINTER_TO_UINT(p2); 

	if (n1>n2)
	{
		return (-1);
	}
	return ((gint) (n2>n1));
}


/******************************************************************************
 *
 *   shared_dirs_unshare_all_flagged
 *
 ***/

static void
shared_dirs_unshare_all_flagged (void)
{
	GSList      *recordlist = NULL;
	GSList      *node;
	GtkTreeIter  iter;
	gboolean     val;

	/* get indices of all directories that are flagged for unsharing */
	val = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(shared_dirs_store), &iter);
	while (val)
	{
		gint      dirindex = -1;
		gboolean  flag = FALSE;

		gtk_tree_model_get ( GTK_TREE_MODEL(shared_dirs_store), &iter,
		                     SHARED_DIRS_IDXNUM_COLUMN, &dirindex,
		                     SHARED_DIRS_UNSHARE_FLAG_COLUMN, &flag,
		                     -1);

		g_return_if_fail (dirindex > 0);	/* core starts counting from 1 here */

		if (flag)
		{
			recordlist = g_slist_insert_sorted (recordlist, GUINT_TO_POINTER(dirindex), helper_send_remove_records_sort_func);
		}

		val = gtk_tree_model_iter_next (GTK_TREE_MODEL(shared_dirs_store), &iter);
	}

	/* now that we have a sorted list with directory indices (highest first), unshare */
	node = recordlist;
	while (node)
	{
		guint	num = GPOINTER_TO_UINT(node->data);

		g_return_if_fail ( num > 0);

		status_message_blue (_("GUI: unsharing directory #%u\n"), num);
		send_unshare_dir_no(num);

		node = node->next;
	}

	g_slist_free(recordlist);

	shared_dirs_send_request (2000);
}


/******************************************************************************
 *
 * shareddirs_send_request 
 *
 * updates the list of shared dirs in X millisecs
 *
 ***/

static gint
shared_dirs_send_request_timeout (gpointer data)
{
	send_get_shared_dirs ();
	return 0; /* don't call us again */
}

static void
shared_dirs_send_request (guint millisecs)
{
	if (millisecs == 0)
	{
		send_get_shared_dirs ();
		return;
	}
	
	(void) g_timeout_add (millisecs, shared_dirs_send_request_timeout, NULL);
}


/******************************************************************************
 *
 *   check_if_path_is_allowed
 *
 *   displays warning message if path is not allowed.
 *
 *   returns 0 if path is not allowed, otherwise 1
 *
 ***/

static gboolean
check_if_path_is_allowed (const gchar *addpath)
{
	const gchar *msg = NULL;

	g_return_val_if_fail ( addpath  != NULL, FALSE );

	/* path entered was '/' (because tailing '/' has been chopped off) */
	if ( !(*addpath) )
	{
		msg = _("I don't think this is a sensible path to share. Sorry.");
	}
	else if (     strncmp(addpath, "/dev",     4) == 0 
	           || strncmp(addpath, "/etc",     4) == 0  
	           || strncmp(addpath, "/sbin",    5) == 0  
	           || strncmp(addpath, "/bin",     4) == 0  
	           || strncmp(addpath, "/boot",    5) == 0  
	           || strncmp(addpath, "/proc",    5) == 0
	           || strncmp(addpath, "/usr/bin", 8) == 0  
	           || strncmp(addpath, "/usr/sbin",9) == 0  
	           || strcmp (addpath, "/home")       == 0 )
	{
		msg = _("I don't think this is a sensible path to share. Sorry.");
	}

	/* absolute path? */
	else if ( CORE_IS_LOCAL  &&  !g_path_is_absolute(addpath) )
	{
		msg = _("Please specify an absolute path (e.g. starting with '/')");
	}

	/* home path shared? -> caution! */
	else if ( strcmp (addpath, g_get_home_dir()) == 0 )
	{
		msg = _("You are trying to share your home "
		        "directory or your _whole_ home directory structure.\n"
		        "This is a security risk, as the donkey will automatically "
		        "include all hidden files and hidden subdirectories,\nwhich "
		        "might contain secret information like usernames and "
		        "passwords (think about the configuration file of\n your "
		        "mail client, or your .ssh subdirectory, or your .gnupg "
		        "subdirectory if you are using pgp public keys...).\n"
		        "However, you can share specific folders from your home "
		        "directory if you enter something like "
		        "\"/home/joe/sharethis\"");
	}

	else if ( CORE_IS_LOCAL  &&  !g_file_test(addpath, G_FILE_TEST_EXISTS) )
	{
		msg = _("The path you entered does not exist.");
	}

	else if ( CORE_IS_LOCAL  &&  !g_file_test(addpath, G_FILE_TEST_IS_DIR) )
	{
		msg = _("The path you entered is not a directory.");
	}

	if ( !msg )
		return TRUE;

	misc_gtk_ok_dialog (GTK_MESSAGE_WARNING, msg);

	return FALSE ;
}



/******************************************************************************
 *
 *   shared_dirs_add_dir
 *
 *   path has previously been returned from fileselector
 *     and is in on-disk encoding
 *
 ***/

static void
shared_dirs_add_dir (const gchar *path, int add_whole_structure) // auxiliary to file select dialg
{
	gchar   *addpath;

	if (!path)
		return;

	addpath = g_strdup(path);

	/* chop of tailing '/' if there */
	if ( addpath[strlen(addpath)-1] == G_DIR_SEPARATOR )
		addpath[strlen(addpath)-1] = 0x00;

	if ( !check_if_path_is_allowed(addpath) )
	{
		g_free(addpath);
		return;
	}

	if ( add_whole_structure )
	{
		status_message_blue (_("Adding directory tree to shared files. If this does not work, it's the core's fault (really)."));
		send_add_shared_dir_structure (addpath);
	}
	else
	{
		send_add_shared_dir (addpath);
	}

	shared_dirs_send_request (3000);

	statusbar_extra_message (_(" The core is hashing the newly shared files..."));

	g_free(addpath);

	/* note: filelist will update automatically, because we'll get a
	 * Done loading shared file message" when the new files have been hashed */
}



/******************************************************************************
 *
 *   helper_flag_all_subdirs_for_unsharing
 *
 ***/
 
static void
helper_flag_all_subdirs_for_unsharing (const gchar *dirname)
{
	GtkTreeIter  iter;
	gint         parentdir_len;
	gboolean     val = FALSE;

	g_return_if_fail (dirname != NULL);

	parentdir_len = strlen(dirname);
	
	val = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(shared_dirs_store), &iter);

	while (val) 
	{
		gchar     *tmpname = NULL;
		gboolean   flag = FALSE;

		gtk_tree_model_get ( GTK_TREE_MODEL(shared_dirs_store), &iter,
		                     SHARED_DIRS_NAME_COLUMN, &tmpname,
		                     SHARED_DIRS_UNSHARE_FLAG_COLUMN, &flag,
		                     -1);

		if ((!flag) && (tmpname))
		{
			gint checkdir_len = strlen (tmpname);

			/* subdir path must have at least 2 chars more than parent dir (slash + char) */
			if (checkdir_len >= (parentdir_len+2))
			{
				/* subdir must have slash where parent dir has terminating zero */
				if (tmpname[parentdir_len] == G_DIR_SEPARATOR)
				{
					/* subdir and parent dir must match at the beginning */
					if (strncmp(tmpname,dirname,parentdir_len)==0)
					{
						/* it's a subdir => flag it for unsharing */
						gtk_list_store_set ( shared_dirs_store, &iter,
						                     SHARED_DIRS_UNSHARE_FLAG_COLUMN, TRUE,
						                     -1);

						/* note: recursion here is not necessary, because
						 * /mnt/foo/blub/one/two/three is still a subdir of
						 * /mnt/foo/blub the way we check for it, so we do
						 * not need to find subdirs of /mnt/foo/blub/one here.
						 */
					}
				}
			}
		}

		G_FREE(tmpname);
	
		val = gtk_tree_model_iter_next (GTK_TREE_MODEL(shared_dirs_store), &iter);

	}
} 





/*----------------------------------------------------------------
 |								  |
 |								  |
 |                            POPUP MENU			  |
 |								  |
 |								  |
 ----------------------------------------------------------------
*/

static void
shared_dirs_fileselector_tree_onOk (GtkWidget *w, GtkFileSelection *fs)
{

	if (shared_dirs_fileselector_lastdir)
		
		g_free (shared_dirs_fileselector_lastdir);

	shared_dirs_fileselector_lastdir =  g_strdup (gtk_file_selection_get_filename (GTK_FILE_SELECTION(fs)));

	shared_dirs_add_dir (shared_dirs_fileselector_lastdir, 1);

	gtk_widget_destroy (GTK_WIDGET(fs));


}

static void
shared_dirs_fileselector_dir_onOk (GtkWidget *w, GtkFileSelection *fs)
{

	  if (shared_dirs_fileselector_lastdir)

                g_free (shared_dirs_fileselector_lastdir);

        shared_dirs_fileselector_lastdir =  g_strdup (gtk_file_selection_get_filename (GTK_FILE_SELECTION(fs)));

        shared_dirs_add_dir (shared_dirs_fileselector_lastdir, 0);

	gtk_widget_destroy (GTK_WIDGET(fs));


}


static void
shared_dirs_fileselector_onCancel (GtkWidget *widget, GtkWidget *dialog)
{
        if (shared_dirs_fileselector_lastdir) 
		
		g_free(shared_dirs_fileselector_lastdir);

        shared_dirs_fileselector_lastdir = g_strdup(gtk_file_selection_get_filename (GTK_FILE_SELECTION(dialog)));
        
	gtk_widget_destroy (dialog);

}

static void
shared_dirs_fileselector_dialog (int add_structure_flag)
{
        GtkWidget       *filew;
        gchar           *title = g_strdup_printf (_(" Select a directory%s to share "),
                                                        (add_structure_flag==0) ? "" : _(" tree"));

        g_return_if_fail (title!=NULL);
        filew = gtk_file_selection_new (title);

        if (!shared_dirs_fileselector_lastdir) 

		shared_dirs_fileselector_lastdir = g_strdup_printf("%s%c", g_get_home_dir(), G_DIR_SEPARATOR);

        gtk_file_selection_set_filename (GTK_FILE_SELECTION(filew), shared_dirs_fileselector_lastdir);

        // cancel = destroy dialogue widget
        
	g_signal_connect (G_OBJECT(GTK_FILE_SELECTION(filew)->cancel_button), "clicked", G_CALLBACK(shared_dirs_fileselector_onCancel),
				filew);
				
        
	// ok button
        
	g_signal_connect (G_OBJECT(GTK_FILE_SELECTION(filew)->ok_button), "clicked",
                (add_structure_flag==0) ? G_CALLBACK(shared_dirs_fileselector_dir_onOk) : 
					  G_CALLBACK(shared_dirs_fileselector_tree_onOk), filew);


	// from the gtk-app-devel-list, 25-03-99, by Kemal 'Disq' Hadimili - thank you. */
        gtk_widget_hide(GTK_FILE_SELECTION(filew)->file_list);
        gtk_widget_hide(GTK_FILE_SELECTION(filew)->file_list->parent);

        gtk_file_selection_hide_fileop_buttons (GTK_FILE_SELECTION(filew));
        gtk_widget_show(filew);
        g_free(title);
}


static void
shared_dirs_menu_onAddDirectoryToShare (GtkWidget *widget, gpointer data)
{
	if (CORE_IS_LOCAL)
	{
		shared_dirs_fileselector_dialog (0);
	}
	else
	{
		misc_gtk_ok_dialog (GTK_MESSAGE_INFO,
			_(" The core seems to be running on a \n"
			  " different computer than this one. \n"
			  " This means I am not able to show \n"
			  " you a directory selector dialog. \n"
			  " \n"
			  " Please type this into the toolbar: \n"
			  "   !a <full directory path>\n\n"));
	}
}



static void
shared_dirs_menu_onAddDirectoryTreeToShare (GtkWidget *widget, gpointer data)
{
	if (CORE_IS_LOCAL)
	{
		shared_dirs_fileselector_dialog (1);
	}
	else
	{
		misc_gtk_ok_dialog (GTK_MESSAGE_INFO,
			_(" The core seems to be running on a \n"
			  " different computer than this one. \n"
			  " This means I am not able to show \n"
			  " you a directory selector dialog. \n"
			  " \n"
			  " Please type this into the toolbar: \n"
			  "   !a+ <full directory path>\n\n"));
	}
}



/******************************************************************************
 *
 *   for_each_helper_flag_selected_dirs_for_unsharing
 *
 ***/

static void
for_each_helper_flag_selected_dirs_for_unsharing (GtkTreeModel *model,
                                                  GtkTreePath *path,
                                                  GtkTreeIter *iter,
                                                  gpointer data)
{
	g_return_if_fail (iter  != NULL);
	g_return_if_fail (model != NULL);

	gtk_list_store_set ( GTK_LIST_STORE(model), iter,
                         SHARED_DIRS_UNSHARE_FLAG_COLUMN, TRUE,
                         -1);
}

/******************************************************************************
 *
 *   for_each_helper_flag_selected_dir_trees_for_unsharing
 *
 ***/

static void
for_each_helper_flag_selected_dir_trees_for_unsharing (GtkTreeModel *model,
                                                       GtkTreePath *path,
                                                       GtkTreeIter *iter,
                                                       gpointer data)
{
	gchar *name = NULL;

	g_return_if_fail (iter != NULL);
	g_return_if_fail (model != NULL);

	gtk_tree_model_get ( GTK_TREE_MODEL(model), iter,
                         SHARED_DIRS_NAME_COLUMN, &name,
                         -1);
	if (name)
	{
		helper_flag_all_subdirs_for_unsharing (name);
		g_free(name);
	}

	/* don't only flag subdirs, but selected dir as well */
	gtk_list_store_set ( GTK_LIST_STORE(model), iter,
                         SHARED_DIRS_UNSHARE_FLAG_COLUMN, TRUE,
                         -1);
}

/******************************************************************************
 *
 *   shared_dirs_menu_onUnshareSelectedDirectories 
 *
 ***/

static void
shared_dirs_menu_onUnshareSelectedDirectories (GtkWidget *widget, gpointer data)
{
	GtkTreeSelection *selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(shared_dirs_view));

	g_return_if_fail (misc_gtk_tree_selection_count_selected_rows (selection) > 0);

	gtk_tree_selection_selected_foreach ( selection,
	                                      for_each_helper_flag_selected_dirs_for_unsharing,
	                                      NULL);

	shared_dirs_unshare_all_flagged();

	statusbar_extra_message (_(" Unshared selected directories."));
}


static void
shared_dirs_menu_onUnshareSelectedDirectoryTrees (GtkWidget *widget, gpointer data)
{
	GtkTreeSelection *selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(shared_dirs_view));

	g_return_if_fail (misc_gtk_tree_selection_count_selected_rows (selection) > 0);

	gtk_tree_selection_selected_foreach ( selection,
	                                      for_each_helper_flag_selected_dir_trees_for_unsharing,
	                                      NULL);

	shared_dirs_unshare_all_flagged();

	statusbar_extra_message (_(" Unshared selected directories and sub-directories. (list update might take a while)"));
}

static void
shared_dirs_menu_onRefresh (GtkWidget *widget, gpointer data)
{
	shared_dirs_send_request (0);
	send_get_shared_files ();
}


void
shared_dirs_menu_onRehashSharedFiles (GtkWidget *widget, gpointer data)
{
	static gboolean working = FALSE;
	time_t          timefrom = time(NULL);

	if (working)
		return;

	working = TRUE;

	status_message_blue ("%s\n", _(" Forcing core to rehash shared files. If you see a nonsense error message, ignore it."));

	/* [Tim] XXX - what is this for??? */
	while ( (gtk_events_pending()) && (time(NULL)-timefrom) < 2 )
		gtk_main_iteration_do(FALSE);

	working = FALSE;

	/* old core versions: recheck all dirs by adding bogus directory                              */
	/* new core versions: add directory that is already shared to
	 *                    make it update the shared files
	 */
	if (core_build_date >= CORE_JANUAR_2003)
	{
		GtkTreeIter   iter;
		gboolean      val;
		gchar        *dirname_original = NULL;

		g_return_if_fail (shared_dirs_store != NULL);

        val = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(shared_dirs_store), &iter);

		g_return_if_fail (val == TRUE);

		gtk_tree_model_get ( GTK_TREE_MODEL(shared_dirs_store), &iter,
	                         SHARED_DIRS_ORIGINAL_STRING_COLUMN, &dirname_original,
	                         -1);

		g_return_if_fail (dirname_original != NULL);

		send_add_shared_dir (dirname_original);

		g_free(dirname_original);

		return;
	}

	send_add_shared_dir (_("/ignore_this_error_message_it_is_to_force_the_core_to_rehash"));
}



/******************************************************************************
 *
 *   shared_dirs_popup_menu
 *
 ***/

static void
shared_dirs_popup_menu (guint button, guint32 activate_time)
{
	GtkTreeSelection    *selection;
	GtkWidget           *menu;
	guint                sel_count;

	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(shared_dirs_view));
	sel_count = misc_gtk_tree_selection_count_selected_rows (selection);

	menu  = gtk_menu_new();

	misc_gtk_add_menu_item ( menu, _(" add directory to share... "),
	                         shared_dirs_menu_onAddDirectoryToShare, NULL, ICON_NONE, TRUE);

	misc_gtk_add_menu_item ( menu, _(" add directory tree to share... "),
	                         shared_dirs_menu_onAddDirectoryTreeToShare, NULL, ICON_NONE, TRUE);

	misc_gtk_add_menu_separator (menu);

	misc_gtk_add_menu_item ( menu, _(" unshare selected directories "),
	                         shared_dirs_menu_onUnshareSelectedDirectories, NULL, ICON_NONE, (sel_count>0) );

	misc_gtk_add_menu_item ( menu, _(" unshare selected directory trees "),
	                         shared_dirs_menu_onUnshareSelectedDirectoryTrees, NULL, ICON_NONE, (sel_count>0) );

	misc_gtk_add_menu_separator (menu);

	misc_gtk_add_menu_item ( menu, _(" refresh list "),
	                         shared_dirs_menu_onRefresh, NULL, ICON_NONE, TRUE );

	misc_gtk_add_menu_separator (menu);

	misc_gtk_add_menu_item ( menu, _(" rehash shared files "),
	                         shared_dirs_menu_onRehashSharedFiles, NULL, ICON_NONE, TRUE );


	gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL, button, activate_time);
}


/******************************************************************************
 *
 *   shared_dirs_onPopupMenu
 *
 ***/

static gint
shared_dirs_onPopupMenu (GtkWidget *widget, gpointer data)
{
	shared_dirs_popup_menu (0, gdk_event_get_time(NULL));
	return TRUE;
}

/******************************************************************************
 *
 *   shared_dirs_onButtonPress
 *
 ***/

static gint
shared_dirs_onButtonPress (GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	if ( event->button != 3 )
		return FALSE;

	g_return_val_if_fail (CORE_IS_ALIVE == TRUE, FALSE);

	shared_dirs_popup_menu (event->button, event->time);
	return TRUE;
}


/******************************************************************************
 *
 *   shared_dirs_create_view 
 *
 *   Creates the GtkTreeView inside a scrolled window.
 *
 ***/

GtkWidget *
shared_dirs_create_view (void)
{
	GtkWidget           *scrollwin;
	GtkTreeViewColumn   *column;
	GtkCellRenderer     *renderer;
	GtkTreeSelection    *selection;

	scrollwin = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scrollwin), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

	shared_dirs_store = gtk_list_store_new ( SHARED_DIRS_N_COLUMNS,
	                                         SHARED_DIRS_TYPES_PLACEHOLDER );

	shared_dirs_view = gtk_tree_view_new_with_model (GTK_TREE_MODEL(shared_dirs_store));
	
	column = gtk_tree_view_column_new();
	gtk_tree_view_column_set_sort_column_id(column, SHARED_DIRS_NAME_COLUMN);

	renderer = gtk_cell_renderer_pixbuf_new();
	gtk_tree_view_column_pack_start(GTK_TREE_VIEW_COLUMN(column), renderer, FALSE);
	gtk_tree_view_column_set_attributes(GTK_TREE_VIEW_COLUMN(column), renderer,
	                                    "pixbuf", SHARED_DIRS_ICON_COLUMN, NULL);
	gtk_tree_view_column_set_title(column, UTF8_SHORT_PRINTF("%s",_(" directory")));

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start (GTK_TREE_VIEW_COLUMN(column), renderer, FALSE);
	gtk_tree_view_column_set_attributes(GTK_TREE_VIEW_COLUMN(column), renderer,
	                                    "text", SHARED_DIRS_NAME_COLUMN, NULL);
	gtk_tree_view_column_set_sort_column_id(column, SHARED_DIRS_NAME_COLUMN);
	gtk_tree_view_append_column(GTK_TREE_VIEW(shared_dirs_view), column);
	
	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(shared_dirs_view));
	gtk_tree_selection_set_mode (selection, GTK_SELECTION_MULTIPLE);

	g_signal_connect (G_OBJECT(shared_dirs_view), "button_press_event", G_CALLBACK(shared_dirs_onButtonPress), NULL);
	g_signal_connect (G_OBJECT(shared_dirs_view), "popup_menu", G_CALLBACK(shared_dirs_onPopupMenu), NULL);

	gtk_container_add (GTK_CONTAINER(scrollwin), shared_dirs_view);

	gtk_widget_show_all (scrollwin);

	/* make sure variables are set to NULL when objects are destroyed */

	g_object_set_data_full (G_OBJECT(shared_dirs_store), "foo1",
	                        &shared_dirs_store, (GDestroyNotify)g_nullify_pointer);

	g_object_set_data_full (G_OBJECT(shared_dirs_view), "foo1",
	                        &shared_dirs_view, (GDestroyNotify)g_nullify_pointer);

	/* make sure this string is freed when view is destroyed (in case string is not empty) */

	g_object_set_data_full (G_OBJECT(shared_dirs_view), "foo2",
	                        shared_dirs_fileselector_lastdir, (GDestroyNotify)g_free);

	dirsht = g_hash_table_new_full ( g_str_hash, g_str_equal, g_free, (GDestroyNotify) gtk_tree_row_reference_free );

	return scrollwin;
}



