/***************************************************************************
              status_page.c  -  GTK2 version of the status page
              -------------------------------------------------
    begin                : Sat Feb 1 2003
    copyright            : (C) 2003 by Tim-Philipp Mller
    email                : t.i.m@orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#define _GNU_SOURCE
#include <stdio.h>		/* for vasprintf() declaration */

#include "colors.h"
#include "global.h"
#include "logfile.h"
#include "mainwindow.h"
#include "misc.h"
#include "misc_strings.h"
#include "options.h"
#include "statusbar.h"
#include "status_page.h"
#include <string.h>
#include <errno.h>

/* local variables */

static GtkWidget    	*view = NULL;		/* status text view widget */
static GtkTextBuffer	*viewbuf = NULL;	/* status text buffer      */

/* local functions */

static void              status_message_coloured (GdkColor *colour, const gchar *msg);

static GtkTextTag       *status_page_lookup_colour_tag (GdkColor *color);

static void              status_select_and_copy_ed2klinks_in_status_text (void);


/******************************************************************************
 *
 * gui_print_error_function
 *
 * This is our new error print handler (for g_printerr())
 *
 * If the window and the status text widget exist, it will output the error
 *	to stderr and the status log (and thus also the status log file)
 * If the window or the status text widget are not initialised yet, output will
 *	go to stderr only
 *
 */

void
gui_print_error_function (const gchar *errstring)
{
	gchar *newline = "";
	g_return_if_fail (errstring!=NULL);
	g_return_if_fail (*errstring!=0x00);

	if (errstring[strlen(errstring)-1] != '\n')
		newline = "\n";

	/* This will also output the line to the logfile */
	status_warning (_("GUI: %s%s"), errstring, newline);

	/* Might contain markup escapes */
	if ( strchr(errstring,'&') != NULL )
	{
		gchar *unesc;
		unesc = misc_strings_unescape_text(errstring);
		if (unesc)
		{
			fprintf(stderr, _("GUI: %s%s"), unesc, newline);
			g_free(unesc);
		}
	}
	else
		fprintf (stderr, _("GUI: %s%s"), errstring, newline);
}


/******************************************************************************
 *
 *   status_select_and_copy_ed2klinks_in_status_text
 *
 *   selects all (consecutive) ed2k-links at the end of the textbuf
 *   and copies them to the clipboard
 *
 ***/

static void
status_select_and_copy_ed2klinks_in_status_text (void)
{
	GtkClipboard  *clipboard;
	GtkTextIter    end, firstlink, emptyiter, previous;
	gint           l, lines;

	g_return_if_fail ( viewbuf != NULL );

	memset(&firstlink, 0x00, sizeof(GtkTextIter));
	memset(&emptyiter, 0x00, sizeof(GtkTextIter));

	lines = gtk_text_buffer_get_line_count(GTK_TEXT_BUFFER(viewbuf));

	gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER(viewbuf), &end);

	memcpy(&previous, &end, sizeof(GtkTextIter));

	for ( l = lines-2; l >= 0; l--)
	{
		GtkTextIter   linestart;
		gchar        *txt;

		gtk_text_buffer_get_iter_at_line (GTK_TEXT_BUFFER(viewbuf), &linestart, l);

		txt = gtk_text_buffer_get_text(GTK_TEXT_BUFFER(viewbuf), &linestart, &previous, FALSE);

		memcpy(&previous, &linestart, sizeof(GtkTextIter));

		if (!txt)
			break;

		/* these are all ascii chars, so utf-8 must match ascii */
		if ( g_ascii_strncasecmp(txt,"ed2k:",5) != 0 )
		{
			g_free(txt);
			break;
		}

		memcpy(&firstlink, &linestart, sizeof(GtkTextIter));

		g_free(txt);
	}

	/* note: we've just printed a link, so there must be at least
	 * one line with an ed2k-link in the buffer                    */

	g_return_if_fail ( memcmp(&firstlink, &emptyiter, sizeof(GtkTextIter)) !=0 );

	gtk_text_buffer_move_mark_by_name (GTK_TEXT_BUFFER(viewbuf), "insert", &firstlink);
	gtk_text_buffer_move_mark_by_name (GTK_TEXT_BUFFER(viewbuf), "selection_bound", &end);

	clipboard = gtk_clipboard_get (gdk_atom_intern("CLIPBOARD",TRUE));
	gtk_text_buffer_copy_clipboard (GTK_TEXT_BUFFER(viewbuf), clipboard);

}



/******************************************************************************
 *
 * status_msg_print_ed2klink
 *
 * Just prints an ed2k://|file| link to the status log
 * Makes sure that the filename doesn't contain illegal characters.
 *
 ***/

void
status_msg_print_ed2klink (const gchar *filename, guint size, const guint8* hash)
{
	gchar *fn, *name_locale, *name;

	g_return_if_fail ( filename != NULL );
	g_return_if_fail ( hash     != NULL );

	fn = misc_strings_unescape_text(filename);	/* just in case */

	g_return_if_fail ( fn != NULL );

	g_strdelimit(fn, "|\\/$~\"&^", ' ');	/* do NOT delimit '%' here */

	name = TO_LOCALE(fn);

	g_return_if_fail ( name != NULL );

	name_locale = misc_strings_escape_ascii_url (name);

	g_return_if_fail ( name_locale != NULL );

	status_msg ("ed2k://|file|%s|%u|%s|/\n", name_locale, size, hash_to_hash_str(hash));

	if (opt_get_bool(OPT_GUI_NO_LINKS_TO_LOGFILE))
	{
		statusbar_extra_message (_(" ed2k-link of this file written to status window and copied to the clipboard."));
	}
	else
	{
		statusbar_extra_message (_(" ed2k-link of this file written to logfile 'gui_statuslog' and to status log."));
	}

	status_select_and_copy_ed2klinks_in_status_text();

	g_free(fn);
	g_free(name);
	g_free(name_locale);
}


/******************************************************************************
 *
 * status_system_error_msg
 *
 * writes a status line in red to the statuslog, including the last error (errno must be set)
 * 'GUI:' is automatically prepended
 *
 */

void
status_system_error_msg (const gchar *msg_text)
{
	status_warning (_("GUI: %s (%s)\n"), (msg_text!=NULL) ? msg_text: "", g_strerror(errno));
}


/******************************************************************************
 *
 * status_parse_error_msg
 *
 * prints a 'GUI: parse error in file X line Y: <line beginning> ...' message
 *
 */

void
status_parse_error_msg (const gchar *infile, gint lineno, const gchar *linestring)
{
	g_return_if_fail (infile!=NULL);
	g_return_if_fail (linestring!=NULL);
	status_msg (_("GUI: parse error in %s file line %d: %20s...\n"), infile, lineno, linestring);
}



/******************************************************************************
 *
 *  status_page_lookup_colour_tag
 *
 */

static GtkTextTag *
status_page_lookup_colour_tag(GdkColor *color)
{
	if (color == &red)
	{
		return gtk_text_tag_table_lookup (gtk_text_buffer_get_tag_table(viewbuf), "RED_TAG");
	}
	else if (color == &blue)
	{
		return gtk_text_tag_table_lookup(gtk_text_buffer_get_tag_table(viewbuf), "BLUE_TAG");
	}
	else if ( color == &purple )
		return gtk_text_tag_table_lookup(gtk_text_buffer_get_tag_table(viewbuf), "PURPLE_TAG");

	return gtk_text_tag_table_lookup(gtk_text_buffer_get_tag_table(viewbuf), "BLACK_TAG");
}


/******************************************************************************
 *
 *   status_page_scroll_to_end
 *
 ***/

static gboolean
status_page_scroll_to_end (gpointer data)
{
	GtkTextIter iter;

	gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (viewbuf), &iter);
	gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW(view), &iter, 0, TRUE, 0.0, 1.0);

	return FALSE;
}

/******************************************************************************
 *
 *   status_message_coloured
 *
 *   Prints a message to the status window in a certain colour.
 *
 *   o will take care to suppress and/or parse certain messages
 *   o will write the message to the status log file if not suppressed.
 *
 */

static void
status_message_coloured (GdkColor *colour, const gchar *msg)
{
	const gchar *coltag = "";
	gchar       *new_msg = NULL;
	gchar       *msg_utf8 = NULL;
	GtkTextIter  end;

	g_return_if_fail ( colour != NULL);
	g_return_if_fail ( msg    != NULL);

	if (colour == &red)
		coltag = _("RED: ");

	if (colour == &blue)
		coltag = _("BLUE: ");

	logfile_write ("%s%s", coltag, msg);

	if ((!window) || (!view) || (!viewbuf))
		return;

	/* output time stamp */

	if ( opt_get_bool(OPT_GUI_LOG_HAVE_TIMESTAMP)
	      && *msg != '\n'
	      && g_ascii_strncasecmp(msg,"ed2k:",5) !=0 )
	{
		const gchar  *timestamp_utf8;
		GtkTextIter   enditer;

		timestamp_utf8 = UTF8_SHORT_PRINTF(logfile_get_timestamp());
		gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER(viewbuf), &enditer);
		gtk_text_buffer_insert_with_tags (GTK_TEXT_BUFFER(viewbuf), &enditer, timestamp_utf8, -1, status_page_lookup_colour_tag (&purple), NULL);
	}

	/* (strip off all potential newlines etc.) */
	new_msg = g_strdup(msg);
	g_return_if_fail (new_msg != NULL);
	(void) g_strdelimit(new_msg, "\n\t", ' ');
	(void) g_strstrip(new_msg);

	/* output to status window */
	msg_utf8 = TO_UTF8(msg);

	gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (viewbuf), &end);
	gtk_text_buffer_insert_with_tags (GTK_TEXT_BUFFER(viewbuf), &end, msg_utf8, -1, status_page_lookup_colour_tag (colour), NULL);

	g_idle_add ( (GSourceFunc) status_page_scroll_to_end, NULL );
	
	/* output to statusbar */
	statusbar_extra_message (" %s", new_msg);

	G_FREE(new_msg);
	FREE_UTF8(msg_utf8);
}

/******************************************************************************
 *
 *   status_msg
 *   status_warning
 *   status_message_blue
 *
 *   print messages to the status page (and to the status bar)
 *
 */

void
status_msg (const gchar *format, ...)
{
	gchar   *output;
	va_list  args;

	g_return_if_fail ( format != NULL );

	va_start(args, format);
	output = g_strdup_vprintf(format,args);
	va_end(args);

	status_message_coloured (&black, output);

	g_free(output);
}

void
status_warning (const gchar *format, ...)
{
	gchar   *output;
	va_list  args;

	g_return_if_fail ( format != NULL );

	va_start(args, format);
	output = g_strdup_vprintf(format,args);
	va_end(args);

	status_message_coloured (&red, output);

	g_free(output);
}

void
status_message_blue (const gchar *format, ...)
{
	gchar   *output;
	va_list  args;

	g_return_if_fail ( format != NULL );

	va_start(args, format);
	output = g_strdup_vprintf(format,args);
	va_end(args);

	status_message_coloured (&blue, output);

	g_free(output);
}



/******************************************************************************
 *
 *   status_page_clear
 *
 *   clears the status window
 *
 */

void
status_page_clear (void)
{
	static gboolean  license_shown = FALSE;
	GtkTextIter      startiter, enditer;
	gint             save_option_logtimestamp = opt_get_bool(OPT_GUI_LOG_HAVE_TIMESTAMP);

	if ((!window)||(!view)||(!viewbuf))
		return;

	opt_set_bool (OPT_GUI_LOG_HAVE_TIMESTAMP, FALSE);

	gtk_text_buffer_get_bounds(viewbuf, &startiter, &enditer);
	gtk_text_buffer_delete(viewbuf, &startiter, &enditer);

	status_warning (" \n                  ");
	gtk_text_buffer_get_end_iter(viewbuf, &enditer);
	gtk_text_buffer_insert_pixbuf(viewbuf, &enditer, get_icon(ICON_TOOLBAR_ED2K_LOGO));

	status_warning ("        ");
	gtk_text_buffer_get_end_iter(viewbuf, &enditer);
	gtk_text_buffer_insert_pixbuf(viewbuf, &enditer, get_icon(ICON_TOOLBAR_OVERNET_LOGO));

	status_warning (" \n \n");

	status_warning (_("Welcome to this %s GUI for eDonkey2000 and Overnet.\n"), OP_SYSTEM_STRING);
	status_warning ("ed2k-gtk-gui version %s, Copyright (c) 2001-2003 Tim-Philipp Mller\n", VERSION);

	if (!license_shown)
	{
		license_shown = TRUE;
		status_warning (" \n");		/* need the space so timestamps don't look funny */
		status_warning (_("This program is free software; you can redistribute it and/or modify\n"));
		status_warning (_("it under the terms of the GNU General Public License as published by\n"));
		status_warning (_("the Free Software Foundation; either version 2 of the License, or\n"));
		status_warning (_("(at your option) any later version.\n"));
		status_warning (" \n");		/* need the space so timestamps don't look funny */
		status_warning (_("This program is distributed in the hope that it will be useful, but\n"));
		status_warning (_("without any warranty; without even the implied warranty of merchantability\n"));
		status_warning (_("or fitness for a particular purpose. See the GNU General Public License\n"));
		status_warning (_("for more details: http://www.gnu.org\n"));
		status_warning (" \n");		/* need the space so timestamps don't look funny */
		status_warning (_("By using this program you agree to be bound by the GNU General Public License.\n"));
	}

	status_warning (" \n");		/* need the space so timestamps don't look funny */
	status_msg     (_("It is written in plain C using the Gimp Toolkit (GTK+) and the\n"));
	status_msg     (_("gnet library. Project homepage: http://ed2k-gtk-gui.sourceforge.net\n"));
	status_msg     (" \n");		/* so the statusbar is empty */
	status_message_blue ( _("Click on the eDonkey/Overnet logo in the top left corner to access extra functions.\n"));
	status_message_blue ( _("Type 'help' in the top toolbar to see what you can do with it and how to use it.\n"));
	status_msg          (" \n");		/* so the statusbar is empty */

	opt_set_bool (OPT_GUI_LOG_HAVE_TIMESTAMP, save_option_logtimestamp);
}


/******************************************************************************
 *
 *  status_page_print_clientID_info
 *
 *  prints the clientID and firewalled/not firewalled info to the statusbar
 *
 *  if idstring is NULL, it prints the last available ID info, otherwise
 *   it saves idstring as new ID info and does not print it out.
 *
 ***/

static gchar        *idinfo = NULL;

void
status_page_print_clientID_info (const gchar *idstring)
{
	if (idstring)
	{
		G_FREE(idinfo);
		idinfo = g_strdup(idstring);
		return;
	}

	if ( idinfo  &&  *idinfo != 0x00 )
		statusbar_extra_message(UTF8_SHORT_PRINTF("%s", idinfo));
	else
		statusbar_extra_message(UTF8_SHORT_PRINTF("%s",_("No clientID information available yet.")));

}


/******************************************************************************
 *
 *  status_page_create
 *
 *  create the status window notebook page and returns it.
 *
 ***/

GtkWidget *
status_page_create (void)
{
	GtkWidget   *frame;
	GtkWidget   *scrollwin;
	GtkTextTag  *tag;

	frame = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);

	/* create non-editable text widget with word-wrap and add it to packing table */
	view = gtk_text_view_new();
	gtk_text_view_set_editable (GTK_TEXT_VIEW(view), FALSE);
	gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW(view), FALSE);
	gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW(view), GTK_WRAP_WORD);

	viewbuf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
	gtk_text_buffer_set_text(viewbuf, " ", -1);
	g_object_set (G_OBJECT(view), "left-margin", 15, NULL);
	g_object_set (G_OBJECT(view), "right-margin", 15, NULL);

	/* create the tag table with some handy tags */

	tag = gtk_text_buffer_create_tag (GTK_TEXT_BUFFER (viewbuf), "RED_TAG",  "foreground-gdk", &red, NULL);
	tag = gtk_text_buffer_create_tag (GTK_TEXT_BUFFER (viewbuf), "BLUE_TAG", "foreground-gdk", &blue, NULL);
	tag = gtk_text_buffer_create_tag (GTK_TEXT_BUFFER (viewbuf), "BLACK_TAG", "foreground-gdk", &black, NULL);
	tag = gtk_text_buffer_create_tag (GTK_TEXT_BUFFER (viewbuf), "PURPLE_TAG", "foreground-gdk", &purple, "size-points", 7.0, NULL);

	g_object_set_data_full (G_OBJECT(view), "foo1",
	                        &view, (GDestroyNotify)g_nullify_pointer);

	g_object_set_data_full (G_OBJECT(viewbuf), "foo1",
	                        &viewbuf, (GDestroyNotify)g_nullify_pointer);


	status_page_clear();

	scrollwin = gtk_scrolled_window_new (NULL,NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scrollwin), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER(scrollwin), view);
	gtk_container_add (GTK_CONTAINER(frame), scrollwin);


	gtk_widget_show_all (frame);

	return frame;
}







