/***************************************************************************
                          uploads.c  -  description
                             -------------------
    begin                : Sat Feb 8 2003
    copyright            : (C) 2003 by Tim-Philipp Mller
    email                : t.i.m@orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "global.h"
#include "options.h"
#include "misc_gtk.h"
#include "misc_strings.h"
#include "shared_files.h"
#include "uploads.h"
#include "uploadstats.h"
#include "send_command.h"

enum
{
	UPLOADS_FILENAME_COLUMN = 0,
	UPLOADS_INDEX_COLUMN,
	UPLOADS_USERNAME_COLUMN,
	UPLOADS_SPEED_COLUMN,
	UPLOADS_FILEHASH_COLUMN,
	UPLOADS_USERHASH_COLUMN,
	UPLOADS_SPEED_STRING_COLUMN,
	UPLOADS_SLOTID_COLUMN,
	UPLOADS_N_COLUMNS
};

#define UPLOADS_TYPES_PLACEHOLDER \
	G_TYPE_STRING,\
	G_TYPE_INT,\
	G_TYPE_STRING,\
	G_TYPE_FLOAT,\
	G_TYPE_POINTER,\
	G_TYPE_POINTER,\
	G_TYPE_STRING,\
	G_TYPE_INT


static guint                 uploads_store_items = 0;
static GtkListStore         *uploads_store = NULL;
static GtkTreeModel         *uploads_store_sort = NULL;
static GtkWidget            *uploads_view = NULL;



static gboolean     uploads_record_is_complete (GtkTreeIter *iter, const guint8 **hash, gchar **name);

static gboolean     uploads_remove_iter (GtkTreeIter *iter);

static gboolean     uploads_get_iter_from_slot_id (guint32 slotid, GtkTreeIter *newiter);

/******************************************************************************
 *
 *  uploads_view_save_settings
 *
 *  saves uploads view settings: column widths and sort column
 *
 */

void
uploads_view_save_settings (void)
{
	misc_gtk_save_tree_view_settings ( uploads_view,
	                                   uploads_store,
	                                   OPT_GUI_COLUMN_WIDTHS_UPLOADS,
	                                   OPT_GUI_SORT_TYPE_UPLOADS,
	                                   OPT_GUI_SORT_COL_UPLOADS);
}


/******************************************************************************
 *
 *  uploads_view_restore_settings
 *
 *  Note: setting column widths only seems to work after the
 *        widget has been realised.
 *
 */

static void
uploads_view_restore_settings (void)
{
	misc_gtk_restore_tree_view_settings ( uploads_view,
	                                      uploads_store,
	                                      OPT_GUI_COLUMN_WIDTHS_UPLOADS,
	                                      OPT_GUI_SORT_TYPE_UPLOADS,
	                                      OPT_GUI_SORT_COL_UPLOADS,
	                                      UPLOADS_FILENAME_COLUMN);
}


/******************************************************************************
 *
 *  uploads_decrease_all_higher_indices
 *
 *  if an upload is removed, we need to move down the index
 *  numbers of all following uploads by one (as the core
 *  sends us upload status information by index number)
 *
 */

static void
uploads_decrease_all_higher_indices(guint index)
{
	gint         idx;
	gboolean     val;
	GtkTreeIter  iter;

	g_return_if_fail (uploads_store != NULL);

	val = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(uploads_store), &iter);
	while (val)
	{
		idx = -1;

		gtk_tree_model_get ( GTK_TREE_MODEL(uploads_store), &iter,
		                     UPLOADS_INDEX_COLUMN, &idx, -1);

		if (idx > index)
		{
			gtk_list_store_set ( uploads_store, &iter,
			                     UPLOADS_INDEX_COLUMN, (idx-1), -1);
		}

		val = gtk_tree_model_iter_next (GTK_TREE_MODEL(uploads_store), &iter);
	}
}


/******************************************************************************
 *
 *   uploads_get_iter_from_index
 *
 *
 */

static gboolean
uploads_get_iter_from_index (guint index, GtkTreeIter *newiter)
{
	GtkTreeIter  iter;
	gboolean     val;

	g_return_val_if_fail (uploads_store != NULL, FALSE);
	g_return_val_if_fail (newiter != NULL, FALSE);

	val = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(uploads_store), &iter);

	while (val)
	{
		gint idx = -1;
		gtk_tree_model_get ( GTK_TREE_MODEL(uploads_store), &iter,
		                     UPLOADS_INDEX_COLUMN, &idx, -1);
		if (idx == index)
		{
			*newiter = iter;
			return TRUE;
		}

		val = gtk_tree_model_iter_next (GTK_TREE_MODEL(uploads_store), &iter);
	}

	g_return_val_if_reached(FALSE);	/* the userhash should really have been in there */
}


/******************************************************************************
 *
 *   uploads_add_new_upload
 *
 *
 */

void
uploads_add_new_upload (const guint8 *userhash, const gchar *filename, const gchar *username)
{
	GtkTreeIter  iter;
	gchar       *fn_utf8, *un_utf8;

	g_return_if_fail (userhash!=NULL);
	g_return_if_fail (filename!=NULL);
	g_return_if_fail (username!=NULL);
	g_return_if_fail (uploads_store!=NULL);

	gtk_list_store_append (uploads_store, &iter);

	fn_utf8 = TO_UTF8(filename);
	un_utf8 = TO_UTF8(username);

	gtk_list_store_set ( uploads_store, &iter,
	                     UPLOADS_INDEX_COLUMN, uploads_store_items,
	                     UPLOADS_FILENAME_COLUMN, fn_utf8,
	                     UPLOADS_USERNAME_COLUMN, un_utf8,
	                     UPLOADS_USERHASH_COLUMN, g_memdup(userhash,16),
	                     UPLOADS_FILEHASH_COLUMN, g_new0(guint8,16),
	                     -1);

	uploads_store_items++;

	FREE_UTF8(fn_utf8);
	FREE_UTF8(un_utf8);
}


/******************************************************************************
 *
 *   uploads_get_iter_from_slot_id    - for the post-v46 core new upload stuff
 *
 ***/

static gboolean
uploads_get_iter_from_slot_id (guint32 slotid, GtkTreeIter *newiter)
{
	GtkTreeIter  iter;
	gboolean     val;

	g_return_val_if_fail (uploads_store != NULL, FALSE);
	g_return_val_if_fail (newiter != NULL, FALSE);

	val = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(uploads_store), &iter);

	while (val)
	{
		guint32 idx = -1;
		gtk_tree_model_get ( GTK_TREE_MODEL(uploads_store), &iter,
		                     UPLOADS_SLOTID_COLUMN, &idx, -1);
		if (idx == slotid)
		{
			*newiter = iter;
			return TRUE;
		}

		val = gtk_tree_model_iter_next (GTK_TREE_MODEL(uploads_store), &iter);
	}

	return FALSE;
}


/******************************************************************************
 *
 *   uploads_change_upload_slot    - for the post-v46 core new upload stuff
 *
 *   NOTE: takes ownership of filehash and userhash!
 *
 ***/

void
uploads_change_upload_slot ( guint32 slotid,
                             const gchar  *filename,
                             guint8 *filehash,
                             const gchar  *username,
                             guint8 *userhash )
{
	GtkTreeIter  iter;
	gchar       *fn_utf8, *un_utf8;

	g_return_if_fail (filename != NULL);
	g_return_if_fail (filehash != NULL);
	g_return_if_fail (userhash != NULL);
	g_return_if_fail (username != NULL);
	g_return_if_fail (uploads_store != NULL);

	if (uploads_get_iter_from_slot_id(slotid, &iter))
		uploads_remove_iter (gtk_tree_iter_copy (&iter));

	gtk_list_store_append (uploads_store, &iter);

	fn_utf8 = TO_UTF8(filename);
	un_utf8 = TO_UTF8(username);

	gtk_list_store_set ( uploads_store, &iter,
	                     UPLOADS_INDEX_COLUMN, uploads_store_items,
	                     UPLOADS_SLOTID_COLUMN, slotid,
	                     UPLOADS_FILENAME_COLUMN, fn_utf8,
	                     UPLOADS_FILEHASH_COLUMN, filehash,
	                     UPLOADS_USERNAME_COLUMN, un_utf8,
	                     UPLOADS_USERHASH_COLUMN, userhash,
	                     -1);

	uploads_store_items++;

	FREE_UTF8(fn_utf8);
	FREE_UTF8(un_utf8);
}

/******************************************************************************
 *
 *   uploads_clear_all
 *
 *
 */

void
uploads_clear_all (void)
{
	if (uploads_store)
	{
		GtkTreeIter  iter;
		gboolean     val;

		val = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(uploads_store), &iter);

		while (val)
		{
			g_idle_add ( (GSourceFunc) uploads_remove_iter, gtk_tree_iter_copy (&iter));

			val = gtk_tree_model_iter_next (GTK_TREE_MODEL(uploads_store), &iter);
		}
	}
	uploads_store_items = 0;
}


/******************************************************************************
 *
 *  uploads_remove_iter
 *
 *  callback for idle timeout to remove multiple iters when
 *  walking through a store or using foo_foreach();
 *
 ***/

static gboolean
uploads_remove_iter (GtkTreeIter *iter)
{
	guint8   *fhash = NULL;
	guint8   *uhash = NULL;
	guint     index = G_MAXUINT;

	g_return_val_if_fail (iter != NULL, FALSE);

	gtk_tree_model_get ( GTK_TREE_MODEL (uploads_store), iter,
	                     UPLOADS_USERHASH_COLUMN, &uhash,
	                     UPLOADS_FILEHASH_COLUMN, &fhash,
	                     UPLOADS_INDEX_COLUMN, &index,
	                     -1);

	g_return_val_if_fail ( index != G_MAXUINT, FALSE );

	uploads_decrease_all_higher_indices(index);
	uploads_store_items--;

	gtk_list_store_remove ( uploads_store, iter);

	gtk_tree_iter_free (iter);

	G_FREE(fhash);
	G_FREE(uhash);

	return FALSE;
}


/******************************************************************************
 *
 *   uploads_remove_upload
 *
 *
 */

void
uploads_remove_upload (const guint8 *userhash)
{
	GtkTreeIter  iter;
	gboolean     val;

	g_return_if_fail (userhash != NULL);
	g_return_if_fail (uploads_store != NULL);

	val = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(uploads_store), &iter);

	while (val)
	{
		guint8 *uhash = NULL;

		gtk_tree_model_get ( GTK_TREE_MODEL(uploads_store), &iter,
		                     UPLOADS_USERHASH_COLUMN, &uhash,
		                     -1);

		g_return_if_fail ( uhash != NULL );

		if (memcmp(uhash, userhash, 16) == 0)
		{
			uploads_remove_iter (gtk_tree_iter_copy (&iter));

			return;
		}

		val = gtk_tree_model_iter_next (GTK_TREE_MODEL(uploads_store), &iter);
	}

	g_return_if_reached();	/* the userhash should really have been in there */
}



/******************************************************************************
 *
 *   uploads_remove_upload_slot  - post v46 core nwe upload stuff
 *
 ***/

void
uploads_remove_upload_slot (guint32 slotid)
{
	GtkTreeIter  iter;

	g_return_if_fail ( uploads_store != NULL );

	if (!uploads_get_iter_from_slot_id(slotid,&iter))
		return;

	uploads_remove_iter (gtk_tree_iter_copy (&iter));
}

/******************************************************************************
 *
 *  uploads_set_speed
 *
 *  If the upload with the given index number was found,
 *   the iter 'retiter' is set and TRUE is returned,
 *   otherwise FALSE is returned.
 *
 ***/

void
uploads_set_speed (gfloat speed, guint index, guint timediff)
{
	const gchar  *utf8;
	GtkTreeIter   iter;
	const guint8 *filehash = NULL;
	gchar        *filename = NULL;

	if (!uploads_get_iter_from_index(index,&iter))
		return;

	if (speed > 0.1)
	{
		utf8 = UTF8_SHORT_PRINTF(_("%.1f kB/s "), speed);
	}
	else
	{
		utf8 = UTF8_SHORT_PRINTF(_("%.0f byte/s "), (speed*1024.0));
	}

	gtk_list_store_set ( uploads_store, &iter,
	                     UPLOADS_SPEED_COLUMN, speed,
	                     UPLOADS_SPEED_STRING_COLUMN, utf8,
	                     -1);

	if (uploads_record_is_complete (&iter, &filehash, &filename))
	{
		if ((filehash)&&(filename))
		{
			ulstats_add_or_increase_record ( filename, filehash, speed, timediff, 0.0, time(NULL), G_MAXULONG );
		}

		G_FREE(filename);
	}

	return;
}


/******************************************************************************
 *
 *  uploads_set_slot_speed   - for the post-v46 core new upload stuff
 *
 *  If the upload with the given index number was found,
 *   the iter 'retiter' is set and TRUE is returned,
 *   otherwise FALSE is returned.
 *
 ***/

void
uploads_set_slot_speed (gfloat speed, guint32 slotid, guint timediff)
{
	const gchar  *utf8;
	GtkTreeIter   iter;
	const guint8 *filehash = NULL;
	gchar        *filename = NULL;

	if (!uploads_get_iter_from_slot_id(slotid,&iter))
		return;

	if (speed > 0.1)
	{
		utf8 = UTF8_SHORT_PRINTF(_("%.1f kB/s "), speed);
	}
	else
	{
		utf8 = UTF8_SHORT_PRINTF(_("%.0f byte/s "), (speed*1024.0));
	}

	gtk_list_store_set ( uploads_store, &iter,
	                     UPLOADS_SPEED_COLUMN, speed,
	                     UPLOADS_SPEED_STRING_COLUMN, utf8,
	                     -1);

	if (uploads_record_is_complete (&iter, &filehash, &filename))
	{
		if ((filehash)&&(filename))
		{
			ulstats_add_or_increase_record ( filename, filehash, speed, timediff, 0.0, time(NULL), G_MAXULONG );
		}

		G_FREE(filename);
	}
}


/******************************************************************************
 *
 *  uploads_record_is_complete
 *
 *  returns TRUE if we have the upload's filehash, and sets hash and name
 *  (caller needs to free the returned name string, but not the hash)
 *
 *  returns FALSE if we haven't found the filehash
 *
 ***/

static gboolean
uploads_record_is_complete (GtkTreeIter *iter, const guint8 **hash, gchar **name)
{
	const guint8    zerohash[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	guint8         *filehash = NULL;
	gchar          *filename = NULL;
	const guint8   *sharedfilehash = NULL;
	static time_t   last_sent_for_sf_update = (time_t) 0;

	g_return_val_if_fail ( hash != NULL, FALSE );
	g_return_val_if_fail ( name != NULL, FALSE );
	g_return_val_if_fail ( iter != NULL, FALSE );

	gtk_tree_model_get ( GTK_TREE_MODEL(uploads_store), iter,
	                     UPLOADS_FILEHASH_COLUMN, &filehash,   /* is G_TYPE_POINTER */
	                     UPLOADS_FILENAME_COLUMN, &filename,
	                     -1 );

	g_return_val_if_fail ( filehash != NULL, FALSE );
	g_return_val_if_fail ( filename != NULL, FALSE );

	if ( memcmp(filehash, zerohash, 16) != 0 )
	{
		*hash = (const guint8*)filehash;
		*name = filename;	/* caller becomes owner of this copy */
		return TRUE;
	}

	/* we don't have the filehash of the file uploaded yet
	 *  => try  to find hash from the shared files          */
	sharedfilehash = shared_files_get_hash_from_filename (filename);
	if (sharedfilehash)
	{
		memcpy (filehash, sharedfilehash, 16);
		*hash = (const guint8*)filehash;
		*name = filename;	/* caller becomes owner of this copy */
		return TRUE;
	}

	/* haven't found it? Probably our 'shared files' list is not up to date
	 * => refresh it (only if it wasn't refreshed in the last three minutes)      */
	if ( (time(NULL) - last_sent_for_sf_update) > (3*60) )
	{
		send_get_shared_files();
		last_sent_for_sf_update = time(NULL);
	}

	G_FREE(filename);

	return FALSE;
}


/******************************************************************************
 *
 *   uploads_update_rules_hint
 *
 ***/

void
uploads_update_rules_hint (void)
{
	gtk_tree_view_set_rules_hint (GTK_TREE_VIEW(uploads_view), opt_get_bool(OPT_GUI_SET_RULES_HINT_ON_LISTS));
}


/******************************************************************************
 *
 *  uploads_create_store_model_and_view
 *
 ***/

GtkWidget *
uploads_create_store_model_and_view (void)
{
	GtkWidget        	*scrollwin;
	GtkCellRenderer  	*renderer;
	GtkTreeViewColumn	*column;
	GtkTreeSelection 	*selection;

	uploads_store = gtk_list_store_new ( UPLOADS_N_COLUMNS,
	                                     UPLOADS_TYPES_PLACEHOLDER );

	uploads_store_sort = gtk_tree_model_sort_new_with_model (GTK_TREE_MODEL(uploads_store));

	uploads_view = gtk_tree_view_new_with_model (GTK_TREE_MODEL(uploads_store));

	uploads_update_rules_hint();

	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(UTF8_SHORT_PRINTF("%s",_("File being uploaded")),
	                                                  renderer, "text", UPLOADS_FILENAME_COLUMN, NULL);
	gtk_tree_view_column_set_sort_column_id(column, UPLOADS_FILENAME_COLUMN);
	gtk_tree_view_append_column(GTK_TREE_VIEW(uploads_view), column);
	gtk_tree_view_column_set_alignment(column, 0.02);
	g_object_set (G_OBJECT(renderer), "xalign", 0.0, NULL);
	g_object_set (G_OBJECT(renderer), "xpad", 2, NULL);

	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(UTF8_SHORT_PRINTF("%s",_("Who")),
	                                                  renderer, "text", UPLOADS_USERNAME_COLUMN, NULL);
	gtk_tree_view_column_set_sort_column_id(column, UPLOADS_USERNAME_COLUMN);
	gtk_tree_view_append_column(GTK_TREE_VIEW(uploads_view), column);
	gtk_tree_view_column_set_alignment(column, 0.02);
	g_object_set (G_OBJECT(renderer), "xalign", 0.00, NULL);
	g_object_set (G_OBJECT(renderer), "xpad", 2, NULL);

	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(UTF8_SHORT_PRINTF("%s",_("Speed")),
	                                                  renderer, "text", UPLOADS_SPEED_STRING_COLUMN, NULL);
	gtk_tree_view_column_set_sort_column_id(column, UPLOADS_SPEED_COLUMN);

	gtk_tree_view_append_column(GTK_TREE_VIEW(uploads_view), column);
	gtk_tree_view_column_set_alignment(column, 0.9);
	g_object_set (G_OBJECT(renderer), "xalign", 1.0, NULL);
	g_object_set (G_OBJECT(renderer), "xpad", 2, NULL);

	g_object_set_data_full (G_OBJECT(uploads_view), "foo1",
	                        &uploads_view, (GDestroyNotify)g_nullify_pointer);

	g_object_set_data_full (G_OBJECT(uploads_store), "foo1",
	                        &uploads_store, (GDestroyNotify)g_nullify_pointer);

	g_object_set_data_full (G_OBJECT(uploads_store_sort), "foo1",
	                        &uploads_store_sort, (GDestroyNotify)g_nullify_pointer);

	scrollwin = gtk_scrolled_window_new (NULL,NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scrollwin), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER(scrollwin), uploads_view);

	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(uploads_view));
	gtk_tree_selection_set_mode (selection, GTK_SELECTION_NONE);

	gtk_widget_show_all (scrollwin);

	uploads_view_restore_settings();

	return (scrollwin);
}



